!+
! Module: epframes_mod
!
! Authors:  Matthias Ehle, Michael Freyberg (mjf@mpe.mpg.de) - XMM-SSC-MPE
!
! Description: Reformats the EPIC PN events list and auxiliary files into 
!              the file format required by subsequent off-line analysis.
!
! Comments:
!    - simple test calls: export SAS_ODF=.
!      epframes eventset=tmp_e.dat gtiset=tmp_g.set set='OAL(EPN,IMAGING,1,4)'
!      epframes eventset=tmp_e.dat gtiset=tmp_g.set set='OAL(EPN,TIMING,1,4)'
!      epframes eventset=tmp_e.dat gtiset=tmp_g.set set=0205_0112310301_PNS00101IME.FIT
!
!    - list of HK parameters to be read is incomplete (not enough ODFs yet)
!      CEQUADSEL, ...
!    - parameter `qualmax' could be related to event attributes or XMMEA_EP
!    - CAL offset correction should be checked: is CAL doing right (have 
!      raised SPR on CCF) ?
!
!-

module epframes_mod

  use eventAttributes          ! packages/saslib/lib/sasdal/eventattributes.f90
  use cal
  use caloaldefs
  use dal
  use oal
  use param
  use errorHandling
  use sla

  implicit none

  private

! Global variables

  character(len=10), parameter :: CodeVersion = '8.46.11'

  integer, parameter :: maxfrm   = 200000000 ,& ! 1150 cts/s telem. limit
                                                ! and 48 hrs max exp. and
                                                ! min. one event per frame
                        maxbuf   = 12800     ,& ! buffer size
!                       maxwrong = 4000000   ,& ! max. number of invalid events
                        maxiwrong = 1000     ,& ! list of invalid events
                                                !
                        nx_ccd = EPN_MAX_X   ,& ! CCD-columns = 64
                        ny_ccd = EPN_MAX_Y      ! CCD-rows   = 200

  integer(kind=Int32) :: CCDno, &        ! number of selected CCD [1..12]
                         expNr, &        ! selected exposure [OAL convention]
                         expNo, &        ! selected exposure [PNx###]
                         instrument, &   ! selected instrument [EPN]
                         nfrm    , &     ! number of frames in input file
                         mfrm    , &     !   ... in output EXPOSURE extension
                         kfrm    , &     !   ... in output EVENTS extension
                         lfrm    , &     !   ... in input PNAUX2 extension
                         nfrmbad    , &  ! number of bad frames in input file
                         ntrow    , &    ! number of rows in input ODF file 
                         ntaux0    , &   ! number of empty+filled frames
                         ntaux1    , &   ! number of rows in input AUX1 file
                         ntaux2    , &   ! number of rows in input AUX2 file
                         ngaux1    , &   ! good AUX1 frames
                         nfaux1    , &   ! false AUX1 frames
                         n_of_fram    , &! number of frames
                         COLperQUAD, &   ! number of columns per quadrant
                         MaxfrcCOLperQUAD = 0,&! number of columns per quadrant
                         maxgoodreset, & ! AUX1 early time reset frame
                         ngti_aux = 0, & ! AUX1 GTIs
                         ntccx = 0, &    ! number of rows in input CCX file
                         ntpmh = 0, &    ! number of rows in input PMH file
                         ntpah = 0, &    ! number of rows in input PAH file
                         ngrow = 0, &    ! number of rows in output file
                         ngbuf = 0, &    ! number of events in output frame
                         i,j,k, igrow, & ! loop indices
                         oldLim, &       ! FrameLim value of previous frame
                         ndsclinq_ave=0,&! discarded columns total quadrant (av)
                         ndsclinq_20 =0,&! discarded columns total quadrant (20)
                         ndsclinq, &     ! discarded columns total quadrant
                         ndsclinc, &     ! discarded columns single CCD
                         qualMax         ! maximum value of qual for good event

  integer(kind=Int16) :: nxdim, nydim, & ! dimensions of the ODF readout window
                         nx0, ny0, &     ! bottom left corner of window
                         mxdim, mydim, & ! dimensions of the ODF readout window
                         mx0, my0, &     ! bottom left corner of window
                         nxmax, nymax, & ! top right crorner of window
                         quadrant,     & ! HK quadrant number [0..3]
                         CCDid   ,     & ! HK CCD number [0..2]
                         revolution,   & ! revolution number
                         loopflag1,    & ! 
                         loopflag2,    & ! flags for first time entering a loop
                         warnFlag, &     ! set flag if wrongpixlimit reached
                         sourceY, &      ! source position for TIMING/BURST
                         sourceYin, &    ! source position for TIMING/BURST
                         MIP_Threshold, &! maximum amplitude for non-MIPs
                         MIP_out, &      ! output method for MIP-map
                         NOMIP_loth, &   ! minimum amplitude for non-crmap
                         NOMIP_hith, &   ! maximum amplitude for non-crmap
                         SOFT_loth, &    ! minimum amplitude for non-crmap-soft
                         SOFT_hith       ! maximum amplitude for non-crmap-soft

  integer(kind=Int16), parameter    ::    evt_phaMin = 30 ! adu
  integer(kind=Int32) :: dliarray(1:ny_ccd), evtarray(1:ny_ccd), &
                         dlisasarray(1:ny_ccd), dli_sum, evt_sum
  real(kind=double)   :: dli_mean, dli_median, dli_var, dli_sigma

  real(kind=double)   :: srcra, srcdec      ! source RA/DEC for TI+BU modes
  real(kind=double)   :: objra, objdec      ! source RA/DEC for TI+BU modes
! real(kind=double)   :: nomra, nomdec      ! source RA/DEC for TI+BU modes
  logical(kind=bool)  :: withsrccoords      ! use source RA/DEC for TI+BU modes
  logical(kind=bool)  :: withsrcrawy        ! use source RAWY for TI+BU modes
  integer             :: xrawsrc, yrawsrc


  logical(kind=bool)  :: automode = .false.   ! determine F1294 and An_CHOP
  logical(kind=bool)  :: autofilter = .false. ! determine FILTER

  real(kind=double), dimension(:), allocatable :: deb, fin          ! GTIs
  integer(kind=Int32) :: ngti, nmipcol

  integer(kind=Int8), parameter    ::    offcollimit = 0
  integer(kind=Int8)  :: ocv(0:nx_ccd-1)   ! offset correction vector 
  logical(kind=bool)  :: on_offcol(0:nx_ccd-1), next_offcol(0:nx_ccd-1)

  integer             :: inmode, insubmode ! readout mode of the current CCD
  integer             :: ccdmode           ! readout mode of the current CCD

  integer(kind=Int32)                        :: ncolOffset
  integer(kind=Int32), dimension(1)          :: dims
  integer(kind=Int32), dimension(2)          :: imageDim


  real(kind=double),   dimension(:), pointer :: &
                       outtimeCol, timeTags, exptimeCol

  real(kind=single),   dimension(:), pointer :: &
                       fracexpCol, timedelCol

  integer(kind=Int32), dimension(:), pointer :: &
                       frameCol, outframeCol, outqualflagCol, frames, &
                       outhistRCol, outhistLCol, frameLimCol, &
                       aux1FrameCol, aux2CycleCol, &
                       outdliodfCol, outdlisasCol

  integer(kind=Int32), dimension(:,:), pointer :: &
                       mip_arrayP, nomip_arrayP, total_arrayP, soft_arrayP

  integer(kind=Int16), dimension(:,:), pointer :: &
                       dli_arrayP

  integer(kind=Int16), dimension(:), pointer :: &
                       aux1ftcoarseCol, aux1ftfineCol, aux2NdisclinCol, &
                       ccx1StcoarseCol, ccx1StfineCol, ccx1EtcoarseCol, &
                       ccx1EtfineCol, ccx1QuadrantCol, ccx1NCycleCol, &
                       outrawXCol, outrawYCol, phaCol, outphaCol, &
                       ndsclin1Col, ndsclin2Col, ndsclin3Col, ndsclin4Col

  integer(kind=Int8),  dimension(:), pointer :: &
                       aux1CcdidCol, aux1QuadrantCol, aux2QuadrantCol, & 
                       rawXCol, rawYCol, outccdnrCol, &
                       outmipdistposCol, outmipdistnegCol, outdistCol


  real(kind=double),   dimension(:), allocatable :: &
                       arrExpTime, aux1GTime, aux2GTime, &
                       aux1gti_b, aux1gti_e, &
                       ccx1Duration, arrFracExp, dummy64

  integer(kind=Int32), dimension(:), allocatable :: &
                       aux1Frame, aux2Cycle, aux1FtTime, aux1GFrame, &
                       aux2gti_b, aux2gti_e, &
                       aux1Bindex_s,aux1Bindex_e, &
                       arrExpFrame, arrFrame, arrFrames, &
                       arrFramelim, arrQual, &
                       qual, ind, iwrong

  integer(kind=Int16), dimension(:), allocatable :: &
                       arrRawX, arrRawY, arrPha, &
                       arrNdsclin1, arrNdsclin2, arrNdsclin3, arrNdsclin4, &
                       aux1Ftcoarse, aux1Ftfine, aux2Ndisclin, &
                       ixx, d_rawXCol, iyy, d_rawYCol, ampl, &
                       ccx1Stcoarse, ccx1Etcoarse, ccx1Stfine,ccx1Etfine, &
                       ccx1NCycle

  integer(kind=Int8),  dimension(:), allocatable :: &
                       aux1FtQual, aux1ccdNr, calRawX, &
                       arrDistP, arrDistN, distp, distn


  integer(kind=Int32):: nwrongCoord, nwrongOrder, nwrongTime, nMIP, &
                        nwrongFTCoarse, nwrongPHA, &
                        mwrong, nwrong, kwrong, wronglimit, &
                        frameNr_first, frameNr_last, &
                        mip_array(1:nx_ccd,1:ny_ccd), &
                        nomip_array(1:nx_ccd,1:ny_ccd), &
                        total_array(1:nx_ccd,1:ny_ccd), &
                        soft_array(1:nx_ccd,1:ny_ccd), &
                        dist_2_mip_hist_r(1:nx_ccd), &
                        dist_2_mip_hist_l(1:nx_ccd)

  integer(kind=Int32):: An_CMLOTH, An_CMCORR, An_EAMIPSEL, An_MAXMIP, An_MIP
  integer(kind=Int32):: HK_CMLOTH, HK_CMCORR, HK_EAMIPSEL, HK_MAXMIP, HK_MIP
  integer(kind=Int32):: F1294, F1052, An_CHOP, An_MAXFRC
  integer(kind=Int32):: OT_CMCORR, OT_CMLOTH, OT_MIP, OT_MAXMIP, OT_EAMIPSEL
  integer(kind=Int32):: OT_CHOP
! integer(kind=Int32):: C_CE_QUADSEL, An_CCD_SWISTAT
  integer(kind=Int32):: Ax_QUAD_CCD_SEL, Ax_LowThresh
  integer(kind=Int32):: F1030,F1030_1,F1030_2
  integer(kind=Int32):: Frames_OffsetMap

  real(kind=single)  :: An_DSLINC_MIN,An_DSLINC_MAX,An_DSLINC_AVE
  real(kind=single)  :: F1122MIN, F1122MAX, F1122AVE
  real(kind=single)  :: F1128MIN, F1128MAX, F1128AVE
  real(kind=single)  :: E_Cx_TEMPQB1_MIN, E_Cx_TEMPQB1_MAX, E_Cx_TEMPQB1_AVE
  real(kind=single)  :: E_Cx_IGRAx_MIN, E_Cx_IGRAx_MAX, E_Cx_IGRAx_AVE
  real(kind=single)  :: V_VC_UHV_MIN, V_VC_UHV_MAX, V_VC_UHV_AVE
  real(kind=single)  :: fc0, fc3

  integer(kind=Int16):: ODF_offset      ! to correct for incorrect ODF PHAs

  real(kind=double)  :: pwrong, deltat, deltat_odf, halfdt, &
                        frmtime, frmtime_seconds, AttTimeTag, &
                        ontime, livetime, rotime, psftime, shftime, &
                        T_Counting_Mode, TimeOffset, pha_offset_time, &
                        good_time_first, good_time_last, oldValue, &
                        telapse, duration, delta_duration

  character(len=1024):: pmhfilename, pahfilename, ccxfilename, &
                        dlifilename, eventmapset, sciencefilename, &
                        messagestr
  character(len=80)  :: history, map_title
  character(len=64)  :: dashline, ddashline
  character(len=32)  :: filter, filter_hk, obsSubMode, enable_of_empty_frames
  character(len=25)  :: date_obs,date_end, obs_start,obs_stop, toffs,toffe
  character(len=24)  :: odf_id
  character(len=22)  :: odf_icd
  character(len=12)  :: An_SENDMODE, SAS_SENDMODE, OT_SENDMODE, &
                        ECACHn_GAINCAME, &
 			                    BPT_CODE     = INVALID_BAD_PIXEL_TABLE_CODE, &
 			                    BPT_CODE_CCF = INVALID_BAD_PIXEL_TABLE_CODE
  integer(kind=Int16):: BPT_CODE_MIN_REV = 293    ! first rev of implementation
  character(len=8)   :: MIP_Method, extname_dli, colname_dli1, colname_dli2
  character(len=4)   :: setup_OCV, ExpIdStr
  character(len=3)   :: setup_SAS, Median_OffsetMap

  logical(kind=bool) :: log, lxcolStore(1:nx_ccd), &
                        frame_array(0:nx_ccd+1,0:ny_ccd+1), &
                        after_mip_frame

  logical(kind=bool), parameter :: only_single      = .TRUE.   ! MIP check
  logical(kind=bool), parameter :: only_after_mip   = .FALSE.  ! MIP check
  logical(kind=bool), parameter :: also_timedel_col = .FALSE.  ! only keyword
  logical(kind=bool), parameter :: empty_frames     = .TRUE.   ! enable
  logical(kind=bool), parameter :: empty_frames_GTI = .TRUE.   ! enable
  logical(kind=bool), parameter :: verify_PMH       = .FALSE.  ! disable
! logical(kind=bool), parameter :: verify_PMH       = .TRUE.   ! enable
  logical(kind=bool), parameter :: enable_PSFtime   = .TRUE.   ! enable
  logical(kind=bool), parameter :: enable_AUX2dtime = .TRUE.   ! enable
  logical(kind=bool), parameter :: enable_offcol_extension = .TRUE. ! enable
  logical(kind=bool), parameter :: enable_dli_extension = .TRUE.    ! enable
  logical(kind=bool), parameter :: enable_OAL_latest = .TRUE.       ! enable
!
! oal-3.92 can handle DISCARDED_LINES data correctly now
!
! logical(kind=bool), parameter :: enable_DISCARDED_LINES = .FALSE.  !
  logical(kind=bool), parameter :: enable_DISCARDED_LINES = .TRUE.   ! enable
  logical(kind=bool), parameter :: enable_DATA_DIAGNOSTIC = .FALSE.  !
! logical(kind=bool), parameter :: enable_DATA_DIAGNOSTIC = .TRUE.   ! enable

  logical(kind=bool) :: CAL_correct_finetime

  logical            :: mipcorr, &       ! logical code of MIP correction
                        mipcols, &       ! logical code of MIP neighbours flag
                        mipnew , &       ! logical code of MIP new onboard SW
                        MIP_ext,&        ! logical code of MIP extension
                        MIP_dist, &      ! logical code of MIP distances
                        MIP_discard, &   ! logical code of MIP omission
                        witheventmap     ! 

  logical            :: correct_ODF, correct_CCF, correct_HK, &
                        show_CCX, show_AUX, show_PMH, show_PAH, show_VE, &
                        exist_ccx, exist_pah, exist_pmh, exist_aux, &
                        exist_dli, bad_dli, illegal_ODF = .false.

  logical            :: scheduled, noPNAUX2, singleBoresightBurst, &
                        doubleWindowTiming, wrongBPTcode, illegalODF, &
			                     CCDlowGain, isFastMode

!x  logical(kind=bool) :: first_GTI = .true.
!x  logical(kind=bool) :: current_GTI = .false.

  logical             :: wrong_srcpos
  logical(kind=bool)  :: tooManyEmptyFrames, noGTIfound, tooManyEvents
  integer(kind=Int32) :: maxManyEvents = 50000000  ! 5000000

  real(kind=double), parameter  :: twopi = 6.283185307179586477D0
  real(kind=double), parameter  :: deg_to_rad = twopi/360.d0

  type(DataSetT)           :: infile, outfile, outgti, &
                              outmip, outnomip, outsoft, outtotal, &
                              dlifile

  type(TableT)             :: inccd, outccd, outexpo, outhist, outoffcol, &
                              outdlimap

  type(BlockT)             :: inblk

  type(ColumnT)            :: ccdnr, rawX, rawY, time, pha, qualflag, &
                              frame, tempInCol, &
                              exptime, fracexp, timedel, framelim, &
                              ndsclin1, ndsclin2, ndsclin3, ndsclin4, &
                              mipdistl, mipdistr, &
                              miphistl, miphistr, miphdist, &
                              dliodf, dlisas

  type(ArrayT)             :: mip, nomip, total, soft, dli

  type(OdfInfoType)        :: odfinfos
  type(ProposalInfoType)   :: proposalinfo
  type(ExposureInfoType)   :: expinfos

  public epframes

contains


!========================================================================
  subroutine epframes
!========================================================================
  type(EquatorialDirectionType),dimension(0:0) :: sourceDir
  character(len=1024)                          :: fileName, auxfilename
  character(len=80)                            :: fileMap

  real(kind=double)  ,dimension(:),allocatable :: itime

  real(kind=single),  dimension(:),pointer     :: eInReal, eOutReal
  integer(kind=int16),dimension(:),pointer     :: eOut
  integer(kind=int16),dimension(:),allocatable :: eIn

  integer(kind=Int32)                          :: ifrm, irow, nbuf, iend, &
                                                  frame_nr, iaux, nfff, &
                                                  ix, iy, &
                                                  oldframe !, newframe

  logical(kind=bool)                           :: aux1alloc, aux2alloc, &
                                                  ccxalloc, &
                                                  pmhalloc, pahalloc, &
                                                  goodtime, &
                                                  goodauxtime, &
                                                  frameCounterError, &
                                                  frameEventError, &
                                                  wrongReadoutOrder, &
                                                  wrongCoord

  integer(kind=Int32)                          :: wrongCoordN, &
                                                  wrongReadoutOrderN
  integer(kind=Int32),parameter                :: wrongCoordMax = 2, &
                                                  wrongReadoutOrderMax = 10


! =============================================================================
! Start the programm
! =============================================================================

  write(errstr,*) 'EPFRAMES ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  write(dashline,'(64a)')  ('-',i=1,64)
  write(ddashline,'(64a)') ('=',i=1,64)

!
! initialize all variables here instead of at the declaration for metatask usage
!
  dliarray(1:ny_ccd) = 0
  dlisasarray(1:ny_ccd) = 0
  evtarray(1:ny_ccd) = 0
  nfrm = 0
  mfrm = 0
  ntrow = 0
  ntaux1 = 0
  ntaux2 = 0
  ntaux0 = 0
  ngaux1 = 0
  nfaux1 = 0
  n_of_fram = 0
  ngti_aux = 0
  ntccx = 0
  ntpmh = 0
  ntpah = 0
  ngrow = 0
  ngbuf = 0
  ndsclinq_ave = 0
  ndsclinq_20 =0
  F1122MIN = -9999.9
  F1122MAX = -9999.9
  F1122AVE = -9999.9
  F1128MIN = -9999.9
  F1128MAX = -9999.9
  F1128AVE = -9999.9
  E_Cx_TEMPQB1_MIN = -9999.9
  E_Cx_TEMPQB1_MAX = -9999.9
  E_Cx_TEMPQB1_AVE = -9999.9
  E_Cx_IGRAx_MIN = -9999.9
  E_Cx_IGRAx_MAX = -9999.9
  E_Cx_IGRAx_AVE = -9999.9
  V_VC_UHV_MIN = -9999.9
  V_VC_UHV_MAX = -9999.9
  V_VC_UHV_AVE = -9999.9
  fc0 = 1.0
  fc3 = 1.0
 	BPT_CODE     = INVALID_BAD_PIXEL_TABLE_CODE
 	BPT_CODE_CCF = INVALID_BAD_PIXEL_TABLE_CODE
  ExpIdStr = 'X000'
  messagestr = ' '

! ---------------------------------------------------------
! Read the parameters and determine relevant files
! ---------------------------------------------------------

  call getParamValues(fileName,auxfilename,CCDno,instrument,expNr,inmode,filter)

  read(odf_icd(19:22),'(a4)') ExpIdStr
  read(odf_icd(1:4),'(i4)') revolution

! ------------------------------------------------------------------------
! Set the window parameters according to documentation: check against ODF
! ------------------------------------------------------------------------
  !
  ! EPN data modes          [taken from $SAS_DIR/include/caloaldefs.f90 1.35]
  ! EMOS/EPN
  !   IMAGING                           =  0 IME
  !   TIMING                            =  1 TIE
  !   REDUCED_IMAGING                   =  2
  !   COMPRESSED_TIMING                 =  3
  !   AUXILIARY                         =  4 AUX
  !   DATA_DIAGNOSTIC                   =  5
  !   DATA_OFFSET_VARIANCE              =  6
  !   COUNTING_CYCLE_REPORT             =  7 CCX
  !   PERIODIC_HOUSEKEEPING             =  8
  !   HBR_CONFIG_HOUSEKEEPING           =  9 HCH
  !   HBR_BUFFER_HOUSEKEEPING           = 10
  !   HBR_THRESHOLD_HOUSEKEEPING        = 11
  !   EXTRAHEATING_CONF_HOUSEKEEPING    = 12
  !   THERMAL_HOUSEKEEPING              = 13
  !   BRIGHT_PIXEL_HOUSEKEEPING         = 14
  !
  ! EPN not in above list
  !   BURST                             = 15 BUE
  !   OFFSET_DATA                       = 16 ODI
  !   NOISE_DATA                        = 17 NOI
  !   DISCARDED_LINES_DATA              = 18 DLI
  !   MAIN_PERIODIC_HOUSEKEEPING        = 19 PMH
  !   ADD_PERIODIC_HOUSEKEEPING         = 20 PAH
  !
  ! EPN instrument modes                     [internal SAS CAL modes]   
  !   PRIME_FULL_WINDOW                 =  0     0
  !   PRIME_FULL_WINDOW_EXTENDED        = 15     1
  !   PRIME_LARGE_WINDOW                = 16     2
  !   PRIME_SMALL_WINDOW                = 17     3
  !   FAST_TIMING                       = 18     4
  !   FAST_BURST                        = 19     5
  !   PRIME_FULL_OFFSET                 = 20
  !   PRIME_LARGE_OFFSET                = 21
  !   OFFSET                            = 22
  !   NOISE                             = 23
  !   DIAGNOSTIC                        = 24
  !   EXTRA_HEATING_DEICING             = 12
  !   EXTRA_HEATING_DECONTAM            = 13
  !   IN_FLIGHT_TEST                    = 14

  select case (insubmode)
    case(PRIME_FULL_WINDOW)
      mx0   = 1
      my0   = 2
      mxdim = 64
      mydim = 199
      COLperQUAD = 3*nx_ccd
      maxgoodreset = 2
    case(PRIME_FULL_WINDOW_EXTENDED)
      mx0   = 1
      my0   = 2
      mxdim = 64
      mydim = 199
      COLperQUAD = 3*nx_ccd
      maxgoodreset = 2
    case(PRIME_LARGE_WINDOW)
      mx0   = 1
      my0   = 102
      mxdim = 64
      mydim = 99
      COLperQUAD = 3*nx_ccd
      maxgoodreset = 4
    case(PRIME_SMALL_WINDOW)
      mx0   = 1
      my0   = 137
      mxdim = 64
      mydim = 64
      COLperQUAD = nx_ccd
      maxgoodreset = 11
    case(FAST_TIMING)
      mx0   = 1
      my0   = 2
      mxdim = 64
      mydim = 199
      COLperQUAD = nx_ccd
      maxgoodreset = 100
    case(FAST_BURST)
      mx0   = 1
      my0   = 2
      mxdim = 64
      mydim = 179
      COLperQUAD = nx_ccd
      maxgoodreset = 100
    case default
      call error('wrongMode','caseDefaultError: window')
  end select
  MaxfrcCOLperQUAD = An_MAXFRC * COLperQUAD

  if (insubmode == FAST_TIMING .or. insubmode == FAST_BURST) then
    isFastMode = .true.
  else
    isFastMode = .false.
  endif
  scheduled            = .true.
  wrong_srcpos         = .true.
  noPNAUX2             = .false.
  CCDlowGain           = .false.
  singleBoresightBurst = .false.
  doubleWindowTiming   = .false.
  wrongBPTcode         = .false.
  illegalODF           = .false.

  sourceDir(0)%ra  = DEG_TO_RAD*srcra
  sourceDir(0)%dec = DEG_TO_RAD*srcdec
  objra            = DEG_TO_RAD*proposalinfo%targetRa
  objdec           = DEG_TO_RAD*proposalinfo%targetDec
  if (.NOT.withsrccoords) then
    sourceDir(0)%ra  = objra
    sourceDir(0)%dec = objdec
  endif
  sourceYin = sourceY

  if (empty_frames) then
    enable_of_empty_frames = 'enabled'
  else
    enable_of_empty_frames = 'not enabled'
  endif
  write(errstr,*) 'handling of empty frames: ',enable_of_empty_frames
  call message(AppMsg,Verbose)


! ---------------------------------------------------------------------------
! Declarations for the DAL, open input events file
! Developer's Note: next routine grew by far too large, therefore it should
!                   be split into smaller pieces in one of the next versions
! ---------------------------------------------------------------------------

  call declareDal(mxdim,mydim,mx0,my0,ntrow,nxdim,nydim,nx0,ny0,&
                  frameNr_first,frameNr_last,insubmode)

!-----------------------------------------------------------
! access and release AUX file, return arrays 'aux1*' 'aux2*'
!-----------------------------------------------------------

  ndsclinq_ave = 0
  ndsclinq = 0
  ndsclinc = 0
  noGTIfound = .true.
  CAL_correct_finetime = .true.

  if (exist_aux) then
    call readAux(auxfilename,ntaux0,ntaux1,ntaux2,aux1alloc,aux2alloc,&
       CCDNo,maxgoodreset)
  else
    aux1alloc = .false.
    aux2alloc = .false.
  endif


! ------------------------------------------------------------
! get the integration time per frame for fine time correction
!   -> try to compute frame time (= time between time words)
! ------------------------------------------------------------

  if (verbosity(AppMsg) > Noisy) call printNominalFrameTime(f1294)

  call getNominalFrameTime(insubmode,f1294,fc0,frmtime,rotime,psftime,shftime)
  frmtime_seconds = frmtime*1.d-3
  deltat = frmtime


!
! should be converted to a WARNING once somewhat correct ODFs are available
! and frame time [ms] and integration times are confirmed
! However, frame times given in ODF are completely wrong (sometimes 2 or 3 ms
! instead of 73.36 ms for FF mode), therefore only use nominal values !
!

  if (abs((deltat_odf-deltat)) > 0.5) then
! if (abs((deltat_odf-deltat)/deltat) > 0.001) then
    if (correct_ODF) then
      write(errstr,'(1x,a,1x,a,f9.4,a,f9.4,a)') &
                'Frame time in ODF does not match nominal value. Check ODF !',&
                'Nominal: ',deltat,' [ms],   ODF: ',deltat_odf,' [ms]'
      call message(AppMsg,Sparse)
!
! apparently the ODS system is not able to produce correct header keywords
! so it is still worthwhile to mention this, however, to reduce the number
! of warnings this has now been turned down to a "sparse message" (2002-01-17)
!
!     call warning('frametimeMismatch')
    else
      call message(AppMsg,Verbose, &
                ' Frame time of ODF does not match nominal value. Check ODF !')
      write(errstr,'(4x,a,f9.4,a,f9.4,a)') &
                'Nominal: ',deltat,' [ms],   ODF: ',deltat_odf,' [ms]'
      call message(AppMsg,Verbose)
    endif
  endif
! deltat = deltat_odf

!
! convert from [ms] to [s]
!
  shftime = shftime*1.d-3
  deltat  = deltat*1.d-3
  halfdt  = deltat*0.5d0


! -------------------------------------------
! fill in mode dependent lifetime factor FC0
! -------------------------------------------

  if (fc0 < 0) then
    call warning('wrongFC0','No mode dependent lifetime factor, set FC0 = 1')
    FC0 = 1.0
  endif
  FC3 = 1./(1+An_CHOP)


! ------------------------------------------------------
! access and release PMH, PAH, CCX files, return arrays
! ------------------------------------------------------

  toffs = " "
  toffe = " "
  if (exist_pmh) then
    call readPMH(pmhfilename,ntpmh,pmhalloc,Quadrant,CCDid,&
                 date_obs,date_end,toffs,toffe)
  else
    pmhalloc = .false.
  endif

  if (exist_pah) then
    call readPAH(pahfilename,ntpah,pahalloc,Quadrant,CCDid,&
                 date_obs,date_end,toffs,toffe)
  else
    pahalloc = .false.
  endif

  if (An_EAMIPSEL .le. 0 .or. An_EAMIPSEL .gt. 3) then
    if (HK_EAMIPSEL .ne. An_EAMIPSEL) then
      write(errstr,*) 'PAH file gives EAMIPSEL =',An_EAMIPSEL, &
             ', will use fallback by parameter value =',HK_EAMIPSEL
      call message(AppMsg,Sparse)
      An_EAMIPSEL = HK_EAMIPSEL
    endif
  endif

  if (An_MIP .le. 3000 .or. An_MIP .gt. 4000) then
    if (HK_MIP .ne. An_MIP) then
      write(errstr,*) 'PAH file gives MIP =',An_MIP, &
             ', will use fallback by parameter value =',HK_MIP
      call message(AppMsg,Sparse)
      An_MIP = HK_MIP
      if (HK_CMCORR .ne. An_CMCORR) then
        write(errstr,*) 'PAH file gives CMCORR =',An_CMCORR, &
             ', will use fallback by parameter value =',HK_CMCORR
        call message(AppMsg,Sparse)
        An_CMCORR = HK_CMCORR
      endif
    endif
  endif

  if (An_MAXMIP < 63) then
    if (HK_MAXMIP .ne. An_MAXMIP) then
      write(errstr,*) 'PAH file gives MAXMIP =',An_MAXMIP, &
             ', will use fallback by parameter value =',HK_MAXMIP
      call message(AppMsg,Sparse)
      An_MAXMIP = HK_MAXMIP
    endif
  endif

  if (trim(An_SENDMODE) == "MIP CORREC.1") then
    call warning('MIPrejection','MIP CORREC.1: row-wise MIP rejection is unusual.')
  endif

! ------------------------------------------------------------------
! read the BadPixelTable code from HK and CCF file and do comparison
! ------------------------------------------------------------------

  BPT_CODE_CCF = CAL_getAduconvCode()
  if (verbosity(AppMsg) > Verbose) then
    write(errstr,*) 'BPT_CODE_CCF = ',BPT_CODE_CCF(1:9)
    call message(AppMsg,Sparse)
    write(errstr,*) 'BPT_CODE     = ',BPT_CODE(1:9)
    call message(AppMsg,Sparse)
    write(errstr,*) 'BPT_CODE_UNK = ',INVALID_BAD_PIXEL_TABLE_CODE(1:9)
    call message(AppMsg,Sparse)
  endif
  if (REVOLUTION < BPT_CODE_MIN_REV) then
    BPT_CODE = INVALID_BAD_PIXEL_TABLE_CODE
    write(errstr,*) 'BadPixelTable code not implemented before revolution ', &
                     BPT_CODE_MIN_REV,' [here Rev =',revolution,']'
    call message(AppMsg,Verbose)
  elseif (BPT_CODE_CCF(1:9) == INVALID_BAD_PIXEL_TABLE_CODE(1:9)) then
    write(errstr,*) 'BadPixelTable code ',BPT_CODE(1:12), &
                    ' is not coded in ADUCONV.CCF'
    call message(AppMsg,Sparse)
  elseif (BPT_CODE_CCF(1:9) .ne. BPT_CODE(1:9)) then
    write(errstr,*) 'BadPixelTable code differs in PAH.FITS [', &
          BPT_CODE(1:9),'] and ADUCONV.CCF [',BPT_CODE_CCF(1:9),']'
    call warning('BPTcode')
    call message(AppMsg,Sparse,' ')
  else
    write(errstr,*) 'BadPixelTable code: ',BPT_CODE(1:12)
    call message(AppMsg,Sparse)
  endif

  if (exist_ccx) then
    call readCCX(ccxfilename,ntccx,ccxalloc,Quadrant,T_Counting_Mode)
  else
    ccxalloc = .false.
  endif

! -----------------------------------
! print information about offset maps
! -----------------------------------

  if (OT_CHOP > -1) then
    write(errstr,*) 'Offset table calculation time interval: ',&
             trim(toffs), ' - ', trim(toffe)
    call message(AppMsg,Sparse)
    write(errstr,'(1x,a,i4,a,i2,a,a)') '  calculation with ',Frames_OffsetMap, &
                    ' frames and chopper = ',OT_CHOP, &
                    ' with median correction = ',trim(Median_OffsetMap)
    call message(AppMsg,Sparse)
    write(errstr,*) '  offset = ',OT_CMCORR, &
              '  lower threshold = ',OT_CMLOTH, &
              '  MIP rejection method: ',OT_SENDMODE
    call message(AppMsg,Sparse)
    if (index(OT_SENDMODE,'NO') > 0) then
    else
      write(errstr,*) '  MIP threshold = ', OT_MIP, &
              '  maximum number = ', OT_MAXMIP, &
              '  neighbouring columns/lines = ',OT_EAMIPSEL
      call message(AppMsg,Sparse)
    endif
  endif


!----------------------------------------------------
! determine raw pixel coordinates for source position
!----------------------------------------------------

  if (withsrccoords.or..NOT.withsrcrawy) then
    call SourcePositionOnDetector(sourceDir,AttTimeTag,xrawsrc,yrawsrc)
    if (xrawsrc .ge. 1 .and. xrawsrc .le. EPN_MAX_X .and. &
        yrawsrc .ge. 1 .and. yrawsrc .le. EPN_MAX_Y) then
      wrong_srcpos = .false.
      sourceY = yrawsrc
      if (withsrccoords) then
        write(errstr,'(1x,a,f10.6,a,f10.6,a,i2,a,i3,a)')    &
             'Requested source position (',srcra,',',srcdec,   &
             ') corresponds to RAW pixel (',xrawsrc,',',yrawsrc,') of CCD 4'       
      else
        write(errstr,'(1x,a,f10.6,a,f10.6,a,i2,a,i3,a)')    &
             'Requested source position (',proposalinfo%targetRa,&
             ',',proposalinfo%targetDec,   &
             ') corresponds to RAW pixel (',xrawsrc,',',yrawsrc,') of CCD 4'       
      endif
      call message(AppMsg,Sparse)
    else
      if (isFastMode) then
        call warning("srcpos", &
               "Could not be determined from attitude, outside chip boundaries")
        wrong_srcpos = .true.
        write(errstr,'(1x,a,f11.7,a,f11.7)') &
          ' SourceDir  RA = ',sourceDir(0)%ra/deg_to_rad,&
          '  DEC = ',sourceDir(0)%dec/deg_to_rad
        call message(AppMsg,Sparse)
        write(errstr,'(1x,a,f11.7,a,f11.7)') &
          ' SourcePos  RA = ',srcra, '  DEC = ',srcdec
        call message(AppMsg,Sparse)
        write(errstr,'(1x,a,f11.7,a,f11.7)') &
          ' TargetDir  RA = ',proposalinfo%targetRa,&
          '  DEC = ',proposalinfo%targetDec
        call message(AppMsg,Sparse)
      endif
    endif
  else
    wrong_srcpos = .false.
    sourceY = sourceYin
  endif

!-----------------------------------------------------------
! access and release AUX file, return arrays 'aux1*' 'aux2*'
!-----------------------------------------------------------
!-- this part of the code has been moved

! -----------------------------------------------
! Determine frame boundaries of (existing) events
! -----------------------------------------------

  call defineFrames

  iend        = -1
  ngrow       = 0
  igrow       = 0
  ngbuf       = 0
  kfrm        = 0
  lfrm        = 0

! initialize statistic counters:

  nwrongFTCoarse = 0
  nwrongPHA   = 0
  nwrongCoord = 0
  nwrongOrder = 0
  nwrongTime  = 0
  nMIP        = 0
  nfrmbad     = 0
  nfff        = 0  ! not used yet
  nmipcol     = 0
  wrongCoordN        = 0
  wrongReadoutOrderN = 0

  loopflag1 = 0
  loopflag2 = 0
  nwrong    = 0
  kwrong    = 0
  warnFlag  = 0

  if (ntrow > 0) then

! ------------------------------------------------------
! determine PHA offset due to incorrectly processed ODFs
! ------------------------------------------------------

! reprocessing at Vilspa had introduced newly incorrect PHA values
! therefore make check also for "nominal correct ODFs"
!
!   if (.not. correct_ODF) &
    call ODF_offsetPHA(phaCol,ntrow,ODF_icd,ODF_offset)
!
! allocate output arrays
!
    call message(AppMsg,Noisy,"... allocating output event arrays ...")

    allocate(arrRawX(0:ntrow-1),arrRawY(0:ntrow-1),arrFrame(0:ntrow-1))
    allocate(arrPha(0:ntrow-1),arrQual(0:ntrow-1))
    if (MIP_dist) allocate(arrDistP(0:ntrow-1),arrDistN(0:ntrow-1))

    arrRawX  = 0
    arrRawY  = 0
    arrFrame = 0
    arrPha   = 0
    arrQual  = 0

    call message(AppMsg,Noisy,"... allocating output frame arrays ...")
    if (empty_frames) then
      allocate(arrFracExp(0:mfrm-1), arrFramelim(0:mfrm-1),&
               arrNdsclin1(0:mfrm-1),arrNdsclin2(0:mfrm-1),&
               arrNdsclin3(0:mfrm-1),arrNdsclin4(0:mfrm-1),&
               arrFrames(0:nfrm-1))
      if (.NOT. enable_AUX2dtime) then
        arrFracExp(0:mfrm-1) = 1.0d0-dble(ndsclinq_ave)/maxfrcCOLperQUAD
      else
        arrFracExp(0:nfrm-1) = 1.0d0
      endif
    else
      allocate(arrFracExp(0:nfrm-1), arrFramelim(0:nfrm-1),&
               arrNdsclin1(0:nfrm-1),arrNdsclin2(0:nfrm-1),&
               arrNdsclin3(0:nfrm-1),arrNdsclin4(0:nfrm-1))
      arrFracExp(0:nfrm-1) = 1.0d0
    endif

    arrNdsclin1 = 0
    arrNdsclin2 = 0
    arrNdsclin3 = 0
    arrNdsclin4 = 0
    arrFramelim = 0

    allocate(iwrong(0:maxiwrong+1))
    iwrong(0:maxiwrong+1) = 0

    lxcolStore(1:nx_ccd)           = .true.
    mip_array(1:nx_ccd,1:ny_ccd)   = 0
    nomip_array(1:nx_ccd,1:ny_ccd) = 0
    soft_array(1:nx_ccd,1:ny_ccd)  = 0
    total_array(1:nx_ccd,1:ny_ccd) = 0
    frame_array(0:nx_ccd+1,0:ny_ccd+1) = .false.

! --------------------------------------------------------------------------
! Read table columns IX or IY, depending on MIP rejection scheme An_SENDMODE
! Start loop over all frames
! --------------------------------------------------------------------------

    if (trim(An_SENDMODE) == 'MIP CORREC.2') then
      do i = 0,ntrow-1
        ix = d_rawXCol(i)+1
        if (ix .ge. 1 .and. ix .le. nx_ccd .and. phaCol(i) .ge. evt_phaMin) &
           evtarray(ix) = evtarray(ix)+1
      enddo
    elseif (trim(An_SENDMODE) == 'MIP CORREC.1') then
      do i = 0,ntrow-1
        iy = d_rawYCol(i)+1
        if (iy .ge. 1 .and. iy .le. ny_ccd .and. phaCol(i) .ge. evt_phaMin) &
           evtarray(iy) = evtarray(iy)+1
      enddo
    elseif (trim(SAS_SENDMODE) == 'MIP CORREC.2') then
      do i = 0,ntrow-1
        ix = d_rawXCol(i)+1
        if (ix .ge. 1 .and. ix .le. nx_ccd .and. phaCol(i) .ge. evt_phaMin) &
           evtarray(ix) = evtarray(ix)+1
      enddo
    elseif (trim(SAS_SENDMODE) == 'MIP CORREC.1') then
      do i = 0,ntrow-1
        iy = d_rawYCol(i)+1
        if (iy .ge. 1 .and. iy .le. ny_ccd .and. phaCol(i) .ge. evt_phaMin) &
           evtarray(iy) = evtarray(iy)+1
      enddo
    else
    endif

! -------------------------- 
! Read table columns, 
! Start loop over all frames
! --------------------------

    iaux = 0
    call message(AppMsg,Noisy,'... starting loop over input frames ...')

    oldframe = -1
    goodtime = .true.
    do ifrm = 1, nfrm  ! loop input frames

      irow = ind(ifrm)
      frame_nr      = frameCol(irow-1)
      if (frame_nr > aux1frame(ntaux1-1)) then
        frameCounterError = .false.
        frameEventError = .true.
      elseif (frame_nr < oldframe) then
        frameCounterError = .true.
        frameEventError = .false.
      else
        frameCounterError = .false.
        frameEventError = .false.
      endif

    if (.not.frameEventError) then

      do while (iaux < ntaux1 .and. aux1frame(iaux) < frame_nr)
!       if (iaux > ntaux1-2) then
!         print*,'REVOLUTION = ',revolution
!         write(errstr,*) 'iaux,ntaux1,aux1frame(iaux),frame_nr = ', &
!                          iaux,ntaux1,aux1frame(iaux),frame_nr
!         call warning('EvtFrame_1')
!         print*,'EvtFrame_1: ', &
!                         'iaux,ntaux1,aux1frame(iaux),frame_nr = ', &
!                          iaux,ntaux1,aux1frame(iaux),frame_nr
!       endif
        iaux = iaux + 1
      enddo

      if (verbosity(AppMsg) > Noisy) then
        write(errstr,*) '.not.frameEventError for IAUX = ',iaux
        call message(AppMsg,Noisy)
      endif

      if (aux1frame(iaux) > frame_nr) then
        if(aux1ftcoarse(iaux) < 32767) then
          write(errstr,*) 'Event Frame_Nr not found in AUX file: ',frame_nr

! warning due to quite a number of incorrect ODFs
          call warning('misFrame')        
! activate error again after reprocessing of ODFs (which is unlikely to happen)
!         call error('misFrame')
          frameCounterError = .true.
        else
          frameCounterError = .true.
        endif
      endif

    endif

!     call checkAuxTime(goodauxtime)
!     goodtime = .true.

      goodauxtime = (aux1ftcoarse(iaux) < 32767)
      goodtime = (aux1FtQual(iaux) <= 1)
      if (frameCounterError .or. frameEventError) goodtime = .false.

      if (goodauxtime .and. goodtime) then

        oldframe = frame_nr

        if (aux1ftcoarse(iaux) > F1052) then
          write(errstr,*) 'Illegal FTCOARSE/F1052 time value in AUX file: ',& 
                          'IAUX, FTCOARSE = ', iaux,aux1ftcoarse(iaux)
          call warning('wrongFTCoarse')
        endif

        if (ifrm < nfrm) then
          nbuf = ind(ifrm+1) - irow
        else
          nbuf = ntrow - irow + 1
        endif

        allocate(ixx(1:nbuf))
        allocate(iyy(1:nbuf))
        allocate(ampl(1:nbuf))
        allocate(qual(1:nbuf))
        allocate(itime(1:nbuf))
        allocate(distp(1:nbuf))
        allocate(distn(1:nbuf))

        ixx(1:nbuf)   = d_rawXCol(irow-1:nbuf+irow-2)
        iyy(1:nbuf)   = d_rawYCol(irow-1:nbuf+irow-2)
        ampl(1:nbuf)  = phaCol(irow-1:nbuf+irow-2)+ODF_offset

        nwrongPHA = COUNT(MASK = ampl < An_CMLOTH-An_CMCORR)

        if (verbosity(AppMsg) >= Verbose) then
          if (nfrm > 10) then
            if (mod(ifrm,nint(nfrm/10.,Int32)) == 0) then
               write(errstr,'(3x,a,i9,4x,a,i9,3x,f5.1,a,a1)')     &
                    "CCD frame:",ifrm,"event:",irow,              &
                     real(irow)/real(ntrow)*100.0,"% completion"
              call message(AppMsg,Verbose)                 
            endif
          else
            write(errstr,"(3x,a,i9,4x,a,i9,3x,f5.1,a,a1)")       &
                 "CCD frame:",ifrm,"event:",irow,                &
                  real(irow)/real(ntrow)*100.0,"% completion"
            call message(AppMsg,Verbose)
          endif
        endif

! -----------------------
! Search for valid events
! -----------------------

        wrongReadoutOrder = .false.
        wrongCoord = .false.

        call validEvents(irow,nbuf,ixx,iyy,ampl,qual,distp,distn,goodtime, &
                         ngbuf,nmipcol,wrongReadoutOrder,wrongCoord)

        if (wrongCoord) then
          if (wrongCoordN < wrongCoordMax) then
            wrongCoordN = wrongCoordN+1
            write(errstr,*) '  Wrong event coordinates in frame',frame_nr, &
                          'with',nbuf,'events.'
            call message(AppMsg,Sparse)
            do j = 1,nbuf
              write(errstr,'(6x,a,i4,a,i3,a,i4,a,i5,a,i11)') &
                    'I =',j,'  RAWX =',ixx(j)+1,'  RAWY =',iyy(j)+1,&
                    '  PHA =',ampl(j),'  QUAL =',qual(j)
              call message(AppMsg,Sparse)
            enddo
          endif
        endif
        if (wrongReadoutOrder) then
          if (wrongReadoutOrderN < wrongReadoutOrderMax) then
            wrongReadoutOrderN = wrongReadoutOrderN+1
            write(errstr,*) '  Wrong readout order for frame',frame_nr, &
                          'with',nbuf,'events.'
            call message(AppMsg,Sparse)
            do j = 1,nbuf
              write(errstr,'(6x,a,i4,a,i3,a,i4,a,i5,a,i11)') &
                    'I =',j,'  RAWX =',ixx(j)+1,'  RAWY =',iyy(j)+1,&
                    '  PHA =',ampl(j),'  QUAL =',qual(j)
              call message(AppMsg,Sparse)
            enddo
          endif
        endif
!
! GTI handling is now done via AUX1 intervals and not anymore via event
! qualities, this is the only safe way to handle counting mode intervals
! and especially their position in time (accuracy ~ 1s)
!
!x      if ((ngbuf == 0) .and. (empty_frames)) then
!                  --> bad frame
!x        if (.not. first_GTI) then
!x          if (current_GTI) then
! stop current GTI: set eventnumber for later time computation of STOP
!x            current_GTI = .false.
!x          else
! increment eventnumber for later time computation of START
!x          endif
!x        else
!x        endif
!x      elseif ((ngbuf > 0) .and. (empty_frames)) then
!x          if (current_GTI) then
!x          else
! start current GTI: set eventnumber for later time computation of START
!x            current_GTI = .true.         
!x          endif
!x      endif
!       print*,'NGBUF, NBUF, NMIPCOL = ',NGBUF, NBUF, NMIPCOL 
        if (ngbuf > 0) then
          if (ngbuf > nbuf) then          ! should never occur [cannot occur]
!         if (ngbuf == nbuf) then         ! temporarily disabled
!           nothing special to do
          else
            ngbuf = 0
            do i = 1,nbuf
              if (qual(i) <= qualMax) then
                ngbuf = ngbuf+1
                ixx(ngbuf)     = ixx(i)
                iyy(ngbuf)     = iyy(i)
                ampl(ngbuf)    = ampl(i)
                qual(ngbuf)    = qual(i)
                distp(ngbuf)   = distp(i)
                distn(ngbuf)   = distn(i)
              endif
            enddo
          endif
          iend = igrow+ngbuf-1                   ! iend for this frame

          if (verbosity(AppMsg) > Noisy) then
            write(errstr,*) 'NGBUF,ISTART,IEND,IFRM = ',ngbuf,igrow,iend,ifrm
            call message(AppMsg,Noisy)                 
          endif

          arrFrame(igrow:iend)       =  frame_nr
! IF (FRAME_NR < 20) PRINT*,'KFRM,FRAME_NR = ',kfrm, frame_nr
          arrRawX(igrow:iend)        =  ixx(1:ngbuf)+1
          arrRawY(igrow:iend)        =  iyy(1:ngbuf)+1
          arrPha(igrow:iend)         =  ampl(1:ngbuf)
          arrQual(igrow:iend)        =  qual(1:ngbuf)

          if (MIP_dist) then
            arrDistP(igrow:iend)     =  distp(1:ngbuf)
            arrDistN(igrow:iend)     =  distn(1:ngbuf)
          endif

          if (verbosity(AppMsg) > Noisy) then
            i = min(igrow+30,iend)
            write(errstr,*) 'outqualflagCol() = ',arrQual(igrow:i)
            call message(AppMsg,Noisy)                 
            write(errstr,*) 'igrow:iend = ',igrow,iend
            call message(AppMsg,Noisy)                 
          endif

!
! increment number of frames with NGBUF > 0, kfrm = 1... from now on
!
          kfrm = kfrm + 1
          ind(kfrm) = igrow+1
          igrow = iend+1                   ! istart for next frame
!
! store frame number of this frame for later relation to EXPOSURE frames
!
          arrFrames(kfrm-1) = frame_nr     ! frame number of k-th frame [k=0..]
          if (lfrm < ntaux2-1) then
            do while (aux2Cycle(lfrm) < frame_nr)
!             print*,kfrm,lfrm,aux2Cycle(lfrm),frame_nr
              if (lfrm == ntaux2-1) exit
              lfrm = lfrm + 1
            enddo
            ndsclinq_20 = aux2Ndisclin(lfrm)
          else
            ndsclinq_20 = ndsclinq_ave 
          endif

!
! fill in the bit mask NDSCLIN# (#=1..4) for columns containing MIPs:
!
          do j = 1, 16
            if (.not.lxcolStore(j)) then
              arrNdsclin1(kfrm-1) = ibset(arrNdsclin1(kfrm-1),j-1)
            endif
            if (.not.lxcolStore(j + 16)) then
              arrNdsclin2(kfrm-1) = ibset(arrNdsclin2(kfrm-1),j-1)
            endif
            if (.not.lxcolStore(j + 32)) then
              arrNdsclin3(kfrm-1) = ibset(arrNdsclin3(kfrm-1),j-1)
            endif
            if (.not.lxcolStore(j + 48)) then
              arrNdsclin4(kfrm-1) = ibset(arrNdsclin4(kfrm-1),j-1)
            endif
          enddo
!
! subroutine ValidEvents() has only "MIP CORREC.2" enabled (column-wise)
!
          do j = 1,EPN_MAX_X
            if (.not.lxcolStore(j)) dlisasarray(j) = dlisasarray(j)+1
          enddo
!
! FrameLim(0:..) contains number (1..) of last event in frame
! this relates extensions EXPOSUnn to EVENTS
!
          if (kfrm > mfrm) then
            call message(AppMsg,Sparse, &
              'Task does not expect any entries anymore in EXPOSURE extension.')
          else
            arrFramelim(kfrm-1) = iend+1
!
! FracExp(0:..) contains fractional exposure induced by SAS mip rejection
!
            arrFracExp(kfrm-1) = arrFracExp(kfrm-1)*(1.d0-dble(nmipcol)/64.0d0)
!
! FracExp(0:..) now with fractional exposure induced by onboard mip rejection
!
            if (enable_AUX2dtime) then
!             ndsclinq_20 = ndsclinq_ave
              arrFracExp(kfrm-1) = arrFracExp(kfrm-1)* &
                                 (1.0d0-dble(ndsclinq_20)/maxfrcCOLperQUAD)
            endif
          endif
        else
! bad frame on return from "validEvents"
        endif           ! (ngbuf > 0) ?

        nwrong = kwrong
!       if (nwrong > maxwrong) then
!         call message(AppMsg,sparse, &
!                     'Wrong:     coord | order |  time |   MIP | frame')
!         write(errstr,'(6i8)') nwrong,nwrongCoord,nwrongOrder,nwrongTime, &
!                               nMIP,nfrmbad
!         call message(AppMsg,sparse)
!         call error('maxwrong','More wrong events than maxwrong allows')
!       endif

        pwrong = 100*real(nwrong)/real(ntrow)

        if (pwrong > real(wronglimit) .and. nwrong > 10 .and. warnFlag==0) then
          write(errstr,*) 'wrongEvents: Reached wrongpixlimit - ', &
                          'no more detailed messages about wrong events'
          call message(AppMsg,Verbose)
          warnFlag = 1
        endif

        deallocate(ixx, iyy, ampl, qual, itime, distp, distn)

      elseif (.not.frameCounterError) then
        nwrongFTCoarse = nwrongFTCoarse+1
        if (verbosity(AppMsg) > Verbose) then
          write(errstr,*) 'Illegal AUX time:  IAUX, FRAME_NR, COARSE, FINE = ',&
                   iaux, frame_nr,aux1ftcoarse(iaux),aux1ftfine(iaux)
          call message(AppMsg,Verbose)
        endif
      else
        frameCounterError = .false.
        if (verbosity(AppMsg) > Verbose) then
          write(errstr,*) 'Illegal FrameCounter: IAUX, FRAME, COARSE, FINE = ',&
                   iaux, aux1frame(iaux),aux1ftcoarse(iaux),aux1ftfine(iaux)
          call message(AppMsg,Verbose)
        endif
!       if (iaux > ntaux1-2) then
!         print*,'REVOLUTION = ',revolution
!         write(errstr,*) 'iaux,ntaux1,aux1frame(iaux),frame_nr = ', &
!                          iaux,ntaux1,aux1frame(iaux),frame_nr
!         call warning('EvtFrame_2')
!         print*,'EvtFrame_2: ', &
!                         'iaux,ntaux1,aux1frame(iaux),frame_nr = ', &
!                          iaux,ntaux1,aux1frame(iaux),frame_nr
!       endif
        iaux = iaux+1
!       if (iaux > ntaux1) then
!         if (revolution < 26) exit
!       endif
      endif
    enddo

    call message(AppMsg,Noisy,"... leaving loop over frames ...")

!
! iend = row of last (good) event [rows = 0..iend]
! ngrow = number of (good) events [good = 1..ngrow]
!

    ngrow = iend+1
    deallocate(d_rawXCol)
    deallocate(d_rawYCol)

! ------------------------
! End loop over all frames
! ------------------------

  else
!   ngrow = 0
  endif

  write(errstr,'(1x,a,3i9)') 'Check of final indices: NTROW, IEND, NGROW = ', &
             ntrow,iend,ngrow
  call message(AppMsg,Noisy)

! write(errstr,*) 'EXIST_DLI =',exist_dli,'  An_SENDMODE = ',trim(An_SENDMODE)
! call message(AppMsg,Sparse)

  if (.not.exist_dli .AND. trim(An_SENDMODE).ne."NO MIP CORR.") then
    write(errstr,*) trim(sciencefilename),': No DLI file for CCD',CCDno
!   call warning('noDLIfile')
    call message(AppMsg,Sparse-1)
  endif

  if (exist_dli) then
    call SAS_DLIanalysis(dliarray,evtarray,An_SENDMODE, An_EAMIPSEL, &
                             dli_mean,dli_median,dli_var,dli_sigma)
  elseif (trim(An_SENDMODE).ne.'NO MIP CORR.') then
    call message(AppMsg,Sparse,' No DLI file: performing analysis on SAS detected MIPs')
    call SAS_DLIanalysis(dlisasarray,evtarray,SAS_SENDMODE, An_EAMIPSEL, &
                             dli_mean,dli_median,dli_var,dli_sigma)
  elseif (trim(An_SENDMODE).eq.'NO MIP CORR.') then
    call message(AppMsg,Sparse,' No MIP CORR: performing analysis on SAS detected MIPs')
    call SAS_DLIanalysis(dlisasarray,evtarray,SAS_SENDMODE, An_EAMIPSEL, &
                             dli_mean,dli_median,dli_var,dli_sigma)
  else
  endif

! --------------------
! clear up some memory
! --------------------

  if (ccxalloc) deallocate(ccx1Stcoarse,ccx1Stfine,ccx1Etcoarse,ccx1Etfine, &
                           ccx1Duration,ccx1NCycle)
  if (aux2alloc) deallocate(aux2cycle,aux2ndisclin,aux2gti_b,aux2gti_e)
  if (aux1alloc) deallocate(aux1Frame,aux1Ftcoarse,aux1Ftfine,aux1FtTime, &
                            aux1FtQual)
  if (empty_frames .and. allocated(ind)) deallocate(ind)

  if (allocated(aux1Gtime))    deallocate(aux1Gtime)
  if (allocated(aux1Bindex_s)) deallocate(aux1Bindex_s)
  if (allocated(aux1Bindex_e)) deallocate(aux1Bindex_e)
  if (allocated(aux1GFrame))   deallocate(aux1GFrame)


! -----------------------------------------------------------
! check allocated arrays to save memory (future developments)
! -----------------------------------------------------------

  if (verbosity(AppMsg) > Verbose) then
    call message(AppMsg,Verbose,'Checking which ALLOCATABLE arrays are still allocated...')
    if (ALLOCATED(deb)) call message(AppMsg,Verbose,' ALLOCATED: deb')
    if (ALLOCATED(fin)) call message(AppMsg,Verbose,' ALLOCATED: fin')
    if (ALLOCATED(arrExpTime)) call message(AppMsg,Verbose,' ALLOCATED: arrExpTime')
    if (ALLOCATED(aux1GTime)) call message(AppMsg,Verbose,' ALLOCATED: aux1GTime')
    if (ALLOCATED(aux2GTime)) call message(AppMsg,Verbose,' ALLOCATED: aux2GTime')
    if (ALLOCATED(aux1gti_b)) call message(AppMsg,Verbose,' ALLOCATED: aux1gti_b')
    if (ALLOCATED(aux1gti_e)) call message(AppMsg,Verbose,' ALLOCATED: aux1gti_e')
    if (ALLOCATED(ccx1Duration)) call message(AppMsg,Verbose,' ALLOCATED: ccx1Duration')
    if (ALLOCATED(arrFracExp)) call message(AppMsg,Verbose,' ALLOCATED: arrFracExp')
    if (ALLOCATED(dummy64)) call message(AppMsg,Verbose,' ALLOCATED: dummy64')
    if (ALLOCATED(aux1Frame)) call message(AppMsg,Verbose,' ALLOCATED: aux1Frame')
    if (ALLOCATED(aux2Cycle)) call message(AppMsg,Verbose,' ALLOCATED: aux2Cycle')
    if (ALLOCATED(aux1FtTime)) call message(AppMsg,Verbose,' ALLOCATED: aux1FtTime')
    if (ALLOCATED(aux1GFrame)) call message(AppMsg,Verbose,' ALLOCATED: aux1GFrame')
    if (ALLOCATED(aux2gti_b)) call message(AppMsg,Verbose,' ALLOCATED: aux2gti_b')
    if (ALLOCATED(aux2gti_e)) call message(AppMsg,Verbose,' ALLOCATED: aux2gti_e')
    if (ALLOCATED(aux1Bindex_s)) call message(AppMsg,Verbose,' ALLOCATED: aux1Bindex_s')
    if (ALLOCATED(aux1Bindex_e)) call message(AppMsg,Verbose,' ALLOCATED: aux1Bindex_e')
    if (ALLOCATED(arrExpFrame)) call message(AppMsg,Verbose,' ALLOCATED: arrExpFrame')
    if (ALLOCATED(arrFrame)) call message(AppMsg,Verbose,' ALLOCATED: arrFrame')
    if (ALLOCATED(arrFrames)) call message(AppMsg,Verbose,' ALLOCATED: arrFrames')
    if (ALLOCATED(arrFramelim)) call message(AppMsg,Verbose,' ALLOCATED: arrFramelim')
    if (ALLOCATED(arrQual)) call message(AppMsg,Verbose,' ALLOCATED: arrQual')
    if (ALLOCATED(qual)) call message(AppMsg,Verbose,' ALLOCATED: qual')
    if (ALLOCATED(ind)) call message(AppMsg,Verbose,' ALLOCATED: ind')
    if (ALLOCATED(iwrong)) call message(AppMsg,Verbose,' ALLOCATED: iwrong')
    if (ALLOCATED(arrRawX)) call message(AppMsg,Verbose,' ALLOCATED: arrRawX')
    if (ALLOCATED(arrRawY)) call message(AppMsg,Verbose,' ALLOCATED: arrRawY')
    if (ALLOCATED(arrPha)) call message(AppMsg,Verbose,' ALLOCATED: arrPha')
    if (ALLOCATED(arrNdsclin1)) call message(AppMsg,Verbose,' ALLOCATED: arrNdsclin1')
    if (ALLOCATED(arrNdsclin2)) call message(AppMsg,Verbose,' ALLOCATED: arrNdsclin2')
    if (ALLOCATED(arrNdsclin3)) call message(AppMsg,Verbose,' ALLOCATED: arrNdsclin3')
    if (ALLOCATED(arrNdsclin4)) call message(AppMsg,Verbose,' ALLOCATED: arrNdsclin4')
    if (ALLOCATED(aux1Ftcoarse)) call message(AppMsg,Verbose,' ALLOCATED: aux1Ftcoarse')
    if (ALLOCATED(aux1Ftfine)) call message(AppMsg,Verbose,' ALLOCATED: aux1Ftfine')
    if (ALLOCATED(aux2Ndisclin)) call message(AppMsg,Verbose,' ALLOCATED: aux2Ndisclin')
    if (ALLOCATED(ixx)) call message(AppMsg,Verbose,' ALLOCATED: ixx')
    if (ALLOCATED(iyy)) call message(AppMsg,Verbose,' ALLOCATED: iyy')
    if (ALLOCATED(d_rawXCol)) call message(AppMsg,Verbose,' ALLOCATED: d_rawXCol')
    if (ALLOCATED(d_rawYCol)) call message(AppMsg,Verbose,' ALLOCATED: d_rawYCol')
    if (ALLOCATED(ampl)) call message(AppMsg,Verbose,' ALLOCATED: ampl')
    if (ALLOCATED(ccx1Stcoarse)) call message(AppMsg,Verbose,' ALLOCATED: ccx1Stcoarse')
    if (ALLOCATED(ccx1Etcoarse)) call message(AppMsg,Verbose,' ALLOCATED: ccx1Etcoarse')
    if (ALLOCATED(ccx1Stfine)) call message(AppMsg,Verbose,' ALLOCATED: ccx1Stfine')
    if (ALLOCATED(ccx1Etfine)) call message(AppMsg,Verbose,' ALLOCATED: ccx1Etfine')
    if (ALLOCATED(ccx1NCycle)) call message(AppMsg,Verbose,' ALLOCATED: ccx1NCycle')
    if (ALLOCATED(aux1FtQual)) call message(AppMsg,Verbose,' ALLOCATED: aux1FtQual')
    if (ALLOCATED(aux1ccdNr)) call message(AppMsg,Verbose,' ALLOCATED: aux1ccdNr')
    if (ALLOCATED(arrDistP)) call message(AppMsg,Verbose,' ALLOCATED: arrDistP')
    if (ALLOCATED(arrDistN)) call message(AppMsg,Verbose,' ALLOCATED: arrDistN')
    if (ALLOCATED(distp)) call message(AppMsg,Verbose,' ALLOCATED: distp')
    if (ALLOCATED(distn)) call message(AppMsg,Verbose,' ALLOCATED: distn')
    if (ALLOCATED(calRawX)) call message(AppMsg,Verbose,' ALLOCATED: calRawX')
    if (ALLOCATED(itime)) call message(AppMsg,Verbose,' ALLOCATED: itime')
    if (ALLOCATED(eIn)) call message(AppMsg,Verbose,' ALLOCATED: eIn')
  endif

! -------------------------
! Correct for offset shifts
! -------------------------

  if (ngrow > 0) then
    if (correct_CCF) then
!
! determine which columns have an offset in the CCF: 
! dummy input to CAL routine and compare input/output amplitudes
!
      allocate(eIn(0:nx_ccd-1))
      allocate(dummy64(0:nx_ccd-1))
      allocate(calRawX(0:nx_ccd-1))
      do i = 0,nx_ccd-1
        eIn(i)     = 0
        dummy64(i) = pha_offset_time
        calRawX(i) = i
      enddo
      call CAL_integerToReal(eIn,eInReal)
      call CAL_offsetCorrect(dummy64,eInReal,eOutReal,calrawX)
      call CAL_realToInteger(eOutReal,eOut)
      do i = 0,nx_ccd-1
        ocv(i) = eOut(i) - eIn(i)
        on_offcol(i) = (ocv(i) > offcollimit)
      enddo
      i = 0
      next_offcol(i) = ( on_offcol(i+1) )
      i = nx_ccd-1
      next_offcol(i) = ( on_offcol(i-1) )
      do i = 1,nx_ccd-2
        next_offcol(i) = ( on_offcol(i-1) .or. on_offcol(i+1) )
      enddo
      call CAL_releaseMemory(eOutReal)
      call CAL_releaseMemory(eInReal)
      call CAL_releaseMemory(eOut)
      deallocate(eIn,dummy64,calRawX)
!
! now do the CAL offset correction with real events
!
      allocate(eIn(0:ngrow-1))
      eIn(0:ngrow-1) = arrPha(0:ngrow-1)

      call message(AppMsg,Noisy,"... entering CAL_integerToReal ...")
      call CAL_integerToReal(eIn,eInReal)

      if (ngrow > maxManyEvents) then
        tooManyEvents = .true.
        dims(1)  = 1
        outccd   = addTable(outfile,"EVENTS",ngrow)
        time     = addColumn(outccd,"TIME",Real64,"s",dims, &
                       comment="Event Time")
        outtimeCol => real64Data(time)
        outtimeCol(0:ngrow-1) = pha_offset_time
      else
        tooManyEvents = .false.
        allocate(dummy64(0:ngrow-1))
        dummy64 = pha_offset_time
      endif

      allocate(calRawX(0:ngrow-1))
      calRawX(0:ngrow-1) = arrRawX(0:ngrow-1)-1

      call message(AppMsg,Noisy,"... entering CAL_offsetCorrect ...")
      if (tooManyEvents) then
        call CAL_offsetCorrect(outtimeCol,eInReal,eOutReal,calRawX)
      else
        call CAL_offsetCorrect(dummy64,eInReal,eOutReal,calrawX)
        if (ALLOCATED(dummy64)) deallocate(dummy64)
      endif

      call CAL_releaseMemory(eInReal)
      if (ALLOCATED(calRawX)) deallocate(calRawX)
      call message(AppMsg,Noisy,"... leaving CAL_offsetCorrect ...")

      call message(AppMsg,Noisy,"... entering CAL_realToInteger ...")
      call CAL_realToInteger(eOutReal,eOut)
      call message(AppMsg,Noisy,"... leaving CAL_realToInteger ...")
      call CAL_releaseMemory(eOutReal)

      arrPha(0:ngrow-1) = eOut(0:ngrow-1)

      if (ALLOCATED(eIn)) deallocate(eIn)
    else
      call message(AppMsg,Noisy,"... entering SAS_offsetCorrect ...")
      call SAS_offsetCorrect(Quadrant,CCDid)
      call message(AppMsg,Noisy,"... leaving SAS_offsetCorrect ...")
    endif

! -----------------------------
! set relevant event attributes
! -----------------------------

    if (verbosity(AppMsg) >= Noisy) then
      write(errstr,'(2x,a,64l1)') 'On_Offset_Column   ',(on_offcol(i),i=0,nx_ccd-1)
      call message(AppMsg,Verbose)
      write(errstr,'(2x,a,64l1)') 'Next_Offset_Column ',(next_offcol(i),i=0,nx_ccd-1)
      call message(AppMsg,Verbose)
    endif

    ncolOffset = 0
    do i = 0,nx_ccd-1
      if (on_offcol(i)) ncolOffset =  ncolOffset+1
    enddo

    do i = 0,ngrow-1
      if (on_offcol(arrRawX(i)-1)) &
          arrQual(i) = ior(arrQual(i),EVATT_ON_OFFSET_COLUMN)
      if (next_offcol(arrRawX(i)-1)) &
          arrQual(i) = ior(arrQual(i),EVATT_NEXT_TO_OFFSET_COLUMN)
    enddo

  endif


! --------------------------
! create output file
! --------------------------

  call createRawOutput(inmode,insubmode,revolution)

  if (ngrow > 0) then

    outFrameCol = arrFrame(0:iend)
!   call release(outframecol)
    deallocate(arrFrame)

    outPhaCol = arrPha(0:iend)
    call release(pha)
    deallocate(arrPha)

    outQualFlagCol = arrQual(0:iend)
!   call release(outQualFlagCol)
    deallocate(arrQual)

    outrawXCol = arrRawX(0:iend)
!   call release(outrawXCol)
    outrawYCol = arrRawY(0:iend)
!   call release(outrawYCol)
    outccdnrCol = CCDno
    call release(ccdnr)

    if (MIP_dist) then
      outmipdistposcol = arrDistP(0:iend)
      outmipdistnegcol = arrDistN(0:iend)
      deallocate(arrDistP,arrDistN)
    endif

  else

    if (ALLOCATED(arrFrame)) deallocate(arrFrame)
    if (ALLOCATED(arrPha))   deallocate(arrPha)
    if (ALLOCATED(arrQual))  deallocate(arrQual)

  endif

  call release(infile)


! -------------------------------------------------------------------
! Add attributes TIMESYS, MJDREF, TIMEUNIT, TSTART, TSTOP and TELAPSE
! -------------------------------------------------------------------

  call writeTIMEattributes(outfile,outccd,1)

  call setAttribute(outccd,"TRESET", &
           0.0,"s", "Reset time offset added to PN times")
  call setAttribute(outccd, "TSTART", &
           0.0, comment="Start time of first frame")
  call setAttribute(outccd, "TSTOP", &
           0.0, comment="End time of last frame")
  call setAttribute(outccd, "TELAPSE", &
           0.0,"s", comment="Time elapsed between TSTART and TSTOP")

  call setAttribute(outccd, "MIP_THRS", MIP_Threshold," ",&
                            "maximum PHA for non-MIPs")
  call setAttribute(outccd, "MIP_METH", MIP_Method, "method to handle MIPs")
  call setAttribute(outccd, "QUAL_MAX", qualMax," ","maximum allowed quality")
  call setAttribute(outccd, "CCDSMODE", insubmode," ","CCD sub-mode")


! -------------------------------------------------------------------
! Add attributes related to HK files to be evaluated in later tasks
! -------------------------------------------------------------------

    call writeHKattributes


! ----------------------------
! Collect FLAG informations...
! ----------------------------

    if (nwrong > 0) then
     if (verbosity(AppMsg) >= Noisy) then
      call message(AppMsg,Noisy,dashline)
      call message(AppMsg,Noisy,'Infos about wrong events:')
      call message(AppMsg,Noisy,dashline)
      if (nwrong > 1) then
        write(errstr,*) 'the following ',nwrong,' events are wrong:'
        call message(AppMsg,Noisy)
        if (nwrong > maxiwrong) then
          write(errstr,*) ' [only ',maxiwrong,' of them are shown below]'
          call message(AppMsg,Noisy)
        endif
      else
        call message(AppMsg,Noisy,'the following event is wrong:')
      endif
      mwrong = min(nwrong,maxiwrong)
      j = INT(mwrong/10,Int32)
      if (mwrong < 10) then
        write(errstr,*) iwrong(1:mwrong)
        call message(AppMsg,Noisy)
      else
        do i = 0, j - 1
          write(errstr,*) iwrong(i * 10 + 1:(i + 1) * 10)
          call message(AppMsg,Noisy)
        enddo
        write(errstr,*) iwrong(j * 10 + 1:mwrong)
        call message(AppMsg,Noisy)
      endif
      write(errstr,"(a,f5.1,a)") " corresponding to ",pwrong,"% of all events."
      call message(AppMsg,Noisy)
     elseif (verbosity(AppMsg) == Verbose) then
       call message(AppMsg,Verbose,dashline)
       write(errstr,*) "total number of wrong events :   ", nwrong
       call message(AppMsg,Verbose)
     endif

     if (verbosity(AppMsg) >= Verbose) then
      if (nMIP > 0) then
       write(errstr,*) "number of detected MIP events:   ", nMIP
       call message(AppMsg,Verbose)
      endif
      if (NDSCLINC > 0) then
       write(errstr,*) "number of rejected MIP columns:  ", NDSCLINC
       call message(AppMsg,Verbose)
      endif
      if (nwrongPHA > 0) then
       write(errstr,*) "Events with negative PHA:        ", nwrongPHA
       call message(AppMsg,Verbose)
      endif
      if (nwrongCoord > 0) then 
       write(errstr,*) "events with wrong coordinates:   ", nwrongCoord
       call message(AppMsg,Verbose)
      endif
      if (nwrongOrder > 0) then
       write(errstr,*) "events in wrong readout order:   ", nwrongOrder
       call message(AppMsg,Verbose)
      endif
      if (nwrongTime > 0) then
       write(errstr,*) "events in wrong timing order:    ", nwrongTime
       call message(AppMsg,Verbose)
      endif
      if (nwrongFTCoarse > 0) then
       write(errstr,*) "Frames with false coarse time:   ", nwrongFTCoarse
       call message(AppMsg,Verbose)
      endif
      if (nfrmbad > 0) then
       write(errstr,*) "Number of bad frames:            ", nfrmbad 
       call message(AppMsg,Verbose)
      endif
      write(errstr,*)  "Total number of input frames:    ", nfrm
      call message(AppMsg,Verbose)
      write(errstr,*)  "Total number of output frames:   ", kfrm
      call message(AppMsg,Verbose)
      write(errstr,*)  "Total number of input events:    ", ntrow
      call message(AppMsg,Verbose)
      write(errstr,*)  "Total number of output events:   ", ngrow
      call message(AppMsg,Verbose)
      call message(AppMsg,Verbose,dashline)
     endif
    else
      write(errstr,*)  "Total number of frames:        ", nfrm
      if (ntrow > 0) then
        call message(AppMsg,Verbose)
        call message(AppMsg,Verbose,"There were no wrong events!")
      endif
    endif

    if (ALLOCATED(iwrong)) deallocate(iwrong)

! -----------------------------------------------------------
! Add wrong event statistics to the history and as attributes
! -----------------------------------------------------------

    call setAttribute(outccd,"MIPCOLN",nMIP,"",&
                                 "Number of MIP events")
    call setAttribute(outccd,"WRGAMPLI", nwrongPHA,      "",&
                                 "Number of events with negative PHA")
    call setAttribute(outccd,"WRGCOORD", nwrongCoord,    "",&
                                 "Number of events with wrong coordinates")
    call setAttribute(outccd,"WRGORDER", nwrongOrder,    "",&
                                 "Number of frames in wrong readout order")
    call setAttribute(outccd,"WRGTIME" , nwrongTime,     "",&
                                 "Number of frames in wrong time order")
    call setAttribute(outccd,"WRGCOARS", nwrongFTCoarse, "",&
                                 "Number of frames with FFFF coarse time")
    call setAttribute(outccd,"FRAMEBAD", nfrmbad,        "",&
                                 "Number of bad frames")
    call setAttribute(outccd,"FRAMETOT", nfrm,           "",&
                                 "Total number of input frames")
    call setAttribute(outccd,"EMPTYFRM", trim(enable_of_empty_frames),&
                                 "Handling of empty frames")

    write(history,"(a,f5.1,a)") "epframes: ",pwrong,"% of the events are wrong"
    call addHistory(outccd,history)



  if (MIP_dist) then
    do i = 1,nx_ccd
      outdistCol(i-1)  = i
      outhistLCol(i-1) = dist_2_mip_hist_l(i)
      outhistRCol(i-1) = dist_2_mip_hist_r(i)
    enddo
    call setAttributes(attributable(outhist),attributable(outccd))
  endif


  call setAttribute(outccd,"WINDOWX0",nx0,"", &
                    "X coordinate of bottom left corner of window")
  call setAttribute(outccd,"WINDOWY0",ny0,"", &
                    "Y coordinate of bottom left corner of window")
  call setAttribute(outccd,"WINDOWDX",nxdim,"", &
                    "Size, along x-axis, of window")
  call setAttribute(outccd,"WINDOWDY",nydim,"", &
                    "Size, along y-axis, of window")

  call setAttribute(rawX, "TLMIN", nx0,   comment="Minimum for RAWX")
  call setAttribute(rawY, "TLMIN", ny0,   comment="Minimum for RAWY")

  call setAttribute(rawX, "TLMAX", nxmax, comment="Maximum for RAWX")
  call setAttribute(rawY, "TLMAX", nymax, comment="Maximum for RAWY")


! ------------------------------------------------------------------
! Convert frame counter to time tag for event frame numbers
! ------------------------------------------------------------------

!-UL   call OAL_setState(filename)
!-UL   call OAL_selectFile(CCDno, fileName)

    if (ngrow > 0) then
      call release(frame)
      call frameCountersToTimes(frames,timeTags) 

      outtimeCol(0:ngrow-1)=timeTags(0:ngrow-1)
      telapse = outtimeCol(ngrow-1) - outtimeCol(0) + (2.0 * halfdt)

      call setAttribute(outccd, "TSTART", &
             outtimeCol(0) - deltat,comment="Start time of first frame")
!            outtimeCol(0) - halfdt,comment="Start time of first frame")
      call setAttribute(outccd, "TSTOP", &
             outtimeCol(ngrow-1)         , comment="End time of last frame")
!            outtimeCol(ngrow-1) + halfdt, comment="End time of last frame")
      call setAttribute(outccd, "TELAPSE", &
             telapse,"s", &
             comment="Time elapsed between TSTART and TSTOP")     
!     call setAttribute(outccd, "TRESET", TimeOffset,"s", &
!            comment="Reset time offset added to PN times")
!     call message(AppMsg,Verbose,"... written TimeOffset ...")

      delta_duration = telapse-duration
      if (delta_duration > 0) then
        if (delta_duration > 10) then
          write(errstr,*) 'TELAPSE [',telapse,'] > DURATION [',duration, &
           ']. Event times may be incorrect or exposure lasted longer than scheduled.'
          call warning("tooLongExposure")
        else
          write(errstr,*) 'TELAPSE [',telapse,'] > DURATION [',duration, ']'
          call message(AppMsg,Sparse)        
        endif
      endif
    endif

!
! couple of code checks during task development (and Solaris/OSF consistency)
!

  if (verbosity(AppMsg) > Noisy) then
    write(errstr,'(3x,a,3i11)') &
         'Index check: NTROW, NGROW, NFRM    = ',ntrow,ngrow,nfrm
    call message(AppMsg,Noisy)
    write(errstr,'(3x,a,3i11)') &
         'Index check: IGROW, NGROW, KFRM    = ',igrow,ngrow,kfrm
    call message(AppMsg,Noisy)                 
  endif

  

! ---------------
! Create GTI file
! ---------------

  if (empty_frames) then
    call determineGTIfromAUX(ngti)
  else
    call determineGTIfromEvents(ngti)
  endif
  call createGTI(ngti)


  if (ALLOCATED(arrRawX))      deallocate(arrRawX)
  if (ALLOCATED(arrRawY))      deallocate(arrRawY)
  if (allocated(aux1gti_b))    deallocate(aux1gti_b)
  if (allocated(aux1gti_e))    deallocate(aux1gti_e)



! --------------------------------
! Release EVENTS extension columns
! --------------------------------

  call message(AppMsg,Verbose,"... releasing events columns ...")

  call release(time)
  call release(frame)
  call release(rawX)
  call release(rawY)
! call release(pha)
! call release(ccdnr)
  call release(qualflag)


! -----------------------
! Fill EXPOSURE extension
! -----------------------

  call message(AppMsg,Verbose,"... preparing exposure extension ...")

  if (verbosity(AppMsg) > Noisy) then
    write(errstr,'(3x,a,3i11)')     &
             'Index check: NTROW, NGROW, KFRM    = ', NTROW, NGROW, KFRM
    call message(AppMsg,Noisy)
  endif

!!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!
!!  if (kfrm > 0) then                  !!! for 0 events KFRM=1 on solaris !!
!
!!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

  exptimeCol => Real64Data(exptime)
  exptimeCol = 0.

  if (empty_frames) then
!   exptimeCol(0:mfrm-1) = arrExpTime(0:mfrm-1)
    exptimeCol(0:ntaux0-1) = arrExpTime(0:ntaux0-1)
  else
    if ((ngrow > 0) .and. (kfrm > 0)) then
      do i = 1, kfrm
        if ( ind(i)-1 > ngrow-1) then
          if (verbosity(AppMsg) >=Verbose) then
            write(errstr,'(3x,a,3i11)')     &
               'Index check: > i,ind(i)-1,kfrm = ', i,ind(i)-1,kfrm
            call message(AppMsg,Verbose)
          endif
          exptimeCol(i-1) = -1
        else
          exptimeCol(i-1) = timeTags(ind(i)-1)
        endif
      enddo
      call OAL_releaseMemory(timeTags)
    endif
  endif

  if (ALLOCATED(ind)) deallocate(ind)
  if (ALLOCATED(arrExpTime))  deallocate(arrExpTime)

  call release(exptime)

  if (also_timedel_col) then
    timedelCol => Real32Data(timedel)
    timedelCol =  FRMTIME_SECONDS * fc0
    call release(timedel)
  endif
!
  if (insubmode == FAST_TIMING .or. insubmode == FAST_BURST .or. &
           .NOT. enable_PSFtime) then 
    call setAttribute(outccd, "TIMEDEL", FRMTIME_SECONDS * fc0, "s", &
               "Length of exposure entry interval")
  else
    call setAttribute(outccd, "TIMEDEL", PSFTIME*1.D-3, "s", &
               "Length of exposure entry interval")   
  endif

  if (empty_frames) then

    fracexpCol  => Real32Data(fracexp)
    ndsclin1Col => Int16Data(ndsclin1)
    ndsclin2Col => Int16Data(ndsclin2)
    ndsclin3Col => Int16Data(ndsclin3)
    ndsclin4Col => Int16Data(ndsclin4)
    framelimCol => Int32Data(framelim)

    fracexpCol(0:ntaux0-1) = 0.0
    ndsclin1Col(0:ntaux0-1) = 0
    ndsclin2Col(0:ntaux0-1) = 0
    ndsclin3Col(0:ntaux0-1) = 0
    ndsclin4Col(0:ntaux0-1) = 0
    framelimCol(0:ntaux0-1) = 0

    oldValue = 0.0d0
    oldLim = 0    ! initialization in `epevents' !!!
    k = 0
    do i = 0,ntaux0-1
      if (arrExpFrame(i) > -1) then
        if (k > nfrm-1) exit
        if (arrExpFrame(i) < arrFrames(k)) then
          fracExpCol(i)  = oldValue
          ndsclin1Col(i) = 0
          ndsclin2Col(i) = 0
          ndsclin3Col(i) = 0
          ndsclin4Col(i) = 0
          frameLimCol(i) = oldLim
        else
          fracExpCol(i)  = arrFracExp(k)
          ndsclin1Col(i) = arrNdsclin1(k)
          ndsclin2Col(i) = arrNdsclin2(k)
          ndsclin3Col(i) = arrNdsclin3(k)
          ndsclin4Col(i) = arrNdsclin4(k)
          frameLimCol(i) = arrframelim(k)
          oldValue       = arrFracExp(k) 
          oldLim         = arrframelim(k)    
          k = k+1
        endif
      else
        fracExpCol(i)  = oldValue
        ndsclin1Col(i) = 0
        ndsclin2Col(i) = 0
        ndsclin3Col(i) = 0
        ndsclin4Col(i) = 0
        frameLimCol(i) = oldLim
      endif
    enddo

    call release(fracexp)
    if (allocated(arrFracExp)) deallocate(arrFracExp)
    call release(ndsclin1)
    if (allocated(arrNdsclin1)) deallocate(arrNdsclin1)
    call release(ndsclin2)
    if (allocated(arrNdsclin2)) deallocate(arrNdsclin2)
    call release(ndsclin3)
    if (allocated(arrNdsclin3)) deallocate(arrNdsclin3)
    call release(ndsclin4)
    if (allocated(arrNdsclin4)) deallocate(arrNdsclin4)
    call release(framelim)
    if (allocated(arrFramelim)) deallocate(arrFramelim)

    if (ALLOCATED(arrExpFrame)) deallocate(arrExpFrame)
    if (ALLOCATED(arrFrames))   deallocate(arrFrames)

  else

    if (ALLOCATED(arrExpFrame)) deallocate(arrExpFrame)
    if (ALLOCATED(arrFrames))   deallocate(arrFrames)

    fracexpCol => Real32Data(fracexp)
    fracexpCol = arrFracExp(0:kfrm-1)
    call release(fracexp)
    if (allocated(arrFracExp)) deallocate(arrFracExp)

    ndsclin1Col => Int16Data(ndsclin1)
    ndsclin1Col = arrNdsclin1(0:kfrm-1)
    call release(ndsclin1)
    if (allocated(arrNdsclin1)) deallocate(arrNdsclin1)

    ndsclin2Col => Int16Data(ndsclin2)
    ndsclin2Col = arrNdsclin2(0:kfrm-1)
    call release(ndsclin2)
    if (allocated(arrNdsclin2)) deallocate(arrNdsclin2)

    ndsclin3Col => Int16Data(ndsclin3)
    ndsclin3Col = arrNdsclin3(0:kfrm-1)
    call release(ndsclin3)
    if (allocated(arrNdsclin3)) deallocate(arrNdsclin3)

    ndsclin4Col => Int16Data(ndsclin4)
    ndsclin4Col = arrNdsclin4(0:kfrm-1)
    if (allocated(arrNdsclin4)) deallocate(arrNdsclin4)
    call release(ndsclin4)

    framelimCol => Int32Data(framelim)
    framelimCol = arrframelim(0:kfrm-1)
    call release(framelim)
    if (allocated(arrFramelim)) deallocate(arrFramelim)


  endif

  if (ontime < 0) ontime = 0.0
  call setAttribute(outccd, "ONTIME", ontime, "s", "Total good times")

  if (empty_frames) then
    livetime = max(ontime*FC0,0.0d0)
    call setAttribute(outccd, "LIVETIME", livetime, "s", &
               "Good times * FC0")
  else
    livetime = max((ontime-T_Counting_Mode)*FC0,0.0d0)
    call setAttribute(outccd, "LIVETIME", livetime, "s", &
               "(Good times - Counting Mode) * FC0")
  endif

  call setAttribute(outccd, "EXPOSURE", livetime, "s", &
               "temporary = LIVETIME")

  call setAttribute(outccd, "COUNTMOD",T_Counting_Mode, "s", &
               "Intervals in Counting Mode of this quadrant")

  call setAttribute(outccd, "NDSCLINQ", ndsclinq, " ", &
               "MIP rejected columns total quadrant (AUX)")
  call setAttribute(outccd, "NDSCLINA", ndsclinq_ave, " ", &
               "MIP rejected columns per AN_MAXFRC frames (AUX)")
  call setAttribute(outccd, "NDSCLINC", ndsclinc, " ", &
               "MIP rejected columns single CCD (SAS)")

  if (illegalODF) &
        call setAttribute(outccd, "WARN_101", "illegalODF", &
        comment="important warning in task epframes")    

  if (noPNAUX2) &
        call setAttribute(outccd, "WARN_102", "noPNAUX2", &
        comment="important warning in task epframes")    

  if (singleBoresightBurst) &
        call setAttribute(outccd, "WARN_103", "singleBoresightBurst", &
        comment="important warning in task epframes")    

  if (doubleWindowTiming) &
        call setAttribute(outccd, "WARN_104", "doubleWindowTiming", &
        comment="important warning in task epframes")    

  if (wrongBPTcode) &
        call setAttribute(outccd, "WARN_105", "wrongBPTcode", &
        comment="important warning in task epframes")    

  if (CCDlowGain) &
        call setAttribute(outccd, "WARN_106", "CCDlowGain", &
        comment="important warning in task epframes")    



  call addAttributes(attributable(outexpo),attributable(outccd))

  if (MIP_ext) then
    call addAttributes(attributable(mip),attributable(outccd))
    write(map_title,*) 'MIPs only (PHA $>',MIP_threshold, '$ adu)'
    call setAttribute(mip, "MAPLATEX", map_title(2:), &
               comment="Description of event map")
    write(fileMap,'(a,i2.2)') 'mipmap',CCDno
    call setAttribute(mip, "FILENAME", trim(fileMap), &
               comment="Identifier of event map")
    mip_arrayP   = mip_array

    call addAttributes(attributable(nomip),attributable(outccd))
    write(map_title,*) 'Singles, no MIPs, $',NOMIP_loth,'-',NOMIP_hith,'$ adu'
    call setAttribute(nomip, "MAPLATEX", map_title(2:), &
               comment="Description of event map")
    write(fileMap,'(a,i2.2)') 'nomipmap',CCDno
    call setAttribute(nomip, "FILENAME", trim(fileMap), &
               comment="Identifier of event map")
    nomip_arrayP = nomip_array

    call addAttributes(attributable(soft),attributable(outccd))
    write(map_title,*) 'Singles, no MIPs, $',SOFT_loth,'-',SOFT_hith,'$ adu'
    call setAttribute(soft, "MAPLATEX", map_title(2:), &
               comment="Description of event map")
    write(fileMap,'(a,i2.2)') 'soft',CCDno
    call setAttribute(soft, "FILENAME", trim(fileMap), &
               comment="Identifier of event map")
    soft_arrayP  = soft_array

    call addAttributes(attributable(total),attributable(outccd))
    write(map_title,*) 'All events, not recombined'
    call setAttribute(total, "MAPLATEX", map_title(2:), &
               comment="Description of event map")
    write(fileMap,'(a,i2.2)') 'total',CCDno
    call setAttribute(total, "FILENAME", trim(fileMap), &
               comment="Identifier of event map")
    total_arrayP = total_array

    if (MIP_out == 2) then
      call release(outmip)
    endif
  endif

  call release(outexpo)

  if (enable_offcol_extension) then
    if (.not. hasBlock(outfile,"OFFSETS")) &
      outoffcol = addTable(outfile,"OFFSETS",ncolOffset)
    call addAttributes(attributable(outoffcol),attributable(outccd))
  endif

  if (enable_dli_extension) then
!   write(extname_dli,'(a,i2.2)') 'DLIMAP',CCDno
    write(extname_dli,'(a)') 'DLIMAP'
    if (.not. hasBlock(outfile,extname_dli)) &
      outdlimap = addTable(outfile,extname_dli,EPN_MAX_Y)
    call addAttributes(attributable(outdlimap),attributable(outccd))
  endif

! call release(time)
! call release(frame)
! call release(rawX)
! call release(rawY)
! call release(pha)
! call release(ccdnr)
! call release(qualflag)

  if (MIP_dist) then
    call release(mipdistl)
    call release(mipdistr)
    call release(miphdist)
    call release(miphistl)
    call release(miphistr)
    call release(outhist)
  endif

  call release(outccd)

  if (enable_offcol_extension .and. ncoloffset > 0) then
    rawX = addColumn(outoffcol,"RAWX",Integer16,"pixel",dims, &
                       comment="CCD column number")
    outrawXCol => int16Data(rawX)
    pha      = addColumn(outoffcol,"OFFSETX",Integer16,"channel",dims, &
                       comment="Column offset")
    outphaCol => int16Data(pha)
    ccdnr    = addColumn(outoffcol,"CCDNR",Integer8,"",dims, &
                       comment="Overall CCD Number")
    call setAttribute(ccdnr, "TLMIN",    1, comment="Minimum for CCDNR")
    call setAttribute(ccdnr, "TLMAX",   12, comment="Maximum for CCDNR")
    outccdnrCol => int8Data(ccdnr)

    k = 0
    do i = 0,nx_ccd-1
      if (on_offcol(i)) then
        outrawXCol(k)  = i+1
        outphaCol(k)   = ocv(i)
        outccdnrCol(k) = CCDno
        k = k+1
      endif
    enddo
    call release(outoffcol)
  endif

  if (enable_dli_extension) then
!   write(colname_dli1,'(a,i2.2)') 'DLIODF',CCDno
    write(colname_dli1,'(a)') 'DLIODF'
    dliodf      = addColumn(outdlimap,colname_dli1,Integer32,"columns",dims, &
                       comment="Column rejections in ODF")
    outdliodfCol => int32Data(dliodf)

!   write(colname_dli2,'(a,i2.2)') 'DLISAS',CCDno
    write(colname_dli2,'(a)') 'DLISAS'
    dlisas      = addColumn(outdlimap,colname_dli2,Integer32,"columns",dims, &
                       comment="Column rejections in SAS")
    outdlisasCol => int32Data(dlisas)

    do i = 0,EPN_MAX_Y-1
      outdliodfCol(i)  = dliarray(i+1)
      outdlisasCol(i)  = dlisasarray(i+1)
    enddo
    call release(outdlimap)
  endif

  call release(outfile)

  call message(AppMsg,Sparse,'EPFRAMES finished')

  end subroutine epframes






! ==============================
! Parameter reading, get ODFinfo
! ==============================
  subroutine getParamValues(fileName,auxfilename,CCDno,instrument,& 
           expNr,inmode,filter)

! fileName : Name of the input ODF/SDF file
! CCDno    : Number of selected CCD

  character(len=*),   intent(out) :: fileName,auxfilename,filter
  integer(kind=Int32),intent(out) :: CCDno, instrument, expNr
  integer,            intent(out) :: inmode

  integer(kind=Int16)           :: nExps, nCCDs
  integer                       :: dataMode, ip ! ,nfile
  character(len=256)            :: shortfileName, tmpIPPVname
  character(len=24)             :: cp
  logical(kind=bool)            :: lexist

  call message(AppMsg,Noisy,"... entering getParamValues...")

  call OAL_expandFileName('odf(xmm, summary)', fileName)
  write(errstr,*) "ODF summary file is   : ",&
          trim(fileName(index(fileName, '/', .true.)+1:))
  call message(AppMsg,Verbose)
  call message(AppMsg,Verbose,"")

  call printODFinfo(odfinfos)
  call printPROPOSALinfo(proposalinfo)

! ----------------------------------
! Expand input filename if necessary
! ----------------------------------

! print*,'shortfileName ..'
  shortfileName = trim(stringParameter('set'))
! print*,'oal_expand'
  call OAL_expandFileName(shortfileName,fileName)

! print*,'odfinfo ..'
  ip = len_trim(odfinfos%directoryName)+1
  if (ip > 1024) then
    write(errstr,*) 'odfinfos%directoryName is too long, it exceeds 1024 characters:'
    call message(AppMsg,Sparse)
    call message(AppMsg,Sparse,odfinfos%directoryName)
  endif
! print*,'trim ..'
  cp = trim(filename(ip+1:ip+22))
  write(odf_id,'(a4,a2,a10,a2,a6)') cp(1:4),'\_',cp(6:15),'\_',cp(17:22)
  odf_icd = cp
  scheduled = (odf_icd(19:19) == "S")

! -------------------------------------------
! define the OAL state (needed to read IPPVs)
! -------------------------------------------

! print*,'oal_setstate ..'
  call OAL_setState(filename)

! ---------------------------------------------------------------
! Read ODF/CAL/HK parameters (will probably be later no need for)
! ---------------------------------------------------------------

! print*,'correct ..'
  correct_ODF = .TRUE.
  correct_HK  = .TRUE.
  correct_CCF = .TRUE.
  setup_OCV   =  "CCF"

  correct_ODF    = booleanParameter("odfok")
  correct_HK     = booleanParameter("hkok")
  correct_CCF    = booleanParameter("ccfok")
  setup_OCV      = trim(stringParameter("setupbpx"))
  write(setup_SAS,'(l1,l1,l1)') correct_ODF,correct_HK,correct_CCF

!
! if incomplete ODF (for whatever reason): try to extract values from the rest
!
  automode       = .FALSE.                          ! future development
  autofilter     = .FALSE.                          ! future development
  automode       = booleanParameter("automode")     ! F1294 and An_CHOP
  autofilter     = booleanParameter("autofilter")   ! FILTER

  show_CCX       = booleanParameter("showccx")
  show_AUX       = booleanParameter("showaux")
  show_PMH       = booleanParameter("showpmh")
  show_PAH       = booleanParameter("showpah")
  show_VE        = booleanParameter("showve")

! ----------------------------------------------------
! Read HK parameters (will later be read from HK file)
! ----------------------------------------------------

  Frames_Offsetmap = -99
  F1030       = -1
  F1052       = 32400
  F1294       = 0
  An_CHOP     = 0
  An_CMLOTH   = 532
  An_CMCORR   = 512
  An_EAMIPSEL = 1
  An_MAXMIP   = 63
  An_MIP      = 3000+An_CMCORR
  An_MAXFRC   = 20
  An_SENDMODE = 'MIP CORREC.2'
  SAS_SENDMODE = 'MIP CORREC.2'
  ECACHn_GAINCAME = 'HIGH'
  OT_CHOP     = -1
  OT_CMLOTH   = -1
  OT_CMCORR   = -1
  OT_EAMIPSEL = -1
  OT_MAXMIP   = -1
  OT_MIP      = -1
  OT_SENDMODE = 'NO MIP CORREC'
!
  An_CMCORR      = intParameter("ancmcorr")
  An_EAMIPSEL    = intParameter("aneamipsel")
  An_MAXMIP      = intParameter("anmaxmip")
  An_MIP         = intParameter("anmip")
  An_MAXFRC      = intParameter("anmaxfrc")
  An_SENDMODE    = trim(stringParameter("ansendmode"))
! An_CCD_SWISTAT = intParameter("anccdswistat")                ! is always ON
! C_CE_QUADSEL   = intParameter("cequadsel")
  E_Cx_TEMPQB1_AVE = realParameter("ecntempqb1")
  F1128AVE       = realParameter("hcettmpfpf")

  HK_CMLOTH   = An_CMLOTH  
  HK_CMCORR   = An_CMCORR  
  HK_EAMIPSEL = An_EAMIPSEL
  HK_MAXMIP   = An_MAXMIP  
  HK_MIP      = An_MIP

  write(tmpIPPVname,'(a)') 'FILTER'
  if (OAL_hasIPPV(tmpIPPVname)) then
    filter = OAL_getIPPVstring(tmpIPPVname)
    write(errstr,*) trim(tmpIPPVname),' = ',filter
    call message(AppMsg,Noisy)
  else
    filter = trim(stringParameter("f1118"))  ! UNKNOWN
    if (correct_ODF) then
      write(errstr,*) 'Parameter FILTER not found, ', &
                      'will set it to ',filter
      call warning('missingIPPV')
    else
      call message(AppMsg,Noisy+1,'Valid EPIC PN filter names:')
      call message(AppMsg,Noisy+1,filterId(OPEN))
      call message(AppMsg,Noisy+1,filterId(CLOSED))
      call message(AppMsg,Noisy+1,filterId(THIN1))
      call message(AppMsg,Noisy+1,filterId(THIN2))
      call message(AppMsg,Noisy+1,filterId(MEDIUM))
      call message(AppMsg,Noisy+1,filterId(THICK))
      call message(AppMsg,Noisy+1,filterId(CAL_OPEN))
      call message(AppMsg,Noisy+1,filterId(CAL_CLOSED))
      call message(AppMsg,Noisy+1,filterId(CAL_THIN1))
      call message(AppMsg,Noisy+1,filterId(CAL_THIN2))
      call message(AppMsg,Noisy+1,filterId(CAL_MEDIUM))
      call message(AppMsg,Noisy+1,filterId(CAL_THICK))
    endif
  endif

  write(tmpIPPVname,'(a)') 'FRAME_TIME_PARAMETER'
  if (OAL_hasIPPV(tmpIPPVname)) then
    f1294 = OAL_getIPPVint(tmpIPPVname)
    write(errstr,*) trim(tmpIPPVname),' = ',f1294
    call message(AppMsg,Noisy)
    if ((f1294 < 0) .OR. (f1294 > 15)) then
      write(errstr,*) 'The IPPV parameter ',trim(tmpIPPVname), &
        ' has an impossible integer value of ',f1294, &
        ' (allowed 0..15), and thus the', &
        ' ODF summary file may be incorrect. ', &
        ' The task may crash as a consequence of this inconsistency.', &
        ' Please, check the value of F1294 in the corresponding PMH file',&
        ' and compare this with the value in the SUM.SAS file.'
      call error('illegalIPPVvalue')
    endif
  else
    F1294 = intParameter("f1294")            ! 0 or 3 or 5
    if (correct_ODF) then
      write(errstr,*) 'Parameter FRAME_TIME_PARAMETER not found, ', &
                      'will set it to ',F1294
      call warning('missingIPPV')
    endif
  endif

  write(tmpIPPVname,'(a)') 'CLOCK_RESET_TIME_COARSE'
  if (OAL_hasIPPV(tmpIPPVname)) then
    f1030 = OAL_getIPPVint(tmpIPPVname)
    write(errstr,*) trim(tmpIPPVname),' = ',f1030
    call message(AppMsg,Noisy)
  else
    if (correct_ODF) then
      write(errstr,*) 'Parameter ',trim(tmpIPPVname),' not found, ', &
                      'will set it to ',F1030
      call warning('missingIPPV')
    endif
  endif

  write(tmpIPPVname,'(a)') 'CLOCK_WRAP_AROUND'
  if (OAL_hasIPPV(tmpIPPVname)) then
    f1052 = OAL_getIPPVint(tmpIPPVname)
    write(errstr,*) trim(tmpIPPVname),' = ',f1052
    call message(AppMsg,Noisy)
  else
    F1052 = intParameter("f1052")            ! 32400
    if (correct_ODF) then
      write(errstr,*) 'Parameter ',trim(tmpIPPVname),' not found, ', &
                      'will set it to ',F1052
      call warning('missingIPPV')
    endif
  endif


! ----------------------
! initialize output file
! ----------------------
 
  call message(AppMsg,Noisy,'Accessing input file ...')
  infile = dataSet(fileName,READ,USE_ENVIRONMENT)
  inblk  = block(infile, 0, READ)
  inccd  = table(infile, name(inblk) )
  deltat_odf = real64Attribute(inccd,'FRMTIME')
  ccdmode = int16Attribute(inccd,'CCDMODE')
! print*,'ccdmode = ',ccdmode

! -------------------------------------
! derive temporarily correct frame time
! -------------------------------------
  select case (ccdmode)
    case(0)
      insubmode = PRIME_FULL_WINDOW
    case(1)
      insubmode = PRIME_FULL_WINDOW_EXTENDED
    case(2)
      insubmode = PRIME_LARGE_WINDOW
    case(3)
      insubmode = PRIME_SMALL_WINDOW
    case(4)
      insubmode = FAST_TIMING
    case(5)
      insubmode = FAST_BURST
    case default
      write(errstr,*) 'caseDefaultError: CCDMODE = ',ccdmode
      call error('wrongMode')
  end select
  call getNominalFrameTime(insubmode,f1294,fc0,frmtime,rotime,psftime,shftime)

  call message(AppMsg,Noisy,'Initializing output file ...')
  call initRawOutput

! ---------------------------------------------------
! Specify which instrument/modes one is interested in
! ---------------------------------------------------

! call message(AppMsg,Noisy,' ... CAL_state 1')
! call CAL_setState(fileName)
  call CAL_setState(outfile)
! call message(AppMsg,Noisy,' ... CAL_state 2')
  call CAL_setState(instrument=EPN, randomize = .false.)
! call message(AppMsg,Noisy,' ... CAL_state 3')
  call CAL_getState(ccdModeId=insubmode)


!----------------------------------------------------------
! incorrect eFF in `odffixed' ODFs: so use this workaround:
!----------------------------------------------------------

  if (insubmode /= PRIME_FULL_WINDOW_EXTENDED .and. F1294 /= 0) then
    call message(AppMsg,Sparse,' ')
    write(errstr,*) 'Inconsistent CCDMODE: F1294 = ',F1294, &
              ', INSUBMODE = ',insubmode,' = ',modeId(insubmode)
    call message(AppMsg,Sparse)
    write(errstr,*) 'task will assume PRIME_FULL_WINDOW_EXTENDED ( =', &
              PRIME_FULL_WINDOW_EXTENDED,') in the following.'
    call message(AppMsg,Sparse)
    call warning('changedMode','Data mode has been changed from '// &
              'PRIME_FULL_WINDOW to PRIME_FULL_WINDOW_EXTENDED')
    call message(AppMsg,Sparse,' ')
    insubmode = PRIME_FULL_WINDOW_EXTENDED
    call CAL_setState(instrument=EPN, ccdModeId = insubmode)
  endif

!--------------------------------------
! ... and for the opposite way as well:
!--------------------------------------

  if (insubmode == PRIME_FULL_WINDOW_EXTENDED .and. F1294 == 0) then
    call message(AppMsg,Sparse,' ')
    write(errstr,*) 'Inconsistent CCDMODE: F1294 = ',F1294, &
              ', INSUBMODE = ',insubmode,' = ',modeId(insubmode)
    call message(AppMsg,Sparse)
    write(errstr,*) 'task will assume PRIME_FULL_WINDOW ( =', &
              PRIME_FULL_WINDOW,') in the following.'
    call message(AppMsg,Sparse)
    call warning('changedMode','Data mode has been changed from '// &
              'PRIME_FULL_WINDOW_EXTENDED to PRIME_FULL_WINDOW')
    call message(AppMsg,Sparse,' ')
    insubmode = PRIME_FULL_WINDOW
    call CAL_setState(instrument=EPN, ccdModeId = insubmode)
  endif

  if (.not. verify_PMH) call deleteBlock(outfile,"EVENTS")  ! not needed anymore, will be re-created

! call message(AppMsg,Noisy,' ... CAL_state 4')
  call CAL_getState(ccdModeId=insubmode)
  obsSubMode = modeId(insubmode)
  write(errstr,*) 'CAL is set to ccdModeID = ',insubmode,' = ',obsSubMode
  call message(AppMsg,Verbose)
  call message(AppMsg,Verbose,' ')
! call message(AppMsg,Noisy,' ... CAL_state done')


  sciencefilename = trim(fileName(index(fileName, '/', .true.)+1:))
  write(errstr,*) 'EPIC pn Science File : ',&
          trim(fileName(index(fileName, '/', .true.)+1:))
  call message(AppMsg,Verbose)


! ----------------------
! check instrument modes
! ----------------------

  if (insubmode == FAST_TIMING) then
    inmode = TIMING
  elseif (insubmode == FAST_BURST) then
    inmode = BURST
  elseif (insubmode == PRIME_FULL_WINDOW .or. &
          insubmode == PRIME_FULL_WINDOW_EXTENDED .or. &
          insubmode == PRIME_LARGE_WINDOW .or. &
          insubmode == PRIME_SMALL_WINDOW) then
    inmode = IMAGING
  elseif (insubmode == PRIME_FULL_OFFSET .or. &
          insubmode == PRIME_LARGE_OFFSET .or. &
          insubmode == OFFSET .or. &
          insubmode == NOISE .or. &
          insubmode == DIAGNOSTIC .or. &
          insubmode == EXTRA_HEATING_DEICING .or. &
          insubmode == EXTRA_HEATING_DECONTAM .or. &
          insubmode == IN_FLIGHT_TEST) then
    call error('wrongMode','Not scientific data mode')
  else
    call error('wrongMode','Illegal or missing type of data mode')
  endif


  nExps = OAL_selectScope(EPN, inmode)
  call OAL_getState(instrument,dataMode)
  write(errstr,*) "Number of EPN exposures in mode ",dataMode,": ",nExps 
  call message(AppMsg,Noisy)
  call message(AppMsg,Noisy,"")

  call message(AppMsg,Noisy,"Infos extracted from ODF with OAL_getState:")
  write(errstr,*) "INSTRUMENT: ",instrument
  call message(AppMsg,Noisy)
  write(errstr,*) "DATAMODE  : ",dataMode
  call message(AppMsg,Noisy)
  call message(AppMsg,Noisy,"")

! ---------------------------------------------------
! Specify which exposure and CCD one is interested in 
! ---------------------------------------------------

  scheduled = .true.
  call OAL_getState(instrument,dataMode,expNr,scheduled,CCDno)
! call OAL_getState(instrument,dataMode,expNr,CCDno)
  nCCDs = OAL_selectExposure(expNr)
  call printEXPOSUREinfo(expinfos,nCCDs)

  quadrant = int16Attribute(inccd,"QUADRANT")
  ccdid    = int16Attribute(inccd,"CCDID")

  if (OAL_hasIPPV("LOWER_THRESHOLD_Q1_CCD3")) then
    write(errstr,*) 'LowerThresholdIPPV uses inconsistent CCD numbering, ', &
                    'task will add 1 to the hardware CCD number.'
    call message(AppMsg,Sparse)
    write(tmpIPPVname,'(a,i1,a,i1)') 'LOWER_THRESHOLD_Q',quadrant,'_CCD',CCDid+1
  else
    write(tmpIPPVname,'(a,i1,a,i1)') 'LOWER_THRESHOLD_Q',quadrant,'_CCD',CCDid
  endif
  if (OAL_hasIPPV(tmpIPPVname)) then
    An_CMLOTH = OAL_getIPPVint(tmpIPPVname)
  else
    An_CMLOTH = intParameter("ancmloth")           !   532
    if (correct_ODF) then
      write(errstr,*) 'Parameter ',trim(tmpIPPVname),' not found, ', &
                      'will set it to ',An_CMLOTH
      call warning('missingIPPV')
    endif
  endif
!
! 8-bit chopper value [0-255]
!
  write(tmpIPPVname,'(a,i1)') 'QUADRANT_UNDERSAMPLING_Q',quadrant
  if (OAL_hasIPPV(tmpIPPVname)) then
    An_CHOP = OAL_getIPPVint(tmpIPPVname)
  else
    An_CHOP = intParameter("anchop")
    if (correct_ODF) then
      write(errstr,*) 'Parameter ',trim(tmpIPPVname),' not found, ', &
                      'will set it to ',An_CHOP
      call warning('missingIPPV')
    endif
  endif


! ---------------------------
! Select Data from HK files
! Select Data from AUX file
! ---------------------------

  auxfilename = 'dummy'
! call OAL_setState(instrument,AUXILIARY)
  lexist = OAL_hasAssociatedSet(AUXILIARY)
  if (lexist) then
!   ip = OAL_selectScope(instrument,AUXILIARY)
!   if (ip < 1) call message(AppMsg,Sparse,'AUX: ODF might be inconsistent.')
!   call OAL_selectFile(0,auxfilename)
    auxfilename = OAL_selectAssociatedSet(AUXILIARY)
    write(errstr,*) 'EPIC pn Auxil File   : ',&
!           trim(auxfilename(len_trim(odfinfos%directoryName)+1:))
          trim(auxfileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    exist_aux = .true.
  else
    call message(AppMsg,Sparse,&
            'OAL: no AUX file present, ODF might be inconsistent.')
    exist_aux = .false.
  endif

  ccxfilename = 'dummy'
! call OAL_setState(instrument,COUNTING_CYCLE_REPORT)
  lexist = OAL_hasAssociatedSet(COUNTING_CYCLE_REPORT)
  if (lexist) then
    ccxfilename = OAL_selectAssociatedSet(COUNTING_CYCLE_REPORT)
!   ip = OAL_selectScope(instrument,COUNTING_CYCLE_REPORT)
!   if (ip < 1) call message(AppMsg,Sparse,'CCX: ODF might be inconsistent.')
!   call OAL_selectFile(0,ccxfilename)
    write(errstr,*) 'EPIC pn Counting Mode: ',&
!           trim(ccxfilename(len_trim(odfinfos%directoryName)+1:))
          trim(ccxfileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    exist_ccx = .true.
  else
    ip = index(auxfilename,'AUX.FIT')
    write(ccxfilename,'(a,a)') auxfilename(0:ip-1),'CCX.FIT'
    write(errstr,*) 'EPIC pn CCX missing  : ',&
!           trim(ccxfilename(len_trim(odfinfos%directoryName)+2:))
          trim(ccxfileName(index(fileName, '/', .true.)+2:))
    call message(AppMsg,Noisy)
    exist_ccx = .false.
  endif

  bad_dli = .false.
  exist_dli = .false.
  if (enable_DISCARDED_LINES) then
!
! a good test file is 0086_0124711401_PNU00807DLI.FIT
!
  dlifilename = 'dummy'
  lexist = OAL_hasAssociatedSet(DISCARDED_LINES_DATA)
  if (lexist) then
    bad_dli = .true.
    dlifilename = OAL_selectAssociatedSet(DISCARDED_LINES_DATA)
    write(errstr,*) 'EPIC pn Disc. Lines  : ',&
          trim(dlifileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    dlifile = dataSet(dlifilename,READ,USE_ENVIRONMENT)
    write(errstr,*) "PNDLI1: ",hasBlock(dlifile,"PNDLI1")
    call message(AppMsg,Noisy)
    dli = array(dlifile,'PNDLI1')
    dli_arrayP => uint16Array2Data(dli)
    dliarray(1:ny_ccd) = dli_arrayP(0,0:ny_ccd-1)
    where (dliarray < 0) dliarray = dliarray+65536   ! uint16 -> int32
    if (verbosity(AppMsg) >= Noisy) then
      do i = 1,nx_ccd
        write(errstr,'(1x,i3,i6)') i,dliarray(i)
        call message(AppMsg,Sparse)
      enddo
    endif
    do i = 1,ny_ccd
      if (i.ne.dliarray(I)-1-ccdid*ny_ccd) bad_dli = .false.
    enddo
    if (bad_dli) then
      write(errstr,*) trim(dlifilename), &
          ' contains only dummy quadrant running column index, will be ignored.'
      call warning('badDLI')
    endif
    call release(dli)
    call release(dlifile)
    if (bad_dli) then
      exist_dli = .false.
    else
      exist_dli = .true.
    endif
  else
    if (inmode == IMAGING .and. insubmode .ne. PRIME_SMALL_WINDOW) then
      call message(AppMsg,Verbose,'... found no DLI file')
    endif
    exist_dli = .false.
  endif

  endif

  if (enable_DATA_DIAGNOSTIC) then

  dlifilename = 'dummy'
! call OAL_setState(instrument,DATA_DIAGNOSTIC)
  lexist = OAL_hasAssociatedSet(DATA_DIAGNOSTIC)
  if (lexist) then
    dlifilename = OAL_selectAssociatedSet(DATA_DIAGNOSTIC)
!   ip = OAL_selectScope(instrument,DATA_DIAGNOSTIC)
!   if (ip < 1) call message(AppMsg,Sparse,'DLI: ODF might be inconsistent.')
!   call OAL_selectFile(CCDNo,dlifilename)
    write(errstr,*) 'EPIC pn Discarded Lines: ',&
!           trim(dlifilename(len_trim(odfinfos%directoryName)+1:))
          trim(dlifileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    exist_dli = .true.
  else
    ip = index(filename,'E.FIT')
    read(filename(ip-7:ip-5),'(i3)') expNo
    write(dlifilename,'(a,a,i3.3,i2.2,a)') &
       filename(0:ip-11),'PNU',CCDNo+EXPNo,CCDNo,'DLI.FIT'
    call OAL_selectFile(CCDNo,dlifilename)
    write(errstr,*) 'EPIC pn DLI missing  : ',&
!           trim(dlifilename(len_trim(odfinfos%directoryName)+2:))
          trim(dlifileName(index(fileName, '/', .true.)+2:))
    call message(AppMsg,Noisy)
    exist_dli = .false.
  endif

  endif

  pmhfilename = 'dummy'
! call OAL_setState(instrument,MAIN_PERIODIC_HOUSEKEEPING)
  lexist = OAL_hasAssociatedSet(MAIN_PERIODIC_HOUSEKEEPING)
  if (lexist) then
    pmhfilename = OAL_selectAssociatedSet(MAIN_PERIODIC_HOUSEKEEPING)
    write(errstr,*) 'EPIC pn Main HK File : ',&
!           trim(pmhfilename(len_trim(odfinfos%directoryName)+1:))
          trim(pmhfileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    exist_pmh = .true.
    if (verify_PMH) then
      call verifyPMH(pmhfilename,f1052)
      call CAL_setState(outfile)
      call CAL_setState(instrument=EPN, randomize = .false.)
    endif
  else
    call message(AppMsg,Sparse,&
            'OAL: no PMH file present, ODF might be inconsistent.')
    exist_pmh = .false.
  endif
  if (verify_PMH) call deleteBlock(outfile,"EVENTS")  ! not needed anymore, will be re-created

  pahfilename = 'dummy'
! call OAL_setState(instrument,ADD_PERIODIC_HOUSEKEEPING)
  lexist = OAL_hasAssociatedSet(ADD_PERIODIC_HOUSEKEEPING)
  if (lexist) then
    pahfilename = OAL_selectAssociatedSet(ADD_PERIODIC_HOUSEKEEPING)
    write(errstr,*) 'EPIC pn AddP HK File : ',&
!           trim(pahfilename(len_trim(odfinfos%directoryName)+1:))
            trim(pahfileName(index(fileName, '/', .true.)+1:))
    call message(AppMsg,Verbose)
    exist_pah = .true.
  else
    if (exist_pmh) then
      ip = index(pmhfilename,'PMH.FIT')
    else
      ip = 5
    endif
    write(pahfilename,'(a,a)') pmhfilename(0:ip-1),'PAH.FIT'
    write(errstr,*) 'EPIC pn PAH missing  : ',&
            trim(pahfileName(index(fileName, '/', .true.)+2:))
    call message(AppMsg,Noisy)
    if (correct_odf) call message(AppMsg,Sparse,&
            'OAL: no PAH file present, ODF might be inconsistent.')
    exist_pah = .false.
  endif

  call message(AppMsg,Verbose,'')

! ---------------------------
! Select Data from single CCD
! ---------------------------

! call OAL_setState(filename)
! call OAL_selectFile(CCDno, fileName)


!------------------------------------------------------------
! quadrant and CCD selection: D210|C210|B210|A210: A,B,C,D = Q0..3
!------------------------------------------------------------
  if (insubmode == PRIME_FULL_WINDOW .or. &
      insubmode == PRIME_FULL_WINDOW_EXTENDED .or. &
      insubmode == PRIME_LARGE_WINDOW) then
    AX_QUAD_CCD_SEL = -32768
  elseif (insubmode == PRIME_SMALL_WINDOW .or. &
      insubmode == FAST_TIMING .or. &
      insubmode == FAST_BURST) then
    AX_QUAD_CCD_SEL = 144
  else
    AX_QUAD_CCD_SEL = 0
  endif

  Ax_LowThresh = An_CMLOTH - An_CMCORR 

! ----------------------------
! Read wrongpixlimit parameter
! ----------------------------

  wronglimit = intParameter('wrongpixlimit')

! --------------------------------
! Read srcposition parameter:
! to be used in TIMING/BURST mode
! --------------------------------

  withsrcrawy  = booleanParameter('withsrcrawy')
  sourceY = intParameter('srcposition')

  if (sourceY < 181 .and. inmode == BURST) then
    call warning('srcpos', &
                 'BURST mode: expected source position: 181-200')
  endif

  withsrccoords  = booleanParameter('withsrccoords')
  srcRA  = realParameter('srcra')
  srcDEC = realParameter('srcdec')

! ------------------------------------------------------------
! Read MIP and quality parameter
!   for analysis of commissioning and calibration phase data
!-------------------------------------------------------------

  witheventmap  = booleanParameter('witheventmap')
  eventmapset   = trim(stringParameter('eventmapset'))
  if (witheventmap) then
    MIP_out = 2
  else
!   MIP_out       = intParameter('crmap') 
    MIP_out       = 0
  endif
  MIP_ext       = (MIP_out > 0)
  qualMax       = intParameter('qualmax')

  MIP_Threshold = intParameter('mipthreshold')
  MIP_Method    = trim(stringParameter('mipmethod'))
  MIP_Dist      = booleanParameter('mipdist')
  MIP_Discard   = booleanParameter('mipdiscard')
  NOMIP_loth    = intParameter('nomipthresholdlow')
  NOMIP_hith    = intParameter('nomipthresholdhigh')
  SOFT_loth     = intParameter('softthresholdlow')
  SOFT_hith     = intParameter('softthresholdhigh')

  select case (MIP_Method)
    case("none")
      mipcorr = .false.
      mipcols = .false.
      mipnew  = .false.
      MIP_Method = 'NONE'
    case("onboard")
      mipcorr = .true.
      mipcols = .false.
      mipnew  = .true.
      MIP_Method = 'ONBOARD'
      MIP_Threshold = An_MIP - An_CMCORR
    case("com")
      mipcorr = .true.
      mipcols = .true.
      mipnew  = .false.
      MIP_Method = 'COM'
    case("sas")
      mipcorr = .true.
      mipcols = .false.
      mipnew  = .false.
      MIP_Method = 'SAS'
  end select

  end subroutine getParamValues




! =============================================================================
! print ODF infos
! =============================================================================
  subroutine printODFinfo(odfinfos)

  type(OdfInfoType)     ,intent(out)   :: odfinfos 

  call message(AppMsg,Noisy+1,"... entering OAL_odfInfo...")
  call OAL_odfInfo(odfinfos)

  if (verbosity(AppMsg) >= Verbose) then 
    write(errstr,*) "XMM ODF in directory  : ",trim(odfinfos%directoryName)
    call message(AppMsg,Verbose)  
    write(errstr,*) "revolution number     :",odfinfos%revolutionNr    
    call message(AppMsg,Verbose)
    write(errstr,*) "proposal id           : ",odfinfos%proposalId
    call message(AppMsg,Verbose)
    write(errstr,*) "total number of files :",odfinfos%nrFiles
    call message(AppMsg,Verbose)
    write(errstr,*) "observation id        : ",odfinfos%observationId
    call message(AppMsg,Verbose)
    write(errstr,*) "observation start time: ", &
                 odfinfos%observationStartTime
    call message(AppMsg,Verbose)
    write(errstr,*) "observation end time  : ", &
                 odfinfos%observationEndTime
    if (odfInfos%isSlew) then
      write(errstr,*) "type of ODF           : SLEW"
    else
      write(errstr,*) "type of ODF           : POINTING"  
    endif
    call message(AppMsg,Verbose)
  endif

  end subroutine printODFinfo 




! =============================================================================
! print PROPOSAL infos
! =============================================================================
  subroutine printPROPOSALinfo(proposalinfo)

  type(ProposalInfoType),intent(out)   :: proposalinfo

  call message(AppMsg,Noisy+1,"... entering OAL_proposalInfo...")
  call OAL_proposalInfo(proposalinfo)

  if (verbosity(AppMsg) >= Verbose) then 
    write(errstr,*) "name of the GO        : ",trim(proposalinfo%GoName)
    call message(AppMsg,Verbose)
    write(errstr,*) "institute of the GO   : ",trim(proposalinfo%GoInstitute)
    call message(AppMsg,Verbose)
    write(errstr,*) "postal addr. of the GO: ",trim(proposalinfo%GoAddress)
    call message(AppMsg,Verbose)
    write(errstr,*) "e-mail addr. of the GO: ",trim(proposalinfo%GoEmail)
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,*) "AO of proposal        : ",trim(proposalinfo%AO)
    call message(AppMsg,Verbose)
    write(errstr,*) "science type          : ",trim(proposalinfo%scienceType)
    call message(AppMsg,Verbose)  
    write(errstr,*) "name of the target    : ",trim(proposalinfo%targetName)
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,"(a,f10.5)") " Ra of target     [deg]: ",proposalinfo%targetRa
    call message(AppMsg,Verbose)
    write(errstr,"(a,f10.5)") " Dec of target    [deg]: ",proposalinfo%targetDec
    call message(AppMsg,Verbose)
    write(errstr,"(a,f10.5)") " Ra of boresight  [deg]: ",proposalinfo%boresightRa
    call message(AppMsg,Verbose)
    write(errstr,"(a,f10.5)") " Dec of boresight [deg]: ",proposalinfo%boresightDec
    call message(AppMsg,Verbose)
    write(errstr,"(a,f10.1)") " duration of obser. [s]: ",proposalinfo%obsDuration
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,*) "pos. angle constraints: ", &
          trim(proposalinfo%posAngleConstLo), trim(proposalinfo%posAngleConstHi)
    call message(AppMsg,Verbose)
    write(errstr,*) "pos. angle orig. ref. : ", &
          trim(proposalinfo%posAngleOrigRef)
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,*) "instrument priorities : emos1: ",proposalinfo%emos1Prio
    call message(AppMsg,Verbose)
    write(errstr,*) "                        emos2: ",proposalinfo%emos2Prio
    call message(AppMsg,Verbose)
    write(errstr,*) "                        epn  : ",proposalinfo%epnPrio
    call message(AppMsg,Verbose)
    write(errstr,*) "                        rgs1 : ",proposalinfo%rgs1Prio
    call message(AppMsg,Verbose)
    write(errstr,*) "                        rgs2 : ",proposalinfo%rgs2Prio
    call message(AppMsg,Verbose)
    write(errstr,*) "                        om   : ",proposalinfo%omPrio
    call message(AppMsg,Verbose)    
    call message(AppMsg,Verbose,"")
  endif

  end subroutine printPROPOSALinfo




! =============================================================================
! print EXPOSURE infos
! =============================================================================
  subroutine printEXPOSUREinfo(expinfos,nCCDs)

  integer(kind=Int16),intent(in)          :: nCCDs
  type(ExposureInfoType),intent(out)      :: expinfos

  call message(AppMsg,Noisy+1,"... entering OAL_exposureInfo...")
  call OAL_exposureInfo(expInfos)

  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,"(a,i2,a,i2,a)") " exposure number       : ", &
          expInfos%number," (with ",nCCDs," active CCDs)"
    call message(AppMsg,Verbose)
    write(errstr,*)  "exposure scheduled?   :", expInfos%scheduled
    call message(AppMsg,Noisy)
    if (expInfos%type == 0) then
      call message(AppMsg,Noisy," type of exposure      : SCIENCE")
    else
      call message(AppMsg,Noisy," type of exposure      : CALIBRATION")
    endif
    write(errstr,*) "active filter         : ", filter
    call message(AppMsg,Noisy)
    write(errstr,*) "scheduled start time  : ", &
          expInfos%scheduledStartTime(1:23)
    call message(AppMsg,Noisy)
    write(errstr,*) "actual    start time  : ", &
          expInfos%actualStartTime(1:23)
    call message(AppMsg,Noisy)
    write(errstr,*) "scheduled end   time  : ", &
          expInfos%scheduledEndTime(1:23)
    call message(AppMsg,Noisy)
    write(errstr,*) "actual    end   time  : ", &
          expInfos%actualEndTime(1:23)
    call message(AppMsg,Noisy)
    write(errstr,*) "scheduled length of exposure [s]: ", &
          toTimeTag(expInfos%scheduledEndTime(1:23)) - &
          toTimeTag(expInfos%scheduledStartTime(1:23))
    call message(AppMsg,Noisy)
    write(errstr,*) "actual    length of exposure [s]: ", &
          toTimeTag(expInfos%actualEndTime(1:23)) - &
          toTimeTag(expInfos%actualStartTime(1:23))
    call message(AppMsg,Noisy)
    call message(AppMsg,Noisy,"")
  endif

  end subroutine printEXPOSUREinfo 



! =============================================================================
! Declare DAL: open input events file
! =============================================================================
  subroutine declareDal(mxdim,mydim,mx0,my0,ntrow,nxdim,nydim,nx0,ny0,&
                        frameNr_first,frameNr_last,insubmode)

! ntrow    : Number of rows in the input file [1..ntrow]

  integer            ,intent(in)    :: insubmode
  integer(kind=Int16),intent(in)    :: mxdim, mydim, mx0, my0
  integer(kind=Int16),intent(out)   :: nxdim, nydim, nx0, ny0
  integer(kind=Int32),intent(out)   :: ntrow,frameNr_first,frameNr_last
  real(kind=double)                 :: date_sec
  integer(kind=Int16)               :: iymin,iymax,ixmin,ixmax


  call message(AppMsg,Noisy,"... entering declareDal...")

! infile = dataSet(fileName,READ,USE_ENVIRONMENT)
! select case (inmode)
!   case(IMAGING)
!     inccd = table(infile,"PNIME1")
!   case(TIMING)
!     inccd = table(infile,"PNTIE1")
!   case(BURST)
!     inccd = table(infile,"PNBUE1")
! end select  

  ntrow = numberOfRows(inccd)
  write(errstr,*) "Number of events in input CCD file: ", ntrow
  call message(AppMsg,Verbose)

  if (ntrow == 0) call message(AppMsg,Sparse, &
                              "there are no events in the input file") 


  call message(AppMsg,Noisy,"Infos extracted from the selected CCD file:")

! quadrant = int16Attribute(inccd,"QUADRANT")
! ccdid    = int16Attribute(inccd,"CCDID")

  write(errstr,*) "QUADRANT: ",quadrant
  call message(AppMsg,Noisy)  
  write(errstr,*) "CCDID   : ",CCDid
  call message(AppMsg,Noisy)

!
! XMM-SOC-ICD-0004-SSD "XMM SOC to SSC ICD for ODF and SDF data"
! (Issue 2.4A, 17 Feb 2000, M.Guainazzi) mentions DATE-OBS and DATE-END
! as start and end times of the EXPOSURE (these values are in the ODF)
!
! XMM-SOC-ICD-0006-SSC (v.2.1, Sect.8, p.12) lists only DATE_OBS (but not
! DATE_END, instead DURATION) which should be the start of the observation 
!
  write(DATE_OBS,'(a)') trim(stringAttribute(inccd,"DATE-OBS"))
  write(DATE_END,'(a)') trim(stringAttribute(inccd,"DATE-END"))
  if (DATE_END .le. DATE_OBS) then
    write(errstr,*) 'Illegal header keywords: DATE-OBS >= DATE-END: [', &
                   trim(DATE_OBS),'] [',trim(DATE_END), &
                ']   The ODF is invalid and cannot be processed properly.', &
                ' Please contact the originator of this data.'
    call warning('illegalODF')
    illegal_ODF = .true.
  endif

!
! get elapsed time in particular exposure
!
  call getDateElapse(date_obs,date_end,date_sec)

  write(errstr,*) "DATE-OBS: ",date_obs
  call message(AppMsg,Noisy)
  write(errstr,*) "DATE-END: ",date_end
  call message(AppMsg,Noisy)
  call message(AppMsg,Noisy,"")


! -------------------
! Read table columns
! -------------------

  if (ntrow > 0) then

!
! access EVENTS columns
!

    tempInCol = column(inccd,"FRAME",READ)
    frameCol => int32Data(tempInCol)

    tempInCol = column(inccd,"RAWX",READ)
    rawXCol => int8Data(tempInCol)  
    ixmin = minval(rawXCol(0:ntrow-1))
    ixmax = maxval(rawXCol(0:ntrow-1))

    tempInCol = column(inccd,"RAWY",READ)
    rawYCol => int8Data(tempInCol)
    iymin = minval(rawYCol(0:ntrow-1))
    iymax = maxval(rawYCol(0:ntrow-1))

    allocate(d_rawXCol(0:ntrow-1),d_rawYCol(0:ntrow-1))
    d_rawXCol = INT(rawXCol)
    d_rawYCol = INT(rawYCol)
    do i = 0, ntrow - 1
      if (d_rawXCol(i) < 0) d_rawXCol(i) = d_rawXCol(i) + 256
      if (d_rawYCol(i) < 0) d_rawYCol(i) = d_rawYCol(i) + 256
    enddo
    ixmin = minval(d_rawXCol(0:ntrow-1))
    ixmax = maxval(d_rawXCol(0:ntrow-1))
    iymin = minval(d_rawYCol(0:ntrow-1))
    iymax = maxval(d_rawYCol(0:ntrow-1))

    tempInCol = column(inccd,"ENERGY",READ)
    phaCol => int16Data(tempInCol)

    frameNr_first = frameCol(0)
    frameNr_last  = frameCol(ntrow-1)

    write(errstr,*) 'Event range: RAWX = [',IXMIN,',',IXMAX, &
                              ']  RAWY = [',IYMIN,',',IYMAX,']'
    call message(AppMsg,Noisy)
    nxdim = ixmax-ixmin+1
    nydim = iymax-iymin+1

    if (insubmode == FAST_TIMING) then
      if (nxdim < 64) then  ! classical value
        write(errstr,*) 'Unusual event range: ', &
	         'RAWX = [',IXMIN,',',IXMAX, &
              ']  RAWY = [',IYMIN,',',IYMAX,']. ', &
	    'Dataset may be in new double-window TIMING format. ', &
	    'Window and associated flags may be wrong.'
        call warning('doubleWindowTiming')
      endif
    elseif (insubmode == FAST_BURST) then
      if (nydim < 179) then  ! classical value
        write(errstr,*) 'Unusual event range: ', &
	         'RAWX = [',IXMIN,',',IXMAX, &
              ']  RAWY = [',IYMIN,',',IYMAX,']. ', &
	    'Dataset may be in new SingleBoresight BURST format. ', &
	    'Window and associated flags may be wrong.'
         call warning('singleBoresightBurst')
      endif
    endif
  endif

  nxdim = int16Attribute(inccd,"WINDOWDX")
  nydim = int16Attribute(inccd,"WINDOWDY")

!
! Change WINDOWX0/Y0 to SAS convention of lower left corner at (1,1):
!
  nx0 = int16Attribute(inccd,"WINDOWX0") + 1
  ny0 = int16Attribute(inccd,"WINDOWY0") + 1

! no events from RAWY = 1, window attributes wrong in early data sets

  if (  ny0 == 1)   ny0 = 2
  if (nydim == 200) nydim = 199

  write(errstr,*) "Window size: ",nxdim,"x",nydim," pixels"
  call message(AppMsg,Noisy)
  write(errstr,*) "Bottom left corner of window: (RAWX,RAWY) = (", &
              nx0,",",ny0,")"
  call message(AppMsg,Noisy)
  call message(AppMsg,Noisy,"")




! --------------------------------------
! check windows against nominal values
! --------------------------------------

  if ( (nx0 /= mx0) .or. (ny0 /= my0) .or. &
      (nxdim /= mxdim) .or. (nxdim /= mxdim) ) then
    write(errstr,'(a,9i4)') &
       " Window inconsistency: MODE, X0, Y0, DX, DY (nom,data) =  ", &
       insubmode,mx0,my0,mxdim,mydim,nx0,ny0,nxdim,nydim
    call message(AppMsg,Sparse)
  else
    write(errstr,'(a,5i4)') &
       " Window consistent with nominal values: MODE, X0, Y0, DX, DY =  ", &
       insubmode,mx0,my0,mxdim,mydim
    call message(AppMsg,Noisy)
  endif

  nxmax = nx0 + nxdim - 1
  nymax = ny0 + nydim - 1


  end subroutine declareDal





! =============================================================================
! initialize empty output set with (standard) header keywords, e.g.
!   TELESCOP, INSTRUME, DATATYPE, WINDOW*, FILTER, DATE-OBS,
!   CCDID, QUADRANT, FRMTIME, CCDMODE
! =============================================================================
  subroutine initRawOutput

  call message(AppMsg,Noisy,'Entering initRawOutput...')
  outfile = dataSet(stringParameter("eventset"),CREATE,USE_ENVIRONMENT)
! outfile = dataSet(stringParameter("eventset"),CREATE,LOW_MEMORY)
  call message(AppMsg,Noisy+1,'... done outfile ...')
  outccd = addTable(outfile,"EVENTS",1)
  call message(AppMsg,Noisy+1,'... done outccd ...')
  call setAttributes(attributable(outfile),attributable(infile))
  call message(AppMsg,Noisy+1,'... attr outfile ...')
  call setAttributes(attributable(outccd),attributable(inccd))
  call message(AppMsg,Noisy+1,'... attr outccd ...')
  call setAttribute(outfile,"FILTER",filter, "Filter ID")
  call message(AppMsg,Noisy+1,'... attr CAL ...')
  call setAttribute(outfile,"FRMTIME",NINT(frmtime),"ms",&
                   "Frame Time, nearest integer")
  call message(AppMsg,Noisy,'Leaving initRawOutput...')

  end subroutine initRawOutput





! =============================================================================
! create empty output set 
! =============================================================================
  subroutine createRawOutput(inmode,insubmode,revolution)
! --------------------------------------------------------------
! Create empty output set with (non-standard) header keywords
!   DATE_OBS, DATE_END, SOURCEY, INCCD, OUTFILE
! --------------------------------------------------------------

  integer, intent(in)               :: inmode, insubmode ! readout mode of the current CCD
  integer(kind=Int16), intent(in)   :: revolution  ! revolution number
  integer(kind=Int32), dimension(1) :: dims
  character(len=1024)               :: mipname,nomipname, &
                                       totalname,softname
  character(len=80)                 :: obsid,expname,expid
  character(len=8)                  :: eventMapExtension
! real(kind=double)                 :: duration


  call message(AppMsg,Noisy,'Entering createRawOutput...')

  call setAttribute(outfile,"TELESCOP","XMM","Telescope (mission) name")
  call setAttribute(outfile,"INSTRUME","EPN","Instrument name")
  call setAttribute(outfile,"FILTER",filter, "Filter ID")
  call setAttribute(outfile,"REVOLUT",revolution, &
       comment="XMM revolution number")
  if (hasAttribute(outfile,"CATEGORY")) then
    call deleteAttribute(outfile,"CATEGORY")
    call message(AppMsg,Noisy,'  Deleted attribute CATEGORY')
  endif

  if (odfInfos%isSlew) then
    call setAttribute(outfile,"OBS_MODE","SLEW", &
                     "Observation mode (pointing or slew)")
  else
    call setAttribute(outfile,"OBS_MODE","POINTING", &
                     "Observation mode (pointing or slew)") 
  endif

  write(obsid,"(2a)") trim(odfinfos%proposalId),trim(odfinfos%observationId)
  call setAttribute(outfile,"OBS_ID",obsid,"Observation identifier") 
  expid = stringAttribute(inccd,"EXP_ID")
  call setAttribute(outfile,"EXP_ID",expid,"Exposure identifier")
  call setAttribute(outfile,"EXPIDSTR",expidstr,"Exposure identifier (Xnnn)")

  if (inmode == IMAGING .or. inmode == TIMING .or. inmode == BURST) then
    obsSubMode = modeId(insubmode)
    select case (inmode)
    case(IMAGING)
      call setAttribute(outfile,"CONTENT", &
                     "EPIC IMAGING MODE SINGLE CCD INTERMEDIATE EVENT LIST", &
                     "Content of file")
      call setAttribute(outfile,"DATAMODE","IMAGING", &
                     "Instrument mode (IMAGING, TIMING, BURST, etc.)")
    case(TIMING)
      call setAttribute(outfile,"CONTENT", & 
                        "EPIC TIMING MODE INTERMEDIATE EVENT LIST", &
                        "Content of file")
      call setAttribute(outfile,"DATAMODE","TIMING", &
                        "Instrument mode (IMAGING, TIMING, BURST, etc.)")
    case(BURST)
      call setAttribute(outfile,"CONTENT", & 
                        "EPIC BURST MODE INTERMEDIATE EVENT LIST", &
                        "Content of file")
      call setAttribute(outfile,"DATAMODE","BURST", &
                        "Instrument mode (IMAGING, TIMING, BURST, etc.)")
    end select
    call setAttribute(outfile,"SUBMODE",obsSubMode,"Readout submode")
  endif

  call setAttribute(outfile,"ORIGIN","Leicester/SSC","origin of FITS file")
  call setAttribute(outfile,"OBJECT",trim(proposalinfo%targetName), &
                            "Name of observed object")

  write(OBS_START,'(a)') trim(odfinfos%observationStartTime)
  write(OBS_STOP,'(a)')  trim(odfinfos%observationEndTime)
!
! get elapsed time in particular observation
!
  call getDateElapse(obs_start,obs_stop,duration)

  call setAttribute(outfile,"DATE-OBS",date_obs, &
                            "Start Time (UTC) of exposure")
  call setAttribute(outfile,"DATE-END ",date_end, &
                            "End Time (UTC) of exposure")
  call setAttribute(outfile,"DATE_OBS",OBS_START, &
                            "Date observations were made")
  call setAttribute(outfile,"DATE_END",OBS_STOP, &
                            "Date observations ended")
  call setAttribute(outfile,"DURATION",DURATION,"s", &
                            "Duration of observation")

  call writeTIMEattributes(outfile,outccd,0)

  call setAttribute(outfile,"OBSERVER",trim(proposalinfo%GoName), &
                            "Name of PI")
  call setAttribute(outfile,"RA_OBJ",proposalinfo%targetRa,"deg", &
                            "Ra of the observed object")
  call setAttribute(outfile,"DEC_OBJ",proposalinfo%targetDec,"deg", &
                            "Dec of the observed object")
  call setAttribute(outfile,"RA_NOM",proposalinfo%boresightRa,"deg", &
                            "Ra of boresight")
  call setAttribute(outfile,"DEC_NOM",proposalinfo%boresightDec,"deg", &
                            "Dec of boresight")
  if ((inmode == TIMING .or. inmode == BURST) .or. &
                            (.NOT.wrong_srcpos)) then
     call setAttribute(outfile,"SRCPOS",sourceY,"pixel",&
                            "Assumed RAWY source position for fast modes")
  endif
  if ((inmode == TIMING .or. inmode == BURST) .and. &
                            (withsrccoords.AND..NOT.wrong_srcpos)) then
     call setAttribute(outfile,"SRCPOS_X",xrawsrc,"pixel",&
                            "RAWX source position")
     call setAttribute(outfile,"SRCRA",srcra,"deg",&
                            "Input RA source position")
     call setAttribute(outfile,"SRCDEC",srcdec,"deg",&
                            "Input DEC source position")
  endif

! -------------------------------------------------------------
! write exposure and observation related START and STOP
! -------------------------------------------------------------

  call setAttribute(outfile,"EXPSTART",date_obs, &
                            "Start Time (UTC) of exposure")
  call setAttribute(outfile,"EXPSTOP",date_end, &
                            "End Time (UTC) of exposure")
  call setAttribute(outfile,"OBSSTART",obs_start, &
                            "Date observation was started")
  call setAttribute(outfile,"OBSSTOP ",obs_stop, &
                            "Date observation was stopped")

! call setAttribute(outfile,"BPT_CODE",BPT_CODE,"BPT_CODE F1529_F1530_F1528")

! -------------------------------------------
! Create events table extension (one per CCD)
! -------------------------------------------

  dims(1) = 1

  if (.not. hasBlock(outfile,"EVENTS")) &
    outccd = addTable(outfile,"EVENTS",ngrow)

  call setAttributes(attributable(outccd),attributable(inccd))
  call setAttribute(outccd,"FRMTIME",NINT(frmtime),"ms", &
                           "Frame Time, nearest integer")
  call setAttribute(outccd,"OBS_ID",obsid,"Observation identifier")
  call setAttribute(outccd,"EXP_ID",expid,"Exposure identifier")
  call setAttribute(outccd,"DATAMODE",stringAttribute(outfile,"DATAMODE"), &
                           "Instrument mode (IMAGING, TIMING, FAST, etc.)")
  call setAttribute(outccd,"SUBMODE",stringAttribute(outfile,"SUBMODE"), &
                           "Readout submode")

! -------------------------------------------------------------
! write exposure and observation related START and STOP
! -------------------------------------------------------------

  call setAttribute(outccd,"DATE_OBS",OBS_START, &
                            "Date observations were made")
  call setAttribute(outccd,"DATE_END",OBS_STOP, &
                            "Date observations ended")
  call setAttribute(outccd,"DURATION",DURATION,"s", &
                            "Duration of observation")

  call setAttribute(outccd,"EXPSTART",date_obs, &
                            "Start Time (UTC) of exposure")
  call setAttribute(outccd,"EXPSTOP ",date_end, &
                            "End Time (UTC) of exposure")
  call setAttribute(outccd,"OBSSTART",obs_start, &
                            "Date observation was started")
  call setAttribute(outccd,"OBSSTOP ",obs_stop, &
                            "Date observation was stopped")


  call setAttribute(outccd,"REVOLUT",revolution,comment="XMM revolution number")
  call setAttribute(outccd,"BPT_CODE",BPT_CODE,"BPT_CODE F1529_F1530_F1528")
  call setAttribute(outccd,"SRCPOS",sourceY,"pixel",&
                            "Assumed RAWY source position for fast modes")

  call setAttribute(outccd,"ODFLATEX",odf_id,"ODF identifier")

  call setEventAttribute(outccd,EVATT_ON_OFFSET_COLUMN)
  call setEventAttribute(outccd,EVATT_NEXT_TO_OFFSET_COLUMN)
  call setEventAttribute(outccd,EVATT_IN_SPOILED_FRAME)
  call setEventAttribute(outccd,EVATT_IN_BAD_FRAME)
  call setEventAttribute(outccd,EVATT_OUT_OF_CCD_WINDOW)
  call setEventAttribute(outccd,EVATT_MIP_ASSOCIATED)
  call setEpnRejectionMask(outccd)

! --------------------------
! Create output columns
! --------------------------

  if (.not. hasColumn(outccd,"TIME")) &
     time = addColumn(outccd,"TIME",Real64,"s",dims, comment="Event Time")
  outtimeCol => real64Data(time)

  frame    = addColumn(outccd,"FRAME",Integer32,"counter",dims, &
                       comment="Frame Number")
  outframeCol => int32Data(frame)

  rawX     = addColumn(outccd,"RAWX",Integer16,"pixel",dims, &
                       comment="Event X Position")
  outrawXCol => int16Data(rawX)

  rawY     = addColumn(outccd,"RAWY",Integer16,"pixel",dims, &
                         comment="Event Y Position")   
  outrawYCol => int16Data(rawY)

  pha      = addColumn(outccd,"PHA",Integer16,"channel",dims, &
                       comment="Uncorrected Event Energy")
  call setAttribute(pha, "TLMIN",    0, comment="Minimum for PHA")
  call setAttribute(pha, "TLMAX", 4095, comment="Maximum for PHA")
  outphaCol => int16Data(pha)

  ccdnr    = addColumn(outccd,"CCDNR",Integer8,"",dims, &
                       comment="Overall CCD Number")
  call setAttribute(ccdnr, "TLMIN",    1, comment="Minimum for CCDNR")
  call setAttribute(ccdnr, "TLMAX",   12, comment="Maximum for CCDNR")
  outccdnrCol => int8Data(ccdnr)

  qualflag = addColumn(outccd,EVATT_COLUMN_NAME,Integer32,"",dims, &
                       comment="Event Quality Flag")
  outqualflagCol => int32Data(qualflag)

  if (MIP_dist) then
    mipdistl    = addColumn(outccd,"MIPDISTL",Integer8,"",dims, &
                       comment="distance to the left to next MIP")
    outmipdistposCol => int8Data(mipdistl)

    mipdistr    = addColumn(outccd,"MIPDISTR",Integer8,"",dims, &
                       comment="distance to the right to next MIP")
    outmipdistnegCol => int8Data(mipdistr)

    outhist= addTable(outfile,"MIPDHIST",64)

    miphdist     = addColumn(outhist,"RAWX",Integer8,"pixel",dims, &
                       comment="X_det in CCD")
    outdistCol => int8Data(miphdist)
    outdistCol = 0

    miphistl    = addColumn(outhist,"DHISTL",Integer32,"",dims, &
                       comment="Histogram of possible distances to MIP")
    outhistLCol => int32Data(miphistl)
    outhistLCol = 0

    miphistr     = addColumn(outhist,"DHISTR",Integer32,"",dims, &
                       comment="Histogram of possible distances to MIP")
    outhistRCol => int32Data(miphistr)
    outhistRCol = 0

  endif

! in_events =  clone(stringParameter("eventset"),
!                   "bpxftemp", TEMP, USE_ENVIRONMENT)

! -------------------------------
! Create empty EXPOSURE extension
! -------------------------------
  write(expname,'(a)') "EXPOSURE"

!
! replace `kfrm' by total number of frames 
!
  if (empty_frames) then

!   write(errstr,*) 'EXPOSURE extension: # of frames (=rows) = ',n_of_fram
    write(errstr,*) 'EXPOSURE extension: # of frames (=rows) = ',ntaux0
    call message(AppMsg,Noisy)
!   outexpo = addTable(outfile,expname,n_of_fram,"Time steps and efficiency")
    outexpo = addTable(outfile,expname,ntaux0,"Time steps and efficiency")

  else

    write(errstr,*) 'EXPOSURE extension: # of frames (=rows) = ',kfrm
    call message(AppMsg,Noisy)
    outexpo = addTable(outfile,expname,kfrm,"Time steps and efficiency")

  endif

! call addAttributes(attributable(outexpo),attributable(outccd))

  dims(1) = 1

  exptime  = addColumn(outexpo,"TIME",Real64,"s",dims, &
                       comment="Time of the center of each frame")

  if (also_timedel_col) then
    timedel  = addColumn(outexpo,"TIMEDEL",Real32,"s",dims, &
                       comment="Duration of each frame")
  endif

  fracexp  = addColumn(outexpo,"FRACEXP",Real32,"fraction",dims, &
                       comment="effective time / real time")
  ndsclin1 = addColumn(outexpo,"NDSCLIN1",Integer16,"",dims, &
                       comment="Columns with MIPs: bit array for RAWX  1-16")
  ndsclin2 = addColumn(outexpo,"NDSCLIN2",Integer16,"",dims, &
                       comment="Columns with MIPs: bit array for RAWX 17-32")
  ndsclin3 = addColumn(outexpo,"NDSCLIN3",Integer16,"",dims, &
                       comment="Columns with MIPs: bit array for RAWX 33-48")
  ndsclin4 = addColumn(outexpo,"NDSCLIN4",Integer16,"",dims, &
                       comment="Columns with MIPs: bit array for RAWX 49-64")
  framelim = addColumn(outexpo,"FRAMELIM",Integer32,"",dims, &
                       comment="index of last event in frame")

! ----------------------------------------
! Create empty MIP map (extension of file)
! ----------------------------------------

  if (MIP_ext) then
    imageDim(1) = nx_ccd
    imageDim(2) = ny_ccd
    if (MIP_out == 2) then
      outmip = dataSet(eventmapset,CREATE,USE_ENVIRONMENT)
      call addAttributes(attributable(outmip),attributable(outfile))

      write(eventMapExtension,'(a,i2.2)') 'PEMIPS',CCDno
      mip = addArray(outmip, eventMapExtension, Integer32, imageDim)
      write(eventMapExtension,'(a,i2.2)') 'PEHARD',CCDno
      nomip = addArray(outmip, eventMapExtension, Integer32, imageDim)
      write(eventMapExtension,'(a,i2.2)') 'PESOFT',CCDno
      soft = addArray(outmip, eventMapExtension, Integer32, imageDim)
      write(eventMapExtension,'(a,i2.2)') 'PTOTAL',CCDno
      total = addArray(outmip, eventMapExtension, Integer32, imageDim)
    else
      write(mipname,'(a,i2.2,a)') 'MIPMAP'
      outmip = outfile
      mip = addArray(outmip, mipname, Integer32, imageDim)

      write(nomipname,'(a,i2.2,a)') 'NOMIPMAP'
      outnomip = outfile
      nomip = addArray(outnomip, nomipname, Integer32, imageDim)

      write(softname,'(a,i2.2,a)') 'SOFTMAP'
      outsoft = outfile
      soft = addArray(outsoft, softname, Integer32, imageDim)

      write(totalname,'(a,i2.2,a)') 'TOTALMAP'
      outtotal = outfile
      total = addArray(outtotal, totalname, Integer32, imageDim)
    endif
    mip_arrayP   => int32Array2Data(mip)
    nomip_arrayP => int32Array2Data(nomip)
    soft_arrayP  => int32Array2Data(soft)
    total_arrayP => int32Array2Data(total)
  endif

  call OAL_addCommonAttributes(outfile)

  call message(AppMsg,Noisy, &
                   '... created empty event list and exposure extension ...')

  end subroutine createRawOutput



! =============================================================================
! Add attributes TIMESYS, MJDREF, TIMEUNIT, etc.
! =============================================================================
  subroutine writeTIMEattributes(outset,outtab,pos)

  type(DataSetT),intent(in)   :: outset
  type(TableT)  ,intent(in)   :: outtab
  integer       ,intent(in)   :: pos
  integer                     :: status
  real(kind=double)           :: mjd
  logical, parameter          :: clockapp = .true.

  call sla_cldj(1998,01,01,mjd,status)                     ! 50814
  if (pos == 0) then
    call setAttribute(outset,"TIMESYS","TT", &
                  "XMM time will be TT (Terrestial Time)")
    call setAttribute(outset,"MJDREF",mjd,"d", &
                  "1998-01-01T00:00:00 (TT) expressed in MJD (TT)")
    call setAttribute(outset,"TIMEZERO",0, &
                  comment = "Clock correction (if not zero)")    
    call setAttribute(outset,"TIMEUNIT","s", &
                  "All times in s unless specified otherwise")    
    call setAttribute(outset,"CLOCKAPP",CLOCKAPP,&
                  "Clock correction applied")              ! drifts etc.
    call setAttribute(outset,"TIMEREF","LOCAL", &
                  "Reference location of photon arrival times")
    call setAttribute(outset,"TASSIGN","SATELLITE", &
                  "Location of time assignment")
  else
    call setAttribute(outtab,"TIMESYS","TT", &
                  "XMM time will be TT (Terrestial Time)")
    call setAttribute(outtab,"MJDREF",mjd,"d", &
                  "1998-01-01T00:00:00 (TT) expressed in MJD (TT)")
    call setAttribute(outtab,"TIMEZERO",0, &
                  comment = "Clock correction (if not zero)")    
    call setAttribute(outtab,"TIMEUNIT","s", &
                  "All times in s unless specified otherwise")    
    call setAttribute(outtab,"CLOCKAPP",CLOCKAPP,&
                  "Clock correction applied")              ! drifts etc.
    call setAttribute(outtab,"TIMEREF","LOCAL", &
                  "Reference location of photon arrival times")
    call setAttribute(outtab,"TASSIGN","SATELLITE", &
                  "Location of time assignment")
  endif

  end subroutine writeTIMEattributes




! =====================================
! Convert frame counters to event times
! =====================================
  subroutine frameCountersToTimes(frames,timeTags)

    type(ColumnT)                                :: col
    real(kind=double), dimension(:),   pointer   :: timeTags
    real(kind=double), dimension(:),   pointer   :: obtTags

    integer(kind=Int32), dimension(:), pointer   :: frames
    integer(kind=Int32)                          :: i, j, k, l, m, lastframe
    integer(kind=Int16),dimension(:),allocatable :: helpY

    call message(AppMsg,Noisy,"... entering frameCountersToTimes ...")
!
! j,k,l,m: to avoid internal output overflow in routine "message()"
!

    col = column(outccd, "FRAME", READ)
    frames => int32Data(col)

    if (verbosity(AppMsg) > Noisy) then
      m = min(100,kfrm)
      l = 20
      j = m/l
      k = m-l*j
      call message(AppMsg,Noisy,ddashline)
      write(errstr,*) 'Index check: KFRM, M, L, J, K: ',KFRM, M, L, J, K
      call message(AppMsg,Noisy)
      do i = 1,j
        write(errstr,*) 'Frames: ',frames(l*(i-1):l*i-1)
        call message(AppMsg,Noisy)
      enddo
      write(errstr,*) 'Frames: ',frames(l*j:m-1)
      if (k > 0) call message(AppMsg,Noisy)
    endif

    call message(AppMsg,Verbose,"... entering OAL_frameCountersToObt ...")
    call OAL_frameCounterToObt(frames, obtTags)
    call message(AppMsg,Verbose,"... finished OAL_frameCountersToObt ...")
    if (verbosity(AppMsg) > Noisy) then
      l = 4
      j = m/l
      k = m-l*j
      call message(AppMsg,Noisy,ddashline)
      write(errstr,*) 'Index check: KFRM, M, L, J, K: ',KFRM, M, L, J, K
      call message(AppMsg,Noisy)
      do i = 1,j
        write(errstr,*) 'Frames,obtTags: ', &
              frames(l*(i-1)+1:l*i),obtTags(l*(i-1):l*i-1)
        call message(AppMsg,Noisy)
      enddo
      write(errstr,*) 'Frames,obtTags: ',frames(l*i:m-1),obtTags(l*j:m-1)
      if (k > 0) call message(AppMsg,Noisy)
    endif

    call message(AppMsg,Verbose,"... entering OAL_obtToTimeTag ...")
    call OAL_obtToTimeTag(obtTags, timeTags)
!
!   do l=0,4
!     print*,'OAL_TIME:',l,frames(l),obttags(l),timeTags(l)
!   enddo
!
    call message(AppMsg,Verbose,"... finished OAL_obtToTimeTag ...")
    if (verbosity(AppMsg) > Noisy) then
      l = 2
      j = m/l
      k = m-l*j
      call message(AppMsg,Noisy,ddashline)
      write(errstr,*) 'Index check: KFRM, M, L, J, K: ',KFRM, M, L, J, K
      call message(AppMsg,Noisy)
      do i = 1,j
        write(errstr,*) 'obtTags,TimeTags: ', &
              obtTags(l*(i-1)+1:l*i),timeTags(l*(i-1):l*i-1)
        call message(AppMsg,Noisy)
      enddo
      write(errstr,*) 'obtTags,TimeTags: ', &
            obtTags(l*j:m-1),timeTags(l*j:m-1)
      if (k > 0) call message(AppMsg,Noisy)
    endif

!   call message(AppMsg,Verbose,"... TimeOffset ...")
    TimeOffset = timeTags(0)-obtTags(0)
!   print*,'TimeOffset = ',TimeOffset
!   call message(AppMsg,Verbose,"... entering TimeReset ...")
    call setAttribute(outccd,"TRESET",TimeOffset,"s", &
                     comment="Reset time offset added to PN times")
!   call message(AppMsg,Verbose,"... leaving TimeReset ...")

! fine time correction via CAL or SAS (if any events to be written):

    if (ngrow > 0) then

     cal_correct_finetime = .TRUE.
     if (cal_correct_finetime) then

!     call message(AppMsg,Verbose,"... allocate helpY ...")
      allocate(helpY(0:ngrow-1))
!     call message(AppMsg,Verbose,"... helpY=arrRawY ...")
      helpY(0:ngrow-1) = arrRawY(0:ngrow-1)            ! + 1 is set now earlier

      call message(AppMsg,Verbose,"... entering CAL_correctTime ...")
      if (inmode == IMAGING) then
        call CAL_correctTime(timeTags)
        if (verbosity(AppMsg) > Noisy) then
          l = 3
          j = m/l
          k = m-l*j
          call message(AppMsg,Noisy,ddashline)
          write(errstr,*) 'Index check: KFRM, M, L, J, K: ',KFRM, M, L, J, K
          call message(AppMsg,Noisy)
          do i = 1,j
            write(errstr,*) 'helpY,TimeTags: ', &
                 helpY(l*(i-1):l*i-1),timeTags(l*(i-1):l*i-1)
            call message(AppMsg,Noisy)
          enddo
          write(errstr,*) 'helpY,TimeTags: ', &
                helpY(l*j:m-1),timeTags(l*j:m-1)
          if (k > 0) call message(AppMsg,Noisy)     
        endif
      elseif (inmode == TIMING .or. inmode == BURST) then ! with source position
        call CAL_correctTime(timeTags, helpY, sourceY)    
      endif

      deallocate(helpY)

      call message(AppMsg,Verbose,"... finished CAL_correctTime ...")

    else

      call message(AppMsg,Noisy,"... entering SAS_correctTime ...")

      select case(insubmode)
        case(PRIME_FULL_WINDOW)
          timeTags = timeTags
!
! nothing done yet for FF
!
        case(PRIME_FULL_WINDOW_EXTENDED)
          timeTags = timeTags
!
! nothing done yet for eFF
!
        case(PRIME_LARGE_WINDOW)
          timeTags = timeTags
!
! nothing done yet for LW
!
        case(PRIME_SMALL_WINDOW)
          timeTags = timeTags
!
! nothing done yet for SW
!
!       case(FAST_TIMING)                  ! EPIC-EST-OP-002 Sep.99 Sect.8.3.1
!         timeTags = timeTags - ( 9 - int(helpY/20)*0.5904 +        &
!                  (helpY-int(helpY/20)*20 - sourceY)*0.02952) * 0.001
!
! above formula/document is incorrect: below (hopefully) `correct' formula
!
        case(FAST_TIMING)
          allocate(helpY(0:ngrow-1))
          helpY(0:ngrow-1) = arrRawY(0:ngrow-1)-1 
          timeTags = timeTags + ( (helpY-200) - int(sourceY/10) ) * 29.52d-6
          deallocate(helpY)

        case(FAST_BURST)                   ! EPIC-EST-OP-002 Sep.99 Sect.8.3.2
          allocate(helpY(0:ngrow-1))
          helpY(0:ngrow-1) = arrRawY(0:ngrow-1)-1 
          timeTags = timeTags - ( 4.27 + (helpY - sourceY)*0.00072) * 0.001
          deallocate(helpY)

        case default
          call error('wrongMode','caseDefaultError: fineTime')
      end select

      call message(AppMsg,Noisy,"... leaving SAS_correctTime ...")

    endif

    endif ! ngrow>0 ?

    if (verbosity(AppMsg) > Noisy+1) then
      lastframe = frames(0)
      write(errstr,'(1X, A, I7, 1X, F20.6, 1X, F20.6)') & 
             'Frame-#, obtTag, timeTag: ',frames(0),obtTags(0),timeTags(0)
      call message(AppMsg,Noisy)
      do i=1, numberOfRows(outccd)-1
        if (frames(i) /= lastframe) then
          write(errstr,'(1X,A,I7, 1X, F20.6, 1X, F20.6)') &
             'Frame-#, obtTag, timeTag: ',frames(i),obtTags(i),timeTags(i)
          call message(AppMsg,Noisy)
        endif
        lastframe = frames(i)
      enddo
    endif

    call OAL_releaseMemory(obtTags)
    call release(col)

    call message(AppMsg,Noisy,"... leaving frameCountersToTimes...")

  end subroutine frameCountersToTimes







! ==========================
! Define Frames: event basis 
! ==========================
  subroutine defineFrames

  integer(kind=Int32) :: i, k, jfr

  call message(AppMsg,Noisy,'... entering defineFrames...')

! --------------------------------
! Determine the frame start values
! --------------------------------
! ind() contains the index of the first event in the frame, values: 1..ntrow

  jfr = -1
  k   =  0

  if (ntrow > 0) then   
    allocate(ind(1:ntrow))
    do i = 1 , ntrow
      if (frameCol(i-1) /= jfr) then
        if (k < maxfrm) then
          k = k + 1
          ind(k) = i
        else
          call error('toomany','Too many frames')
        endif
        jfr = frameCol(i-1)
      endif
    enddo
  endif

  nfrm = k

  write(errstr,*) 'Total number of frames:  ',nfrm 
  call message(AppMsg,Noisy)

  end subroutine defineFrames





! =======================
! Search for valid events
! =======================
  subroutine validEvents(irow,nbuf,ixev,iyev,ampl,qual,distp,distn, &
                         goodtime,ngbuf,nmipcol,wrongReadoutOrder, wrongCoord)

  integer(kind=Int32), intent(in)    :: irow, nbuf
  integer(kind=Int16), intent(in)    :: ixev(nbuf), iyev(nbuf), ampl(nbuf)
  integer(kind=Int32), intent(out)   :: qual(nbuf), ngbuf, nmipcol
  integer(kind=Int8),  intent(out)   :: distp(nbuf), distn(nbuf)
  logical(kind=bool) , intent(inout) :: goodtime, wrongReadoutOrder, wrongCoord

  integer(kind=Int16), dimension(:), allocatable :: ixxev, iyyev

  integer(kind=Int32)    :: k, ncol, icol(1:nx_ccd), i, j, nm, kkfrm
  integer(kind=Int16)    :: ix, iy, ix_old, iy_old, x, &
                            pix_code_old, pix_code_new
  logical(kind=bool)     :: lxcol(1:nx_ccd), lycol(1:nx_ccd), lsl, lsingle

  if (loopflag2 == 0) then
    loopflag2 = 1
    call message(AppMsg,Noisy,"... entering validEvents...")
  endif

! Init flag column

  qual = EVATT_CLEAR_ALL_FLAGS

  pix_code_old = 0
  log  = .false.
  nm   = 0
  lsingle = .false.

! initialize the valid columns ( lxcol=true <-> no MIP )

  lxcol(1:nx_ccd) = .true.

  allocate(ixxev(1:nbuf),iyyev(1:nbuf))
  
  ixxev = ixev + 1
  iyyev = iyev + 1

!   if (show_ve) then
!     write(errstr,*) 'IROW=',irow,'NBUF=',NBUF,'GOODTIME=',goodtime,'LOG=',log
!     call message(AppMsg,Sparse)
!     write(errstr,*) 'X=',(IXXEV(J),J=1,MIN(200,NBUF))
!     call message(AppMsg,Sparse)
!     write(errstr,*) 'Y=',(IYYEV(J),J=1,MIN(200,NBUF))
!     call message(AppMsg,Sparse)
!   endif

  do i = 1, nbuf 

    ix = ixxev(i)
    iy = iyyev(i)

    if ((ix < nx0) .or. (ix > nxmax) .or. &
        (iy < ny0) .or. (iy > nymax)) then
      nwrongCoord = nwrongCoord + 1
      log = .true.
!     if (warnFlag == 0) then
        write(errstr,*) 'Error in coordinates: Event ',irow + i - 1,':',ix,iy
        call warning('wrongCoord')
        wrongCoord = .true.
!     endif

      qual(i) = ior(qual(i), EVATT_OUT_OF_CCD_WINDOW)
                                    ! WHOLE FRAME FLAGGING IF EVENT
                                    ! OUTSIDE THE READOUT WINDOW
      cycle
    endif

    if (ampl(i) > MIP_Threshold) then
      nm   = nm + 1
      nMIP = nMIP + 1
      if (verbosity(AppMsg) >= Noisy) then
        if (show_ve) then
          write(errstr,*) "MIP detected in column RAWX =",ix
          call message(AppMsg,Noisy)
        endif
      endif
      if (.not.log) then
        lxcol(ix) = .false.           ! here only column with MIP itself
        if (iy .gt. 1) then
          mip_array(ix,iy) = mip_array(ix,iy)+1
        else
          write(errstr,*) 'Error in MIPcoordinates: MIP ',&
                            irow + i - 1,': ix,iy = ',ix,iy
          call warning('wrongMIPCoord')
        endif
      endif
    endif

    pix_code_new = iy * 100 + ix

    if (pix_code_new <= pix_code_old) then
      nwrongOrder = nwrongOrder + 1
      log = .true.
!     if (warnFlag == 0) then
        write(errstr,*) "Events in wrong order: Event ",&
                         irow + i - 1,": IX,IY =",ix,iy, &
                        "; Previous event: IX,IY =",ix_old, iy_old
        call warning("wrongReadoutOrder")
        wrongReadoutOrder = .true.
!     endif
    endif

    pix_code_old = pix_code_new
    ix_old = ix
    iy_old = iy

    if (.not. goodtime) then
       nwrongTime = nwrongTime + 1
       log = .true.
!      if (warnFlag == 0) then
         write(errstr,*) 'Frames not sorted in time: Event ', irow+i-1 ,':',i
         call warning('wrongTimeOrder')
!      endif
    endif

  enddo   ! end of loop over all events in frame

  if (nm > 0) then
    if (.not. after_mip_frame) then
      after_mip_frame = .true.
    elseif (only_after_mip) then
      after_mip_frame = .false.
    else
      after_mip_frame = .true.
    endif
  else
    if (only_after_mip) then
      if (after_mip_frame) then
        after_mip_frame = .false.
      else
        after_mip_frame = .true.
      endif
    else
      after_mip_frame = .true.
    endif
  endif

!
! determine distance to next MIP event in X direction
!
   do i = 1,nbuf
     ix = ixxev(i)
     if (ix >= 1 .and. ix <= nx_ccd) then
       distp(i) = +99
       distn(i) = +99
       if (nm > 0) then 
         do j = ix,1,-1
           if (.not.lxcol(j)) then
             distp(i) = ix-j
             exit
           endif
         enddo
         do j = ix,64
           if (.not.lxcol(j)) then
             distn(i) = j-ix
             exit
           endif
         enddo
       endif
     else
       distp(i) = +100
       distn(i) = +100
     endif
   enddo

   if (nm .gt. 0) then
     do j = 1,nx_ccd-1
       if (.not.lxcol(j).and.lxcol(j+1)) then
          do i = j+1,nx_ccd
            if (lxcol(i)) then
              dist_2_mip_hist_l(i-j) = dist_2_mip_hist_l(i-j)+1
            else
              exit
            endif
          enddo
       endif
     enddo
     do j = nx_ccd,2,-1
       if (.not.lxcol(j).and.lxcol(j-1)) then
          do i = j-1,1,-1
            if (lxcol(i)) then
              dist_2_mip_hist_r(j-i) = dist_2_mip_hist_r(j-i)+1
            else
              exit
            endif
          enddo  
       endif
     enddo
   endif

!
! collect columns with MIPs
!
   do x = 1, nx_ccd
     if (.not.lxcol(x)) then
       mip_array(x,1) = mip_array(x,1)+1    ! never events in first row
     endif
   enddo

! ----------------------------------------------
! Count invalid events and set whole frame flags
! ----------------------------------------------
   kkfrm = 1
   lsl = .true.  

   if (mipcorr) then

     if (mipnew) then
       if (nm > An_MAXMIP) then
         lxcol(1:nx_ccd) = .false.
         if (verbosity(AppMsg) > Verbose) then
           if (show_ve) then
             write(errstr,*) 'Full MIP Frame: nm > An_MAXMIP: ',nm , An_MAXMIP
             call message(AppMsg,Noisy)
           endif
         endif
       else
         lycol(1:nx_ccd) = lxcol(1:nx_ccd)
         do ix = 1,nx_ccd
           if (.not.lxcol(ix)) then
             do x = -An_EAMIPSEL,+An_EAMIPSEL
               j = ix+x
               if (j >= 1 .and. j <= nx_ccd) lycol(j) = .false.
             enddo
           endif
         enddo
         lxcol(1:nx_ccd) = lycol(1:nx_ccd)
       endif
     endif

! reject the whole frame if there are at least 5 adjacent columns:

     if (mipcols) then

       x = 0
       do while (lsl .and. x < nx_ccd-5)
         x = x + 1
         if (.not.lxcol(x) .and. .not.lxcol(x+1)) then
           if (.not.lxcol(x+3)) then
             if (.not.lxcol(x+4) .and. .not.lxcol(x+5)) lsl = .false.
           endif
         endif
       enddo

       if (.not.lsl) lxcol(1:nx_ccd) = .false.

       lxcolStore(1:nx_ccd) = lxcol(1:nx_ccd)
    
! locate neighbouring columns:
    
       do x = 2, nx_ccd
         if (.not.lxcol(x)) lxcol(x-1) = .false.
       enddo

       do x = nx_ccd-1, 1, -1
         if (.not.lxcol(x)) lxcol(x+1) = .false.
       enddo

     else
       lxcolStore(1:nx_ccd) = lxcol(1:nx_ccd)
     endif

!
! now also neighbouring MIP columns flagged, 
! collect columns to be discarded:
!
     k = 0
     do x = 1, nx_ccd
       if (.not.lxcol(x)) then
         k = k + 1
         icol(k) = x
         ndsclinc = ndsclinc+1
       endif
     enddo
     ncol = k
     nmipcol = ncol

     if (ncol > 0) then
       if (verbosity(AppMsg) > Noisy) then
         write(errstr,*) 'ICOL(1:ncol) = ',icol(1:ncol)
         call message(AppMsg,Noisy) 
       endif
     endif

! flag events in discarded columns:

     do i = 1, nbuf
       ix = ixxev(i)
       iy = iyyev(i)
       lsl = .true.
       if (.not.log) then
         total_array(ix,iy) = total_array(ix,iy)+1
       endif
       k = 0
       do while(lsl .and. k < ncol)
         k = k + 1
         if (ix == icol(k)) then
           qual(i) = ior(qual(i), EVATT_MIP_ASSOCIATED) ! MIP COLUMN REJECTION 
         endif
       enddo
     enddo
   endif

   if (MIP_ext) then
     if (.not.log) then
       do i = 1, nbuf
         ix = ixxev(i)
         iy = iyyev(i)
         if (MIP_discard) then
           if (lxcol(ix)) frame_array(ix,iy) = .true.
         else
           frame_array(ix,iy) = .true.
         endif
       enddo
       do i = 1, nbuf
         ix = ixxev(i)
         iy = iyyev(i)
         if (only_single) then
           lsingle = .not. (frame_array(ix-1,iy) .or. & 
                            frame_array(ix+1,iy) .or. &
                            frame_array(ix,iy-1) .or. &
                            frame_array(ix,iy+1))
         else
           lsingle = .true.
         endif
         if (lxcol(ix) .and. lsingle) then
           if (ampl(i) >= NOMIP_loth .and. ampl(i) <= NOMIP_hith ) then
             if (after_mip_frame) nomip_array(ix,iy) = nomip_array(ix,iy)+1
           endif
           if (ampl(i) >= SOFT_loth .and. ampl(i) <= SOFT_hith ) then
             if (after_mip_frame) soft_array(ix,iy) = soft_array(ix,iy)+1
           endif
         endif
       enddo
       do i = 1, nbuf
         ix = ixxev(i)
         iy = iyyev(i)
         frame_array(ix,iy) = .false.
       enddo
     endif      ! not log
   endif        ! MIP_ext

   deallocate(ixxev,iyyev)

   do i = 1, nbuf
     if (log) then
       qual(i) = ior(qual(i), EVATT_IN_BAD_FRAME)     ! WHOLE FRAME FLAGGING
       qual(i) = ior(qual(i), EVATT_IN_SPOILED_FRAME) ! NOT SORTED IN TIME
     endif
     if (qual(i) /= 0) then
       if (kwrong < maxiwrong) then
         if (kwrong == 0 .or. iwrong(kwrong+1) /= kkfrm) then
           kwrong = kwrong + 1
           iwrong(kwrong) = i + irow - 1
         endif
       else
         kwrong = kwrong + 1
       endif
       kkfrm = kkfrm + 1
     endif
   enddo

! collect global information about good events and frames

   ngbuf = 0
!  if (show_ve) then
!    write(errstr,*) 'LOG=',log,'GOODTIME=',goodtime,'NFRMBAD=',nfrmbad
!    call message(AppMsg,Sparse) 
!  endif
   if (log) then
     nfrmbad = nfrmbad + 1
     goodtime = .false.
   endif
!  if (show_ve) then
!    write(errstr,*) 'LOG=',log,'GOODTIME=',goodtime,'NFRMBAD=',nfrmbad
!    call message(AppMsg,Sparse) 
!  endif
   do i = 1,nbuf
     if (qual(i) <= qualMax) ngbuf = ngbuf+1
!    if (qual(i) == 0) ngbuf = ngbuf+1
   enddo

   if (verbosity(AppMsg) > Noisy) then
     write(errstr,*) 'ngbuf,log,qualMax,nfrmbad = ',ngbuf,log,qualMax,nfrmbad
     call message(AppMsg,Noisy) 
   endif

   end subroutine validEvents






! ========================================================
! Determine CCD-specific GTIs: already done from AUX1 file
! ========================================================
   subroutine determineGTIfromAUX(ngti)

   integer(kind=Int32)                          :: maxgti,il
   integer                                      :: ngti


   call message(AppMsg,Noisy,"... entering determineGTIfromAUX...")

   ngti = ngti_aux
   maxgti = ngti_aux
   allocate(deb(1:maxgti),fin(1:maxgti))
!
! I don't have a good method yet to derive good times for empty input files
!
   if (noGTIfound) then
     ngti = 0
     if (nfrm == 0) then
       call message(AppMsg,Sparse, &
                    'Empty input file, no AUX entries, thus not GTI')
     else
       call warning('EventsNoAUX', &
                    'Non-empty input file, no AUX entries, thus not GTI')
     endif
   endif

   if (noGTIfound) then
     if (.NOT. ALLOCATED(aux1gti_b)) allocate(aux1gti_b(0:0))
     if (.NOT. ALLOCATED(aux1gti_e)) allocate(aux1gti_e(0:0))
     aux1gti_b(0) = 0.0
     aux1gti_e(0) = 0.0
     call message(AppMsg,Noisy,'artificial GTI temporarily created')
   endif

   if (verbosity(AppMsg) >= Verbose) then
     write(errstr,*) 'number of good time intervals  = ',ngti_aux
     call message(AppMsg,Sparse)
     if (ngti > 0) call message(AppMsg,Verbose, &
                ' Index        Start              Stop           Length')
     do il = 1,ngti
       write(errstr,'(i6,2f18.5,f14.5)') &
           il,aux1gti_b(il-1),aux1gti_e(il-1),aux1gti_e(il-1)-aux1gti_b(il-1)
       call message(AppMsg,Verbose)
     enddo
   endif

   ontime = 0
   if (ngti > 0) then
     deb(1:ngti) = aux1gti_b(0:ngti-1) - deltat  ! - t_offset_minus
     fin(1:ngti) = aux1gti_e(0:ngti-1)           ! - t_offset_minus
     do il = 1,ngti
       ontime = ontime + fin(il)-deb(il)
     enddo
   endif
   write(errstr,'(a,f22.5)') 'sum of good time intervals [s]  = ',ontime
   call message(AppMsg,Sparse)
   

   end subroutine determineGTIfromAUX





! =======================================================
! Determine CCD-specific GTIs: obtain from events quality
! =======================================================
   subroutine determineGTIfromEvents(ngti)

   integer(kind=Int32)                          :: maxgti, il, il2 
   integer                                      :: ngti


   call message(AppMsg,Noisy,"... entering determineGTIfromEvents...")

!  maxgti = int(ntrow/2) + 1
   maxgti = int(ngrow/2) + 1
   allocate(deb(1:maxgti),fin(1:maxgti))

   ngti = 0
   il = 0

!
! TIME per event, FrameNr per event, QUAL per event
!
 
!  do while (il < ntrow-1)
   do while (il < ngrow)
     if (verbosity(AppMsg) > Noisy) then
       write(errstr,*) 'Index Check: il,outqualflagCol(il) = ', &
                                     il,outqualflagCol(il)
       call message(AppMsg,Noisy)                    
     endif
!    if (outqualflagCol(il) == 0) then   ! Find first good frame in GTI
     if (outqualflagCol(il) <= qualmax) then   ! Find first good frame in GTI
       ngti = ngti + 1
       if ( ngti > maxgti )  then        ! Too many GTI
         call warning("makeGTI","Maximum number of GTI exceeded")
         exit
       endif
       deb(ngti) = outtimeCol(il) - deltat  ! - t_offset_minus
!      deb(ngti) = outtimeCol(il) - halfdt

! Find last good frame in GTI

!      do il2 = il, ntrow - 2
       do il2 = il, ngrow - 2
         if (iand(outqualflagCol(il2+1), EVATT_IN_BAD_FRAME) > 0 .or. &
             iand(outqualflagCol(il2+1), EVATT_OUT_OF_CCD_WINDOW) > 0 .or. &
             iand(outqualflagCol(il2+1), EVATT_MIP_ASSOCIATED) > 0) then
           if (outframeCol(il2+1) /= outframeCol(il)) exit
         endif
       enddo

       il = il2
       fin(ngti) = outtimeCol(il)           ! + t_offset_plus
!      fin(ngti) = outtimeCol(il) + halfdt
       if (verbosity(AppMsg) > Noisy) then
         write(errstr,*) 'Index Check: ngti,deb(ngti),fin(ngti) = ',  &
                                       ngti,deb(ngti),fin(ngti)
         call message(AppMsg,Noisy)                    
       endif
     endif
     il = il + 1
   enddo

   ontime = 0
   do il = 1,ngti
     ontime = ontime + fin(ngti)-deb(ngti)
   enddo

   end subroutine determineGTIfromEvents





! ============================
! create CCD-specific GTI file
! ============================
   subroutine createGTI(ngti)

   type(TableT)                                 :: outgticcd
   type(ColumnT)                                :: start, stop

   integer                                      :: ngti
   real(kind=double), dimension(:), pointer     :: startCol, stopCol

   character(len=8)                             :: gtiname


   call message(AppMsg,Noisy,"... entering createGTI...")


   outgti = dataSet(stringParameter("gtiset"),CREATE,USE_ENVIRONMENT)
   call addAttributes(attributable(outgti),attributable(outfile)) 

   write(gtiname,'(a,i2.2)') "STDGTI" ,CCDno               ! extension name

   outgticcd = addTable(outgti,gtiname,ngti)

   call addAttributes(attributable(outgticcd),attributable(outccd))

   call setAttribute(outgticcd,"FRMTIME",frmtime,"ms", "Nominal frame time")

   call setAttribute(outgticcd,"HDUCLASS","OGIP", &
                     "File conforms to OGIP/GSFC conventions")
   call setAttribute(outgticcd,"HDUCLAS1","GTI", &
                     "File contains Good Time Intervals")
   call setAttribute(outgticcd,"HDUCLAS2","STANDARD", &
                     "File contains Good Time Intervals")

   dims(1) = 1

   start    = addColumn(outgticcd,"START",Real64,"s",dims, &
                        comment="Start Time of GTI")
   startCol => real64Data(start)
   startCol = 0.

   stop     = addColumn(outgticcd,"STOP", Real64,"s",dims, &
                        comment="End Time of GTI")
   stopCol  => real64Data(stop)
   stopCol  = 0.

   if (ngti > 0) then
     startCol = deb(1:ngti)
     stopCol  = fin(1:ngti)
   endif

   call release(stop)
   call release(start)
   call release(outgticcd)
   call release(outgti)

   if (ALLOCATED(deb)) deallocate(deb)
   if (ALLOCATED(fin)) deallocate(fin)

   end subroutine createGTI




! =============================================================================
! open, read, and close auxiliary file
! =============================================================================
  subroutine readAux(auxfilename,ntaux0,ntaux1,ntaux2,aux1alloc,aux2alloc,& 
     CCDNo,maxgoodreset)

! auxfileName : Name of the input ODF/Auxil file
! ntaux1      : Number of rows in the input file

  character(len=*),   intent(in)   :: auxfilename
  integer(kind=Int32),intent(in )  :: CCDNo, maxgoodreset
  integer(kind=Int32),intent(out)  :: ntaux1,ntaux2,ntaux0
  logical(kind=bool) ,intent(out)  :: aux1alloc,aux2alloc

  real(kind=double), dimension(:), pointer   :: timeTags, obtTags
  real(kind=double)                :: t1,t2,td,tm, &
                                      number_of_frames, time_diff, &
                                      min_diff_frametime, & 
                                      td_aux(1:20), ttmp
  integer(kind=Int32)              :: i,j,k,ftfine32, &
                                      oldtime,newtime,& !,gap(1:10)
                                      oldframe,newframe,fx1052,fx1052half,&
                                      iq0,iq1,iq2,iq4,iq8, fx1052xx, &
                                      itmpmin, itmpmax, itmpval, &
                                      last_good_index, min_diff_frame, n_diff_frame, &
                                      auto_MODE, auto_F1294, auto_CHOP, &
                                      firstGoodIndex, lastGoodIndex, &
                                      nBadIndex, ndaux1
  logical(kind=bool)               :: goodtime,goodframe, &
                                      hasPNAUX2,hasPNAUX1, &
                                      bad_aux1, corruptedAUX1, &
                                      all_good = .false.
  logical(kind=bool), parameter    :: aux1check = .true.
  logical(kind=bool), parameter    :: aux2check = .false., &
                                      negative_obt_debug = .false.
  type(DataSetT)                   :: auxfile
  type(TableT)                     :: aux1table,aux2table


  call message(AppMsg,Noisy,"... entering readAux...")

  time_diff = 0.0d0
  min_diff_frametime = 2.0d+10
  last_good_index = -1
  min_diff_frame = -1
  n_diff_frame = 0
  auto_MODE = -1
  auto_F1294 = -1
  auto_CHOP = -1

  auxfile = dataSet(auxfilename,READ,USE_ENVIRONMENT)
  hasPNAUX1 = hasBlock(auxfile,"PNAUX1")
  if (hasPNAUX1) then
    aux1table = table(auxfile,"PNAUX1")
    ntaux1 = numberOfRows(aux1table)
    write(errstr,*)' PNAUX1 has ',ntaux1,' number of rows'
!   call message(AppMsg,Noisy)           ! number of rows (1..ntaux1)
  else                                   ! should never occur [who knows]
    ntaux1 = 0                           ! no rows
    call warning("noPNAUX1","AUX file does not contain extension PNAUX1.")
  endif

  aux1alloc = .false.
  aux2alloc = .false.
  corruptedAUX1 = .false.

!
! some ODFs have either just one PNAUX1 (no PNAUX2) or two PNAUX1 (second
! apparently containing PNAUX2) extensions: in this case give warning !
! CHECK: can this second PNAUX1 be accessed by block number ?
!
  hasPNAUX2 = hasBlock(auxfile,"PNAUX2")
  if (hasPNAUX2) then
    aux2table = table(auxfile,"PNAUX2")
    ntaux2 = numberOfRows(aux2table)
    write(errstr,*)' PNAUX2 has ',ntaux2,' number of rows'
!   call message(AppMsg,Noisy)           ! number of rows (1..ntaux2)
  else
    ntaux2 = 0                           ! no rows
    write(errstr,*) 'AUX file does not contain extension PNAUX2. ' , &
                    'Exposure values cannot be corrected.'
    call warning("noPNAUX2")
    noPNAUX2 = .true.
    write(errstr,*) 'The ODF is invalid, please contact the originator of this data.'
    call warning('illegalODF')
  endif

!
! access AUX2 columns
!
  j = 0
  if (hasPNAUX2) then
    call message(AppMsg,Noisy,'Accessing AUX2 data...')

    tempInCol = column(aux2table,"CYCLE",READ)
    aux2CycleCol => int32Data(tempInCol)

    tempInCol = column(aux2table,"QUADRANT",READ)
    aux2QuadrantCol => int8Data(tempInCol)

    tempInCol = column(aux2table,"NDISCLIN",READ)
    aux2NdisclinCol => int16Data(tempInCol)

    do i = 0, ntaux2-1
      if (aux2QuadrantCol(i) == Quadrant) then
        j = j+1
      endif
    enddo
  endif

  if (j == 0) then

    write(errstr,*) 'no entries in AUX2 file for CCD = ',CCDNo
    call message(AppMsg,Sparse)
    ntaux2 = 0
    ndsclinq_ave = 0

  else
    aux2alloc = .true.
!
! create scratch data set with frame numbers of NDISCLIN values to be
! able to determine gap times later ??
!
    allocate(aux2Cycle(0:j-1),aux2Ndisclin(0:j-1))
    j = -1
    if (verbosity(AppMsg) >= Noisy) then
      write(errstr,'(a,8x,i12)') ' Total number of rows in AUX2 file:',ntaux2
      call message(AppMsg,Noisy)
    endif
    do i = 0, ntaux2-1
      if (aux2QuadrantCol(i) == Quadrant) then
        j = j+1
        aux2Cycle(j)    = aux2CycleCol(i)
        aux2Ndisclin(j) = aux2NdisclinCol(i)
        if (aux2NdisclinCol(i) > 0) then
          ndsclinq = ndsclinq + aux2NdisclinCol(i)
!         print*,i,aux2NdisclinCol(i)
        endif
      endif
    enddo
    ntaux2 = j+1   ! now this is number of rows that match quadrant number
    ndsclinq_ave = NINT(dble(ndsclinq)/ntaux2)

    if (verbosity(AppMsg) >= Noisy) then
      write(errstr,'(a,1x,i12)') &
           ' Matching CCD number of rows in AUX2 file:',ntaux2
      call message(AppMsg,Noisy)
    endif

    if (ntaux2 < 1) then                  ! should never happen
      write(errstr,*) 'no entries in AUX2 file for CCD = ',CCDNo
      call warning('noAUX2forCCD')
    else
      allocate(aux2gti_b(0:ntaux2-1),aux2gti_e(0:ntaux2-1), &
               aux2Gtime(0:ntaux2-1))
    endif
  endif

  if (hasPNAUX2) then
    call release(aux2table)
!   nullify(aux2CycleCol,aux2QuadrantCol,aux2NdisclinCol)
  endif
!-----------------------------------------------------------------
! access AUX1 columns: do global integrity check
!
  call message(AppMsg,Noisy,'Accessing AUX1 data...')
!
! check whether the quadrant number is "reasonable"  [0..3]
!
  tempInCol = column(aux1table,"QUADRANT",READ)
  aux1QuadrantCol => int8Data(tempInCol)
  itmpmin = 4
  itmpmax = -1
  do i = 0,ntaux1-1
    itmpval = aux1QuadrantCol(i)
    if (itmpval < 0) itmpval = itmpval+256
    if (itmpval > itmpmax) itmpmax = itmpval
    if (itmpval < itmpmin) itmpmin = itmpval
  enddo
  if ((itmpmin < 0 ).or. (itmpmax > 3)) then
    corruptedAUX1 = .true.
    write(errstr,*) 'illegal QUADRANT range = [',itmpmin,',',itmpmax,']'
    call warning('corruptedAUX1')
  endif
!
! check whether the CCD ID is "reasonable"  [0..2]
!
  tempInCol = column(aux1table,"CCDID",READ)
  aux1CcdidCol => int8Data(tempInCol)
  itmpmin = 3
  itmpmax = -1
  do i = 0,ntaux1-1
    itmpval = aux1CcdidCol(i)
    if (itmpval < 0) itmpval = itmpval+256
    if (itmpval > itmpmax) itmpmax = itmpval
    if (itmpval < itmpmin) itmpmin = itmpval
  enddo
  if ((itmpmin < 0 ).or. (itmpmax > 2)) then
    corruptedAUX1 = .true.
    write(errstr,*) 'illegal CCDID range    = [',itmpmin,',',itmpmax,']'
    call warning('corruptedAUX1')
  endif

  allocate(aux1CCDNr(0:ntaux1-1))
  aux1CCDNr = aux1QuadrantCol*3+aux1CcdidCol+1            ! [1..12]
!
! check whether the full second counter is "reasonable"  [0..32767]
!
  tempInCol = column(aux1table,"FTCOARSE",READ)
  aux1FtcoarseCol => int16Data(tempInCol)
  itmpmin = 65536
  itmpmax = -1
  do i = 0,ntaux1-1
    itmpval = aux1FtcoarseCol(i)
    if (itmpval < 0) itmpval = itmpval+65536
    if (itmpval > itmpmax) itmpmax = itmpval
    if (itmpval < itmpmin) itmpmin = itmpval
  enddo
  if ((itmpmin < 0 ).or. (itmpmax > 32767)) then
    corruptedAUX1 = .true.
    write(errstr,*) 'illegal FTCOARSE range = [',itmpmin,',',itmpmax,']'
    call warning('corruptedAUX1')
  endif
!
! check whether the fractional second counter is "reasonable"  [0..65535]
!
  tempInCol = column(aux1table,"FTFINE",READ)
  aux1FtfineCol => uint16Data(tempInCol)
  itmpmin = 65536
  itmpmax = -1
  do i = 0,ntaux1-1
    itmpval = aux1FtfineCol(i)
    if (itmpval < 0) itmpval = itmpval+65536
    if (itmpval > itmpmax) itmpmax = itmpval
    if (itmpval < itmpmin) itmpmin = itmpval
  enddo
  if ((itmpmin < 0 ).or. (itmpmax > 65535)) then
    corruptedAUX1 = .true.
    write(errstr,*) 'illegal FTFINE range   = [',itmpmin,',',itmpmax,']'
    call warning('corruptedAUX1')
  endif
  
  if (corruptedAUX1) then
    call warning('corruptedAUX1',&
        ' processing may fail for this or other CCDs, please check your data.')
  endif

  tempInCol = column(aux1table,"FRAME",READ)
! tempInCol = column(aux1table,"CYCLE",READ)
  aux1FrameCol => int32Data(tempInCol)

  j = 0
  do i = 0, ntaux1-1
    if (aux1CCDNr(i) == CCDNo) then
      j = j+1
    endif
  enddo

  if (j == 0) then

    aux1alloc = .false.
    deallocate(aux1CCDNr)
    write(errstr,*) 'no entries in AUX1 file for CCD = ',CCDNo
    call message(AppMsg,Sparse)
    ntaux1 = 0

  else

    aux1alloc = .true.
    allocate(aux1Frame(0:j-1), aux1Ftcoarse(0:j-1), aux1Ftfine(0:j-1))
    allocate(aux1FtTime(0:j-1), aux1FtQual(0:j-1), aux1GFrame(0:j-1))

    j = -1
    if (verbosity(AppMsg) >= Noisy) then
      write(errstr,'(a,8x,i12)') ' Total number of rows in AUX1 file:',ntaux1
      call message(AppMsg,Sparse)
    endif
    fx1052 = NINT(f1052/20.48d-6)
    fx1052xx = NINT((f1052+256)/20.48d-6)
    fx1052half = NINT(f1052/20.48d-6*0.5d0)
    oldtime   = -1
    oldframe  = -1
    goodtime  = .true.
    goodframe = .true.
    iq0 = 0
    iq1 = 0
    iq2 = 0
    iq4 = 0
    iq8 = 0
    do i = 0, ntaux1-1
      if (aux1CCDNr(i) == CCDNo) then
        j = j+1
        aux1Frame(j)    = aux1FrameCol(i)
        aux1Ftcoarse(j) = aux1FtcoarseCol(i)
        aux1Ftfine(j)   = aux1FtfineCol(i)
        if (aux1Ftfine(j) < 0) then
          ftfine32 = aux1Ftfine(j)+65536
        else
          ftfine32 = aux1Ftfine(j)
        endif
        aux1FtTime(j)   = NINT(aux1Ftcoarse(j)/20.48d-6)+Ftfine32
        newtime  = aux1FtTime(j)
        newframe = aux1Frame(j)
!
! "ordinary" invalid frames:
!
        if (aux1Ftcoarse(j) == 32767) then
          aux1FtQual(j) = 4
          iq4 = iq4+1
          goodtime  = .false.
          goodframe = .false.
!
! frame number decreasing
!
        elseif (newframe < oldframe) then
          aux1FtQual(j) = 8
          iq8 = iq8+1
          goodtime  = .false.
          goodframe = .false.
!
! optimal case: time increasing
!
        elseif (newtime > oldtime) then
          aux1FtQual(j) = 0
          iq0 = iq0+1
          goodtime  = .true.
          goodframe = .true.
!
! time reset after half of F1052 (i.e. > 16200 s)
!
        elseif ((oldtime-newtime) > fx1052half .and. &
!               (oldtime-newtime) < fx1052xx) then
                (oldtime-newtime) < fx1052) then
          aux1FtQual(j) = 1
          iq1 = iq1+1
          if (show_AUX) then
            write(errstr,*) 'i,j,FtTime,Frame,oldtime,oldframe,FtQual= ', &
               i+1,j+1,aux1FtTime(j),aux1Frame(j),oldtime,oldframe,aux1FtQual(j)
            call message(AppMsg,Sparse)
          endif
          goodtime  = .true.
          goodframe = .true.
!
! all the rest, e.g. times > 32400 s, and resets at start of AUX1 file
!
        else
          if (j.gt.maxgoodreset) then
            aux1FtQual(j) = 2
            iq2 = iq2+1
            goodtime  = .false.
            goodframe = .false.
!           print*,i,j,oldtime, newtime,aux1FtQual(j) 
          else
            aux1FtQual(0:j-1) = 2
            iq2 = iq2+1
            goodtime  = .true.
            goodframe = .true.
!           print*,i,j,oldtime, newtime,aux1FtQual(j) 
          endif
        endif

        if (show_AUX) then
          if ((aux1FTCoarse(j) > F1052-10 .or. &
               aux1FTCoarse(j) < 10) .and.     &
              (aux1FTCoarse(j) < 32767 )) then
            write(errstr,*) 'i,j,FtTime,Frame,oldtime,oldframe,FtQual= ', &
               i+1,j+1,aux1FtTime(j),aux1Frame(j),oldtime,oldframe,aux1FtQual(j)
            call message(AppMsg,Sparse)       
          endif
        endif
        if (goodtime)  oldtime  = newtime
        if (goodframe) oldframe = newframe
      endif
    enddo

    if (verbosity(AppMsg) >= Noisy .or. iq8 > 0) then
      call message(AppMsg,Sparse,' AUX1 frame time quality statistics:')
      write(errstr,'(a,t40,i11)') '  Good frames             (FtQual = 0):',iq0
      call message(AppMsg,Sparse)
      write(errstr,'(a,t40,i11)') '  Normal time backjumps   (FtQual = 1):',iq1
      call message(AppMsg,Sparse)
      write(errstr,'(a,t40,i11)') '  Abnormal time backjumps (FtQual = 2):',iq2
      call message(AppMsg,Sparse)
      write(errstr,'(a,t40,i11)') '  Illegal 32767 times     (FtQual = 4):',iq4
      call message(AppMsg,Sparse)
      write(errstr,'(a,t40,i11)') '  Illegal frame order     (FtQual = 8):',iq8
      call message(AppMsg,Sparse)
    endif

    deallocate(aux1CCDNr)
!
! indices of AUX1 arrays start with 0, number of entries = 1..
!
    ntaux1 = j+1                     ! number of CCD entries in AUX1 file (1..)
    if (ntaux1 < 1) then             ! should never happen
      write(errstr,*) 'no entries in AUX1 file for CCD = ',CCDNo
      call warning('noAUX1forCCD')      
    endif

    if (show_AUX) then
      do j = 1,ntaux1-2
        if (aux1FtQual(j) /= 0) then
          call message(AppMsg,Sparse,dashline)
          write(errstr,*) aux1FtQual(j-1),j-1,aux1Frame(j-1), &
                          aux1Ftcoarse(j-1),aux1Ftfine(j-1)
          call message(AppMsg,Sparse)       
          write(errstr,*) aux1FtQual(j),j,aux1Frame(j), &
                          aux1Ftcoarse(j),aux1Ftfine(j)
          call message(AppMsg,Sparse)       
          write(errstr,*) aux1FtQual(j+1),j+1,aux1Frame(j+1), &
                          aux1Ftcoarse(j+1),aux1Ftfine(j+1)
          call message(AppMsg,Sparse)       
          call message(AppMsg,Sparse,dashline)
        endif
      enddo
    endif

    call message(AppMsg,Verbose,"... checking AUX1 frames integrity ...")
    allocate(aux1gti_b(0:ntaux1-1),aux1gti_e(0:ntaux1-1))

    j = -1
    do i = 0, ntaux1-1
!     write(errstr,*) 'i,aux1FtQual(i) = ',i,aux1FtQual(i)
!     call message(AppMsg,Sparse)
      if (aux1FtQual(i) .le. 1) then
        j = j + 1
      endif
    enddo
    ngaux1 = j+1   ! number (1..) of good AUX1 entries (QUAL < 2)

    write(errstr,*) '.. ... number of bad FTCOARSE frames in PNAUX1 ', & 
                    'for this CCD:', ntaux1-ngaux1
    call message(AppMsg,Verbose)

    allocate(aux1Gtime(0:ngaux1-1), & 
             aux1Bindex_s(0:ntaux1-ngaux1), &
             aux1Bindex_e(0:ntaux1-ngaux1))

    j = -1
    k = -1
    bad_aux1 = .false.
    do i = 0, ntaux1-1
      if (aux1FtQual(i) .le. 1) then        ! good frames for OAL
        if (last_good_index .ge. 0 .and. all_good) then
          time_diff = aux1FtTime(i)-aux1FtTime(last_good_index)
          n_diff_frame = n_diff_frame+1
          if ((time_diff > 0) .and. (time_diff < min_diff_frametime)) then
            min_diff_frametime = time_diff
            min_diff_frame = aux1Frame(i)
!           print*,aux1Frame(last_good_index),aux1Frame(i),&
!             aux1FtTime(last_good_index), aux1FtTime(i),time_diff
          endif
        endif
        last_good_index = i
        all_good = .true.
        j = j + 1
        aux1Gframe(j) = aux1Frame(i)
        if (bad_aux1) then
          bad_aux1 = .false.
          aux1Bindex_e(k) = j
        endif
      else                                  ! bad frames for OAL
        all_good = .false.
        if (.not. bad_aux1) then
          bad_aux1 = .true.
          k = k + 1
          aux1Bindex_s(k) = j-1
        else
          if (verbosity(AppMsg) >= Noisy) then
            write(errstr,*) '... ... long bad interval at i,aux1Frame(i),k =', & 
                            i,aux1Frame(i),k
            call message(AppMsg,Sparse)
          endif
        endif
      endif
    enddo
    nfaux1 = k+1        ! number of bad AUX1 intervals
    if (bad_aux1) aux1Bindex_e(k) = -2

    write(errstr,*) '.. ... number of bad PNAUX1 intervals ', & 
                    'for this CCD:', nfaux1
    call message(AppMsg,Verbose)
    if (n_diff_frame > 0) then
      write(errstr,*) '.. ... minimum difference in times of good frames:', &
                NINT(min_diff_frametime), '[units] at FRAME =', &
                min_diff_frame, 'with',n_diff_frame,'analysed intervals'
      call message(AppMsg,Verbose)
    else
    endif

    td_aux(01)  = 3582.2422
    do i = 2,16
      td_aux(i) = td_aux(1)+(i-1)*2048
    enddo
    td_aux(17)  = 2327.375
    td_aux(18)  = 276.9414
    td_aux(19)  = 291.2422
    td_aux(20)  = 212.1367

    do i = 1,20
      if ( ABS(td_aux(i)-1-min_diff_frametime) <= 1 ) then
        auto_CHOP = 0
        if (i > 16) then
          auto_MODE = i-15
          auto_F1294 = 0
        elseif (i == 1) then
          auto_MODE = 0
          auto_F1294 = 0
        else
          auto_MODE = 1
          auto_F1294 = i-1
        endif
        exit
      endif
    enddo
    if (auto_CHOP < 0) then
      do i = 1,20
        do j = 1,127
          ttmp = td_aux(i)*(j+1)
!         print*,i,j,nint(ttmp*100),nint(min_diff_frametime*100)
          if ( ABS(ttmp-min_diff_frametime) <= 1.5 ) then
            auto_CHOP = j
            if (i > 16) then
              auto_MODE = i-15
              auto_F1294 = 0
            elseif (i == 1) then
              auto_MODE = 0
              auto_F1294 = 0
            else
              auto_MODE = 1
              auto_F1294 = i-1
            endif
            exit
          endif
        enddo
      enddo
    endif

    if ((f1294==0) .and. (ccdmode==1)) ccdmode = 0
    if (((ccdmode /= auto_MODE) .or. (F1294 /= auto_F1294) .or.  &
         (An_CHOP /= auto_CHOP)) .and. (auto_chop .ge. 0)) then
      write(errstr,*) 'mode parameters derived from AUX file do not match ',&
         'ODF values: CCDMODE =',auto_MODE,ccdmode,', F1294 =',auto_F1294,&
         f1294,', An_CHOP = ',auto_CHOP,an_CHOP
      call warning('autoModeMismatch')
    endif

    if (auto_CHOP < 0) then
      call message(AppMsg,Sparse,'MODE could not be determined automatically')
    else
      write(errstr,*) 'MODE, F1294, CHOP suggested to be:',&
                       auto_mode,auto_f1294,auto_chop
      call message(AppMsg,Verbose)
      if (automode .AND. (.not.correct_ODF)) then
        ccdmode = auto_MODE
        F1294   = auto_F1294
        An_CHOP = auto_CHOP
      endif
    endif

    if (verbosity(AppMsg) >= Noisy) then
      do k = 0,nfaux1-1
        write(errstr,*) '... ... bad interval indices: k,s,e: ', & 
              k,aux1Bindex_s(k),aux1Bindex_e(k)
        call message(AppMsg,Sparse)
      enddo
    endif

    call message(AppMsg,Verbose,"... ... computing times via OAL ...")
    if (ngaux1 < 3) then
      write(errstr,*) 'Only few good times: NGAUX1 = ',NGAUX1
      call message(AppMsg,Sparse)
    endif
    if (illegal_ODF) then
      write(errstr,*) 'Some SAS packages may now issue fatal errors. ', &
                      'This is an ODF problem and not a SAS problem.'
      call warning('invalidODF')
    endif
    call OAL_frameCounterToObt(aux1Gframe(0:ngaux1-1), obtTags)

!
! -------------- UL -------------->
!
    i = 0
    do while (obtTags(i) < 0 .and. i < ngaux1)
       i = i + 1
    enddo
    if (i >= ngaux1) call error('NoGoodObtValues', 'No good OBT values')
    firstGoodIndex = i

    i = ngaux1-1
    do while (obtTags(i) < 0 .and. i > 0)
       i = i - 1
    enddo
    if (i < 0) call error('NoGoodObtValues', 'No good OBT values')
    lastGoodIndex = i
!
! <-------------- UL --------------
!
    if (show_aux) then
      ndaux1 = min(8604+2,ngaux1)
      write(errstr,*) 'aux1Gframe [',ndaux1,']: ',aux1Gframe(ndaux1-3),aux1Gframe(ndaux1-2),aux1Gframe(ndaux1-1)
      call message(AppMsg,Sparse)
      write(errstr,*) 'obtTags [',ndaux1,']: ',obtTags(ndaux1-3),obtTags(ndaux1-2),obtTags(ndaux1-1)
      call message(AppMsg,Sparse)
    endif

    if (show_aux) then
      if (firstGoodIndex > 0 .or. lastGoodIndex < ngaux1) then
        write(errstr,*) 'First good index =',firstGoodIndex, &
           '  Last good index =', lastGoodIndex , &
           '  Number of NGAUX1 indices =',NGAUX1
        call message(AppMsg,Sparse)
      endif
    endif
!   ngaux1 = lastGoodIndex+1

    nBadIndex = 0
    do i = firstGoodIndex, lastGoodIndex
      if (obtTags(i) < 0) nBadIndex = nBadIndex + 1
    enddo
    if (nBadIndex > 0) then
      write(errstr,*) 'Number of bad OBT values = ',nBadIndex
      call message(AppMsg,Noisy)
    endif
!
    call OAL_obtToTimeTag(obtTags, timeTags)
    if (show_aux) then
      write(errstr,*) 'timeTags [',ndaux1,']: ',timeTags(ndaux1-3),timeTags(ndaux1-2),timeTags(ndaux1-1)
      call message(AppMsg,Sparse)
    endif
    if (insubmode /= FAST_TIMING .AND. insubmode /= FAST_BURST) &
       call CAL_correctTime(timeTags)
!
!   do k=0,4
!     print*,'AUX_TIME:',k,aux1Gframe(k),obttags(k),timeTags(k)
!   enddo
!
    aux1Gtime(0:ngaux1-1) = timeTags(0:ngaux1-1)       ! <----- good AUX1 times
    AttTimeTag = (timeTags(firstGoodIndex)+timeTags(lastGoodIndex))/2.0

    call OAL_releaseMemory(obtTags,timeTags)

    call message(AppMsg,Noisy,"... ... determine average AUX2 DLI rate ...")
    call getNominalFrameTime(insubmode,f1294,fc0,frmtime,rotime,psftime,shftime)
    write(messagestr,'(a,i3,a,i1,a,i4,a,f6.3,a)') &
           ' EPEA setup: mean number of discarded columns in',An_MAXFRC, &
           ' frames in quadrant ',quadrant,': ',ndsclinq_ave, &
           '  [',frmtime*1.d-3*An_MAXFRC,' s]'

    call message(AppMsg,Verbose,"... ... determine AUX1 GTIs ...")
    frmtime_seconds = frmtime*1.d-3
    tm = frmtime_seconds*(An_MAXFRC+0.5)*(An_CHOP+1.5)
    t1 = aux1Gtime(firstGoodIndex)
    pha_offset_time = t1

    good_time_first = t1
    if ( ngaux1 > 0 ) then
      good_time_last  = aux1Gtime(lastGoodIndex)
    else    ! should never occur
      call warning('noNGAUX','No good AUX1 intervals left for analysis')
      good_time_last  = good_time_first
    endif
    number_of_frames = (good_time_last-good_time_first)/frmtime_seconds
    n_of_fram        = NINT(number_of_frames)+1   ! EXPOSURE extension entries
    mfrm             = n_of_fram 

    if (mfrm >= 1) then
!     allocate(arrExpTime(0:mfrm-1),arrExpFrame(0:mfrm-1))
      allocate(arrExpTime(0:mfrm+100),arrExpFrame(0:mfrm+100))
    else
      call message(AppMsg,Sparse,"no frames found for EXPOSURE extension.")        
    endif

    if (verbosity(AppMsg) >= Verbose) then
      call message(AppMsg,Sparse,' ')
      write(errstr,'(4x,a,f16.4)') 'Start time       =',good_time_first
      call message(AppMsg,Sparse)
      write(errstr,'(4x,a,f16.4)') 'End time         =',good_time_last
      call message(AppMsg,Sparse)
      write(errstr,'(4x,a,f16.4)') 'Number of frames =',number_of_frames
      call message(AppMsg,Sparse)
      write(errstr,'(4x,a,f16.6)') 'Frame time [ms]  =',frmtime
      call message(AppMsg,Sparse)
      if (number_of_frames > 0) then
        write(errstr,'(4x,a,f16.6)') 'Best match [ms]  =', & 
            (good_time_last-good_time_first)/nint(number_of_frames)*1.d3
        call message(AppMsg,Sparse)
      else
        call message(AppMsg,Sparse, &
                    'No frames selected. Dataset may contain no good times.')
      endif
    endif

    write(errstr,*) 'FRMTIME = ',frmtime,' An_CHOP = ',An_CHOP, &
                    'An_MAXFRC = ',An_MAXFRC,' TM = ',tm*1.d3
    call message(AppMsg,Noisy)

    j = 0
    aux1gti_b(j) = t1
    do i=1,ngaux1-1
      t2 = aux1Gtime(i)
      td = t2-t1
      if (td > tm) then
!       aux1gti_e(j) = t2-frmtime_seconds
        aux1gti_e(j) = t1
        j = j+1
!       if (show_aux) then
          write(errstr,'(a,f16.4,a,f16.4,a,f10.4,a,f9.2,a)') &
                  'AUX1 gap between ',t1,' and ',t2,' = ',td,' [s] = ', &
                  td/frmtime_seconds,' [frames]'
          call message(AppMsg,Verbose)
!       endif
        aux1gti_b(j) = t2
      else
        aux1gti_e(j) = t2-frmtime_seconds        
      endif
      t1 = t2
    enddo
    ngti_aux = j+1
    if (aux1gti_e(j) < aux1gti_b(j)) then
      call message(AppMsg,Sparse,'Last GTI time may be inaccurate')
      aux1gti_e(j) = aux1gti_b(j)+frmtime_seconds
    endif

    call message(AppMsg,Verbose,' ')

    if (verbosity(AppMsg) >= Noisy) then
      write(errstr,'(a,1x,i12)') &
                           ' Matching CCD number of rows in AUX1 file:',ntaux1
      call message(AppMsg,Noisy) 
    endif

!
! this part shall be re-activated once more information about short-term
! data gaps is available (only gaps > AN_MAXFRC frames can be handled now)
!
!   oldtime = aux1FtTime(0)
!   do i= 1, ntaux1-1
!     if (aux1Ftcoarse(i) .ne. 32767) then
!       do j = 1,10
!         gap(j) = (aux1FtTime(i)-oldtime)/j
!       enddo
!       if (verbosity(AppMsg) > Noisy) then
!         write(errstr,*) 'AUX file time gaps (divided by 1,2,3...)',gap(1:10)
!         call message(AppMsg,Noisy) 
!       endif
!       oldtime = aux1FtTime(i)
!     endif
!   enddo

  endif

  if (hasPNAUX2 .and. NTAUX2 > 0) call pnAUX2times
  if (empty_frames) call fillEmptyFrames(ntaux0)

  if (ALLOCATED(aux2Gtime)) deallocate(aux2Gtime)

  call release(aux1table)
! nullify(aux1QuadrantCol,aux1CcdidCol,aux1FtcoarseCol,& 
!         aux1FtfineCol,aux1FrameCol)

  call release(auxfile)

  if (ntaux1 > 0) then
    noGTIfound = .false.
  else
    if (ntaux2 > 0) noGTIfound = .false.
  endif

  end subroutine readAux
 




! ===========================================================
! get times for AUX2 entries from AUX1 entries
! ===========================================================
  subroutine pnAUX2times

  integer(kind=Int32)            :: i, afrm

  afrm = 0
  call message(AppMsg,Verbose,' ')
  do i = 0,ntaux2-1
    do while ( (afrm < ngaux1) .AND. (aux2Cycle(i) > aux1GFrame(afrm)) )
      afrm = afrm+1
    enddo
    if (afrm >= ngaux1) then
      write(errstr,*) '(PNAUX2cycleException), ', &
                      'AUX2 cycle ',aux2Cycle(i),' (index = ',i, & 
                      ') is larger than any valid AUX1 cycle. ', &
                      'The corresponding AUX1 time is likely to be incorrect.'
      call message(AppMsg,Verbose)
!     call warning("PNAUX2cycleException")
      afrm = ngaux1-1
    endif
    if (aux2Cycle(i) .NE. aux1GFrame(afrm)) then
      if (verbosity(AppMsg) >= Noisy) then
        write(errstr,*) 'I,AFRM = ',i,afrm
        call message(AppMsg,Noisy)
        if (i > 0 .AND. i < ntaux2-2) then
          write(errstr,*) aux2Cycle(i-1),aux2Cycle(i),aux2Cycle(i+1)
          call message(AppMsg,Noisy)
        endif
        if (afrm > 0 .AND. afrm < ntaux1-2) then
          write(errstr,*) aux1GFrame(afrm-1),aux1GFrame(afrm),aux1GFrame(afrm+1)
          call message(AppMsg,Noisy)
        endif
      endif
      if (verbosity(AppMsg) >= Noisy .and. afrm > 0 .AND. afrm < ntaux1-2) then
        write(errstr,*) 'AUX2 cycle ',aux2Cycle(i),' (index = ',i, & 
                      ') does not have an AUX1 entry. ', & 
                      'Closest AUX1 entries (index = ',afrm,'):', &
                      aux1GFrame(afrm-1),aux1GFrame(afrm),aux1GFrame(afrm+1)
        call message(AppMsg,Sparse)
      endif
      cycle
    endif
    aux2Gtime(i) = aux1Gtime(afrm)

    if (i < 0) then
      write(errstr,*) i,afrm,aux1GFrame(afrm),aux1Gtime(afrm)
      call message(AppMsg,Noisy) 
    endif

  enddo

  end subroutine pnAUX2times





! ===========================================================
! fill empty frames arrays: frame number, fractional exposure
! ===========================================================
  subroutine fillEmptyFrames(afrm)

  integer(kind=Int32)            :: i, k, nf, afrm
  real(kind=double)              :: td,t1,t2,fch
  real(kind=double), parameter   :: tiny = 1.d-4

  afrm = 0
  tooManyEmptyFrames = .FALSE.
!
! chopper factor
!
  fch = dble(1+an_chop)*frmtime_seconds
!
! first good time from AUX1 file -> first exposure time
!
  if ( .NOT. ALLOCATED(arrExpTime)) then
    write(errstr,*) 'No frame times could be determined, ',&
                    'array not allocated.'
    call message(AppMsg,Sparse)
  else

    arrExpTime(0)  = aux1Gtime(0)
    arrExpFrame(0) = aux1GFrame(0)
    t1 = aux1Gtime(0)
!
! loop over all good AUX1 entries
!
    do i = 1, ngaux1-1
      t2 = aux1Gtime(i)
      td = t2-t1
!
! nf = number of missing frames (for whatever reason)
!
      nf = NINT(td/frmtime_seconds)-1
      do k = 1,nf
        afrm = afrm + 1
        if (afrm >= mfrm) then
          call message(AppMsg,Sparse,'Too many empty frames to fill in.')
          afrm = afrm-1
          tooManyEmptyFrames = .true.
! print*,"k,nf,i,ngaux1-1,afrm = ",k,nf,i,ngaux1-1,afrm
          exit
        else
          arrExpTime(afrm) = t1+k*fch
          arrExpFrame(afrm) = -1
        endif
      enddo
      if (tooManyEmptyFrames) exit
      afrm = afrm + 1
      arrExpTime(afrm) = t2
      arrExpFrame(afrm) = aux1GFrame(i)
      t1 = t2
    enddo
    if (tooManyEmptyFrames) call message(AppMsg,Verbose, &
      'Probably AUX file is longer than corresponding data'//&
      ' or AUX file contains incorrectly assigned quadrant information.')
!
! now the array arrExpTime(0:afrm) contains all frame times 
! 
    if (verbosity(AppMsg) >= Noisy) then
      call message(AppMsg,Noisy,'k,arrExpTime(k),arrExpFrame(k):') 
      do k = 0, min(10,afrm) ! afrm
        write(errstr,*) k,arrExpTime(k),arrExpFrame(k)
        call message(AppMsg,Noisy) 
      enddo
      do k = afrm-10,afrm
        write(errstr,*) k,arrExpTime(k),arrExpFrame(k)
        call message(AppMsg,Noisy) 
      enddo
    endif

  endif ! ALLLOCATED(arrExpTime)

  write(errstr,*) 'AUX1 good frames from file for quadrant: ',ngaux1
  call message(AppMsg,Verbose)
  write(errstr,*) 'total number of EXPOSUnn frames will be: ',afrm+1
  call message(AppMsg,Verbose)

  end subroutine fillEmptyFrames





! =====================================
! Determine rough exposure/elapsed time
! =====================================
  subroutine getDateElapse(date_obs,date_end,date_sec)

  character(len=*),    intent(in)  :: date_obs,date_end
  real(kind=double),   intent(out) :: date_sec

  integer(kind=Int32):: yy1,mt1,dd1,hh1,mm1,ss1,fs1, &
                        yy2,mt2,dd2,hh2,mm2,ss2,fs2, &
                        yyd,mtd,ddd,hhd,mmd,ssd,fsd, &
                        month(12)
  data        month / 31,28,31,30,31,30,31,31,30,31,30,31/


  read(date_obs,'(i4,5(1x,i2),1x,i3)') yy1,mt1,dd1,hh1,mm1,ss1,fs1
  read(date_end,'(i4,5(1x,i2),1x,i3)') yy2,mt2,dd2,hh2,mm2,ss2,fs2

  yyd = yy2-yy1
  mtd = mt2-mt1
  ddd = dd2-dd1
  hhd = hh2-hh1
  mmd = mm2-mm1
  ssd = ss2-ss1
  fsd = fs2-fs1

! print*,'date_obs = ',yy1,mt1,dd1,hh1,mm1,ss1,fs1
! print*,'date_end = ',yy2,mt2,dd2,hh2,mm2,ss2,fs2
! print*,'date_dif = ',yyd,mtd,ddd,hhd,mmd,ssd,fsd
  
  if (fsd < 0) then
    fsd = fsd + 1000
    ssd = ssd - 1
  endif
  if (ssd < 0) then
    ssd = ssd + 60
    mmd = mmd - 1
  endif
  if (mmd < 0) then
    mmd = mmd + 60
    hhd = hhd - 1
  endif
  if (hhd < 0) then
    hhd = hhd + 24
    ddd = ddd - 1
  endif

!
! valid for years 1997-2099 (no check):
!
  if (ddd < 0) then
    ddd = ddd + month(mt1)
    if (mt1 == 2 .and. mod(yy1,4) == 0) ddd = ddd + 1
    mtd = mtd - 1
  endif
  if (mtd < 0) then
    mtd = mtd + 12
    yyd = yyd - 1
  endif

  if (mtd > 0) then 
    call warning("Date2000", &
            "Quite long exposure: are you really sure you have XMM data ?")
  endif

  if (yyd < 0) then
    call warning("Date2000", &
            "Date_Obs > Date_End: are you really sure you have XMM data ?")
  endif

  if (yyd > 1) then
    call warning("Date2000", &
            "Very long exposure: are you really sure you have XMM data ?")
  endif

  if ((yy1 < 2000) .or. (yy2 < 2000)) then
    call warning("Date2000", &
            "Year is < 2000: are you really sure you have XMM data ?")
  endif

  date_sec = fsd*0.001+ssd+60.0*(mmd+60.0*(hhd+24.0*ddd))

  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,'(1x,a,4(i2.2,a2),a,f8.1,a)') 'Interval = ',ddd,'d ', &
                         hhd,'h ',mmd,'m ',ssd,'s ','= ',date_sec,' [s]'
    call message(AppMsg,Sparse)
  endif

  end subroutine getDateElapse




! ====================================================================
! Determine nominal frame times
! Developer's Note: there will be an update to the EPIC documentation,
!                   this update will be included after publication and
!                   will change the frame times and derived quantities
! Developer's Note: draft received, hard-coded a couple of quantities,
!                   however, these are not the final word
! ====================================================================
  subroutine getNominalFrameTime(insubmode,f1294,fc0,frmtime, &
                                 rotime,psftime,shftime)

  integer         ,    intent(in)  :: insubmode
  integer(kind=int32), intent(in)  :: f1294
  real(kind=double),   intent(out) :: frmtime,shftime,rotime,psftime  ! [ms]
  real(kind=single),   intent(out) :: fc0

  real(kind=double)  :: readout_clock,     & ! time unit of clock (25 MHz) [ms]
                        deltat               ! frame time              [clocks]
  integer(kind=Int32):: readout_nl,        & 
                        readout_fs,        & 
                        readout_ts,        & 
                        readout_ws,        & 
                        readout_wu,        & 
                        readout_t0,        & 
                        c_readout,         & 
                        d_readout,         & 
                        p_readout            

  fc0        =     -1.0
  frmtime    =     -1.0
  readout_nl =    577
  readout_fs =     18
  readout_ts =    739
  readout_t0 =    256
  readout_ws = 262144
  readout_wu =  36202
  readout_clock = 0.04d-3 ! [ms]   25 MHz quartz

  select case (insubmode)
    case(PRIME_FULL_WINDOW) ! 73.36432
      c_readout = 202*readout_nl
      d_readout = 3*readout_nl
      p_readout = 1717555
      deltat = 1834108
      fc0    = 1.d0-d_readout/deltat
    case(PRIME_FULL_WINDOW_EXTENDED) ! 199.1
      c_readout = 202*readout_nl
      d_readout = 3*readout_nl
      p_readout = 1717555+F1294*1048576
      deltat = 1834108+F1294*1048576
      fc0    = 1.d0-d_readout/deltat
    case(PRIME_LARGE_WINDOW) ! 47.66464
      c_readout = 1800
      d_readout = 4*readout_nl+200*readout_fs
      p_readout = 1128604
      deltat = 1191616
      fc0    = 0.94863
!     fc0    = 1.d0-d_readout/deltat
    case(PRIME_SMALL_WINDOW) ! 5.67176
      readout_nl = 576
      c_readout = 1152
      d_readout = 200*readout_fs+4*readout_nl
      p_readout = 99530
      deltat = 141794
      fc0    = 0.7101
!     fc0    = 1.d0-d_readout/deltat
    case(FAST_TIMING) ! 5.96464
      c_readout = 0
      d_readout = 3*readout_ts
      deltat = 149116
      p_readout = deltat-d_readout
      fc0 = 0.99117
!     fc0 = 1.d0-d_readout/deltat
    case(FAST_BURST) ! 4.34456
      c_readout = 0
      d_readout = 200*readout_fs + 177*readout_nl
      deltat = 108614
      p_readout = deltat-d_readout
      fc0 = 0.02983
!     fc0 = 1.d0-d_readout/deltat
    case default
      call error('wrongMode','caseDefaultError: frmtime')
  end select
  frmtime = deltat*readout_clock
  psftime = p_readout*readout_clock
  shftime = 0.0
  rotime  = c_readout*readout_clock

  end subroutine getNominalFrameTime




! =============================
! Display nominal frame times
! =============================
  subroutine printNominalFrameTime(f1294)

  integer(kind=int32),intent(in) :: f1294 
  real(kind=single)  fcx
  real(kind=double)  tf,tr,tp,ts

  call message(AppMsg,Noisy,' ')
  call message(AppMsg,Noisy, &
    ' EPIC PN readout mode: frame time, PSF time, readout time [ms], livetimes')

  call getNominalFrameTime(PRIME_FULL_WINDOW,f1294,fcx,tf,tr,tp,ts)
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'PRIME_FULL_WINDOW           ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  if (f1294 == 0) then
    call getNominalFrameTime(PRIME_FULL_WINDOW_EXTENDED,3,fcx,tf,tr,tp,ts)
  else
    call getNominalFrameTime(PRIME_FULL_WINDOW_EXTENDED,f1294,fcx,tf,tr,tp,ts)
  endif
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'PRIME_FULL_WINDOW_EXTENDED  ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  call getNominalFrameTime(PRIME_LARGE_WINDOW,f1294,fcx,tf,tr,tp,ts)
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'PRIME_LARGE_WINDOW          ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  call getNominalFrameTime(PRIME_SMALL_WINDOW,f1294,fcx,tf,tr,tp,ts)
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'PRIME_SMALL_WINDOW          ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  call getNominalFrameTime(FAST_TIMING,f1294,fcx,tf,tr,tp,ts)
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'FAST_TIMING                 ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  call getNominalFrameTime(FAST_BURST,f1294,fcx,tf,tr,tp,ts)
  write(errstr,'(4x,a,2f11.5,f10.5,f8.5)') &
      'FAST_BURST                  ',tf,tp,tr,fcx
  call message(AppMsg,Noisy)

  call message(AppMsg,Noisy,' ')                 

  end subroutine printNominalFrameTime




! =============================================================================
! add missing column F1052 to main periodic housekeeping file
! =============================================================================
  subroutine verifyPMH(pmhfilename,f1052)

! pmhfileName : Name of the input ODF/PMH1 file

  character(len=*)   ,intent(in)   :: pmhfilename
  integer(kind=Int32),intent(in)   :: F1052

  integer(kind=Int16), dimension(:), pointer :: pmh16iCol
  integer(kind=Int32), dimension(1) :: dims = 1
  integer(kind=Int32)              :: ntpmh
  logical(kind=bool)               :: hasPNPMH1
  character(len=5)                 :: colname
  type(DataSetT)                   :: pmhfile
  type(TableT)                     :: pmh1table
  type(ColumnT)                    :: pmhTimeCol


  call message(AppMsg,Noisy,'... entering verifyPMH...')

  pmhfile = dataSet(pmhfilename,MODIFY,LOW_MEMORY)

  hasPNPMH1 = hasBlock(pmhfile,'PNPMH1')
  if (hasPNPMH1) then
    pmh1table = table(pmhfile,'PNPMH1')
    ntpmh = numberOfRows(pmh1table)
    colname = 'F1052'
    if (.NOT.hasColumn(pmh1table,colname)) then
      pmhTimecol = addColumn(pmh1table,colname,UInteger16, &
                               "sec",dims, comment="CLOCK_WRAP_AROUND")
      pmh16iCol => uint16Data(pmhTimecol)
      pmh16iCol(0:ntpmh-1) = F1052
      call message(AppMsg,Sparse,'... creating F1052...')
    else
      call message(AppMsg,Noisy,'... F1052 already present...')
    endif
    call release(pmh1table)
  endif
  call release(pmhfile)
  call message(AppMsg,Noisy,'... leaving verifyPMH...')

  end subroutine verifyPMH



! =============================================================================
! determine start and end time of offset table calculation
! =============================================================================
  subroutine readPMHoffsetmap(pmh1table,Quadrant,CCDid,istart,idx1,idx2)

! pmh1table : pointer to the input ODF/PMH1 file

! integer(kind=Int16), dimension(:), pointer :: pmh16iCol
  integer(kind=Int8),  dimension(:), pointer :: pmh8iCol
  integer(kind=Int16),intent(in)   :: quadrant, CCDid ! 0..3, 0..2
  integer(kind=Int32),intent(in)   :: istart
  integer(kind=Int32),intent(out)  :: idx1,idx2
  integer(kind=Int32)              :: irow, xmin, xmax, nrow
  logical(kind=bool)               :: ok_hkcol,ok_pmh,ok_ncs,ok_ocs
  character(len=5)                 :: colname,colselect
  character(len=12)                :: OPERATING_MODE, NOISE_CALC, OFFSET_CALC
  character(len=3)                 :: CCD_SWISTAT,c3
  type(TableT),intent(in)          :: pmh1table
! type(ColumnT)                    :: pmhTimeCol
  type(ColumnT)                    :: pmhF1008Col,pmhSelectCol,&
                                      pmhNoiseCalcStatCol,pmhOffsetCalcStatCol

  call message(AppMsg,Noisy,'... entering readPMHoffsetmap...')

  xmin = 0
  xmax = 0

! Ax_CCDn_SWISTAT [1319 1483 1484 1364 1485 1486 1409 1487 1488 1454 1489 1490]
  if (ccdid == 0) then
    write(colselect,'(a1,i4.4)') 'F',1319+45*quadrant
  else
    write(colselect,'(a1,i4.4)') 'F',1482+2*quadrant+ccdid
  endif
  if (hasColumn(pmh1table,colselect)) then
    ok_pmh = .true.
    pmhSelectCol = column(pmh1table,colselect,READ) ! a3 "on" off"
    write(errstr,*) 'Accessing column ',colselect
    call message(AppMsg,Noisy)
  else
    ok_pmh = .false.
    call writeHKerror(colselect,xmin,xmax,0,2)
  endif

! Ax_NOISECALCSTAT [1925 1928 1931 1934]
  write(colname,'(a1,i4.4)') 'F',1925+3*quadrant
  if (hasColumn(pmh1table,colname)) then
    ok_ncs = .true.
    pmhNoiseCalcStatCol = column(pmh1table,colname,READ) ! a12 "selected" "not selected"
    write(errstr,*) 'Accessing column ',colname
    call message(AppMsg,Noisy)
  else
    ok_ncs = .false.
    call writeHKerror(colname,xmin,xmax,0,2)
  endif

! Ax_OFFSETCALCSTAT [1926 1929 1932 1935]
  write(colname,'(a1,i4.4)') 'F',1926+3*quadrant
  if (hasColumn(pmh1table,colname)) then
    ok_ocs = .true.
    pmhOffsetCalcStatCol = column(pmh1table,colname,READ) ! a12 "selected" "not selected"
    write(errstr,*) 'Accessing column ',colname
    call message(AppMsg,Noisy)
  else
    ok_ocs = .false.
    call writeHKerror(colname,xmin,xmax,0,2)
  endif

! D_OPERATING_MODE [1008]
  colname = 'F1008'
  if (hasColumn(pmh1table,colname)) then
    ok_hkcol = .TRUE.
    pmhF1008Col = column(pmh1table,colname,READ) ! a12 "OBSERVATION"
    write(errstr,*) 'Accessing column ',colname
    call message(AppMsg,Noisy)
  else
    ok_hkcol = .false.
    call writeHKerror(colname,xmin,xmax,0,2)
  endif

  idx1 = 0
  idx2 = 0
  nrow = 0
  if (ok_pmh .AND. ok_hkcol) then
    do irow = istart-1, 1, -1
      OPERATING_MODE = trim(stringCell(pmhF1008Col,irow))
      CCD_SWISTAT = trim(stringCell(pmhSelectCol,irow))
      NOISE_CALC = trim(stringCell(pmhNoiseCalcStatCol,irow))
      OFFSET_CALC = trim(stringCell(pmhOffsetCalcStatCol,irow))
      if ( (index(OPERATING_MODE,'OFFSET/NOISE') > 0) .and. &
           (index(OFFSET_CALC(1:8),'SELECTED') > 0) .and. &
           (index(NOISE_CALC,'NOT SELECTED') > 0) .and. &
           (index(CCD_SWISTAT, 'ON') > 0) ) then
        if (idx2 == 0) idx2 = irow
        if (idx2 > 0)  idx1 = irow
      else
        if (idx1 > 0 .and. idx2 > 0) then
          nrow = idx2-idx1+1
          exit
        endif
      endif
    enddo
  endif
! if (nrow > 2) then
!   idx1 = idx1+1
!   idx2 = idx2-1
!   nrow = idx2-idx1+1
! endif

  if (ok_pmh) call release(pmhSelectCol)
  if (ok_hkcol) call release(pmhF1008Col)
  if (nrow > 0) then
!
! Ax_CMOFFS frames for offset map calculation [F1323 1368 1413 1458]
!
    write(colname,'(a2,i3.3)') 'F1',323+quadrant*45
    if (hasColumn(pmh1table,colname)) then
      tempInCol = column(pmh1table,colname,READ) ! i8
      write(errstr,*) 'Accessing column ',colname
      call message(AppMsg,Noisy)
      pmh8iCol => int8data(tempInCol)
      xmin = MINVAL(pmh8iCol)
      xmin = MINVAL(pmh8iCol(idx1:idx2))
      xmax = MAXVAL(pmh8iCol(idx1:idx2))
      call release(tempincol)
      if (xmin == xmax) then
        Frames_OffsetMap = xmin
      else
        Frames_OffsetMap = xmax
        call writeHKerror(colname,xmin,xmax,Frames_OffsetMap,1)
      endif
    else
      call writeHKerror(colname,xmin,xmax,0,2)
    endif

!
! Ax_CPUSTAT median correction [1295 1340 1385 1430]
!
    write(colname,'(a1,i4.4)') 'F',1295+45*quadrant
    if (hasColumn(pmh1table,colname)) then
      pmhSelectCol = column(pmh1table,colname,READ) ! a3 "on" off"
      write(errstr,*) 'Accessing column ',colname
      call message(AppMsg,Noisy)
      Median_OffsetMap = trim(stringCell(pmhSelectCol,idx1))
      do irow = idx1+1,idx2
        c3 = trim(stringCell(pmhSelectCol,irow))
        if (Median_OffsetMap .ne. c3) then
          write(errstr,*) colname,' =  Ax_CPUSTAT is not constant: ', &
               trim(Median_OffsetMap),' and ',trim(c3)
          call warning('notHKconstant')
          exit
        endif
      enddo
      call release(pmhSelectCol)
    else
      call writeHKerror(colname,xmin,xmax,0,2)
    endif
  else
    call warning('HKoffsetmap', &
                 'cannot derive offset map parameters from PMH file')
  endif

  call message(AppMsg,Noisy,'... leaving readPMHoffsetmap...')

  end subroutine readPMHoffsetmap



! =============================================================================
! open, read, and close main periodic housekeeping file
! =============================================================================
  subroutine readPMH(pmhfilename,ntpmh,pmhalloc,Quadrant,CCDid, &
                     tts,tte,toffs,toffe)

! pmhfileName : Name of the input ODF/PMH1 file
! ntpmh       : Number of rows in the input file
!
! further parameters to be read: 
!     C_CE_QUADSEL_n           F1077,1076,1075,1074 ("NOT SELECTED") ! 12A
!     CMS_QUADSELECTEDn        F1293,1292,1291,1290 ("NOT SELECTED") ! 12A

  character(len=*)   ,intent(in)   :: pmhfilename, tts, tte
  character(len=*)   ,intent(out)  :: toffs, toffe
  integer(kind=Int16),intent(in)   :: quadrant, CCDid ! 0..3, 0..2
  integer(kind=Int32),intent(out)  :: ntpmh
  logical(kind=bool) ,intent(out)  :: pmhalloc

  real(kind=single)                :: F1122lim(0:13) = (/ 45.6,51.6,56.6, &
                                      111.6,116.6, 171.6,176.6, 231.6,236.6, &
                                      291.6,296.6, 351.6,356.6, 360.1 /)

  character(len=32)                :: Filter_Lim(0:13) = (/'Unknown  ', &
           'CalClosed','Closed   ','CalThin1 ','Thin1    ', &
           'CalThin2 ','Thin2    ','CalMedium','Medium   ', &
           'CalThick ','Thick    ','CalOpen  ','Open     ','Unknown  ' /)


  real(kind=single),   dimension(:), pointer :: pmh32rCol
  integer(kind=Int8),  dimension(:), pointer :: pmh8iCol
  integer(kind=Int16), dimension(:), pointer :: pmh16iCol
  integer(kind=Int32), dimension(:), pointer :: pmh32iCol
  integer(kind=Int32)              :: irow,istart,iend,xmin,xmax,nrow, &
                                      idx1,idx2

  logical(kind=bool)               :: hasPNPMH1, ok_GAINCAME
  logical(kind=bool)               :: correct_PMH = .true.
  integer(kind=Int32)              :: nchange = 0, ival1 = 0, ival2 = 0
  character(len=5)                 :: colname
  character(len=12)                :: ECACHq_GAINCAME = "HIGH"
  type(DataSetT)                   :: pmhfile
  type(TableT)                     :: pmh1table
  type(ColumnT)                    :: pmhTimeCol, GaincameCol, TempInCol


  call message(AppMsg,Noisy,'... entering readPMH...')

  pmhalloc = .false.
!
! now PMH files are "per observation" rather than "per exposure"
! and therefore these files can become quite large:
! access huge file in LOW_MEMORY mode ($SAS_DIR/include/dal.f90)
! as we only need a few of the (more than 600) columns
!
  pmhfile = dataSet(pmhfilename,READ,LOW_MEMORY)

  hasPNPMH1 = hasBlock(pmhfile,'PNPMH1')
  if (hasPNPMH1) then
    pmh1table = table(pmhfile,'PNPMH1')
    ntpmh = numberOfRows(pmh1table)
    write(errstr,*) 'Number of entries in PMH file:',ntpmh
    call message(AppMsg,Noisy)
  else
    ntpmh = 0
    call warning('noPNPMH1','PHK file does not contain extension PNPMH1.')
  endif

!
! access PMH columns
!
  if (hasPNPMH1) then  
    call message(AppMsg,Noisy,'Accessing PMH data...')

    if (correct_HK) then

! time
      colname = 'TIME '
      if (hasColumn(pmh1table,colname)) then
        pmhTimeCol = column(pmh1table,colname,READ) ! a19 yyyy-mm-ddThh:mm:ss
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        if (stringCell(pmhTimeCol,ntpmh-1) < tte) then
          write(errstr,*) 'The PMH file seems to end prematurely: ', &
                'DATE-END > TIME(PMH)_max: [', &
                   trim(tte),'] [',trim(stringCell(pmhTimeCol,ntpmh-1)), &
                ']   The ODF is invalid, please contact the originator of this data.'
          call warning('illegalODF')
          illegal_ODF = .true.       
        endif
        do irow = 0, ntpmh-1
          if (stringCell(pmhTimeCol,irow) > tts) exit
        enddo
        istart = irow
        do irow = istart+1, ntpmh-1
          if (stringCell(pmhTimeCol,irow) > tte) exit
        enddo
        iend = irow-1
        nrow = iend-istart+1
        write(errstr,*) 'PMH interval to search for HK parameters: ', tts, tte
        call message(AppMsg,Noisy)
        if (istart > ntpmh-1) then
          write(errstr,'(1x,a,a,6x,a)') &
             'PMH interval could not be identified.     Last entry in file:      ', &
              trim(stringCell(pmhTimeCol,ntpmh-1))
          call message(AppMsg,Sparse)
          istart = ntpmh-1
          correct_PMH = .false.
        elseif (iend > ntpmh-1) then
          write(errstr,'(1x,a,a,6x,a)') &
             'PMH interval could not be identified.     Last entry in file:      ', &
              trim(stringCell(pmhTimeCol,ntpmh-1))
          call message(AppMsg,Sparse)
          iend = ntpmh-1
          correct_PMH = .false.
        else
          write(errstr,'(1x,a,a,6x,a)') &
             'PMH interval identified from time tags  : ', &
              trim(stringCell(pmhTimeCol,istart)), &
              trim(stringCell(pmhTimeCol,iend))
          call message(AppMsg,Noisy)
          write(errstr,*) '  this corresponds to indices ', istart, iend
          call message(AppMsg,Noisy)
        endif
        call release(pmhTimeCol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif
    else ! not correct_PMH
      correct_PMH = .false.
    endif

    if (correct_PMH) then
!
! determine offset map parameters: MedianCorrection, NumberOfFrames
!
      call readPMHoffsetmap(pmh1table,Quadrant,CCDid,istart,idx1,idx2)
!
! get start and end time of offset map computation
!
      toffs = trim(stringCell(pmhTimeCol,idx1))
      toffe = trim(stringCell(pmhTimeCol,idx2))
      call release(pmhTimeCol)

      write(errstr,*) 'Offset Interval = ', trim(toffs), ' ', trim(toffe), &
                  '   indices = ',idx1,idx2
      call message(AppMsg,Noisy)

! time reset
      colname = 'F1030'
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! i32
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pmh32iCol => int32data(tempInCol)
        F1030_1 = MINVAL(pmh32iCol)
        F1030_1 = MINVAL(pmh32iCol(istart:iend))
        F1030_2 = MAXVAL(pmh32iCol(istart:iend))
        write(errstr,*) 'F1030 [MIN,MAX] = ',F1030_1,F1030_2
        call message(AppMsg,Noisy)
        ival1 = pmh32iCol(istart)
        do i = istart+1,iend
          ival2 = pmh32iCol(i)
          if (ival2 .ne. ival1) then
            nchange = nchange+1
            write(errstr,*) 'Reset F1030: Time_o  Time_n  Length = ',&
                                   ival1, ival2, ival2-ival1
            call message(AppMsg,Sparse)
            ival1 = ival2
          endif
        enddo
        if (nchange > 1) then
          write(errstr,*) 'Number of F1030 resets:',nchange
          call message(AppMsg,Sparse)
        endif
        if (F1030_1 .ne. F1030_2) then
          if (F1030_2 - F1030_1 > F1052+1) then
            write(errstr,*) &
              'Unusual F1030 values in HK file: large interval found: ',&
              'MIN,MAX,LEN = ',F1030_1,F1030_2,F1030_2-F1030_1
            call message(AppMsg,Sparse)
          elseif (ABS(F1030_2-F1030_1-F1052) > 1) then
            write(errstr,*) &
              'Unusual F1030 values in HK file: unusual interval found: ',&
              'MIN,MAX,LEN = ',F1030_1,F1030_2,F1030_2-F1030_1
            call message(AppMsg,Verbose)
          else
            write(errstr,*) 'Coarse time resets [s]: ',F1030_1,F1030_2
            call message(AppMsg,Verbose)
          endif
        else
          write(errstr,*) 'Coarse time reset [s]: ',F1030_1
          call message(AppMsg,Noisy)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! CountingMode start times: for GTIs ?
      write(colname,'(a4,i1.1)') 'F104',1+2*quadrant+0*ccdid
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! i16 F1041 etc.
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! CountingMode end times: for GTIs ?
      write(colname,'(a4,i1.1)') 'F104',2+2*quadrant+0*ccdid
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! i16 F1042 etc.
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! quadrant wait states [F1294]
      write(colname,'(a5)') 'F1294'
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pmh8iCol => int8data(tempInCol)
        xmin = MINVAL(pmh8iCol)
        xmin = MINVAL(pmh8iCol(istart:iend))
        xmax = MAXVAL(pmh8iCol(istart:iend))
        call release(tempincol)
        if (xmin == xmax) then
          F1294 = xmin
        else
!         F1294 = xmax
     	  call writeHKerror(colname,xmin,xmax,F1294,1)
     	endif
      else
     	call writeHKerror(colname,xmin,xmax,0,2)
      endif

! filter wheel position sensor [F1122]
      write(colname,'(a5)') 'F1122'
      if (hasColumn(pmh1table,colname)) then
  	  	tempInCol = column(pmh1table,colname,READ) ! i8
  	  	write(errstr,*) 'Accessing column ',colname
  	  	call message(AppMsg,Noisy)
  	  	pmh32rCol => real32data(tempInCol)
  	  	xmin = MINVAL(pmh32rCol)
  	  	call getHKstat(pmh32rCol,istart,iend,-0.1,360.1, &
  	  			 F1122MIN,F1122MAX,F1122AVE)
		do i = 13,1,-1
		  if (F1122AVE < F1122LIM(i) .and. F1122AVE .ge. F1122LIM(i-1)) exit
		enddo
		Filter_HK = Filter_Lim(i)
        if (i>0) then
          if (F1122MIN < F1122LIM(i-1) .or. F1122MAX > F1122LIM(i) ) then 
            write(errstr,'(1x,a,f7.3,a,f7.3,a,f7.3,a,f7.3,a)') &
                  'Range of filter wheel position sensor values F1122 [',&
                  F1122MIN,',',F1122MAX,'] exceeds limits [', &
                  F1122LIM(i-1),',',F1122LIM(i),']'
            call warning("notHKconstant")
          endif
        endif
  	  	write(errstr,'(a,f8.3,a,f8.3,a,f8.3,a,a)') &
  	     'H_CE_FWSPOT:  average = ',F1122AVE,' [deg],   range = [', &
  	      F1122MIN,',',F1122MAX,']  ->  FILTER = ',trim(filter_hk)
  	    call message(AppMsg,Sparse)
        if (autofilter) then
          FILTER = Filter_HK
        elseif (FILTER(1:1) == 'U' .or. FILTER(1:1) == 'C') then
          FILTER = Filter_HK
        endif
      else
  	    call writeHKerror(colname,xmin,xmax,0,2)
        Filter_HK = 'UNDEFINED'
      endif

! CCD status [1319 1483 1484 1364 1485 1486 1409 1487 1488 1454 1489 1490]
      if (ccdid == 0) then
        write(colname,'(a2,i3.3)') 'F1',319+45*quadrant
      else
        write(colname,'(a3,i2.2)') 'F14',82+2*quadrant+ccdid
      endif
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! a12 "on" off"
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! ECACHn_GAINCAMEx [1227 1226 1225 1234 1233 1232 1241 1240 1239 1248 1247 1246]
      write(colname,'(a3,i2.2)') 'F12',27+7*quadrant-ccdid
      if (hasColumn(pmh1table,colname)) then
        GaincameCol = column(pmh1table,colname,READ) ! a12
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        ok_GAINCAME = .TRUE.
        ECACHq_GAINCAME = trim(stringCell(GaincameCol,istart))
        do irow = istart+1, iend
          if (trim(stringCell(GaincameCol,irow)) .ne. trim(ECACHq_GAINCAME)) then
            ok_GAINCAME = .FALSE.
            write(errstr,*) colname,' = ECACHq_GAINCAME is not constant: ', &
               trim(ECACHq_GAINCAME),' and ',trim(stringCell(GaincameCol,irow))
            call warning('notHKconstant')
            write(errstr,*) ' ... but at IROW =',irow,'the value of ',&
               colname,' is: ',trim(stringCell(GaincameCol,irow))
            call message(AppMsg,Sparse)
            exit
          endif
        enddo
        if (ok_GAINCAME) ECACHn_GAINCAME = ECACHq_GAINCAME
        if (ECACHq_GAINCAME(1:4) .ne. "HIGH") then
          write(errstr,*) 'CCD',ccdNO,'is not in HIGH gain mode: ',&
                           trim(ECACHq_GAINCAME)
          call warning("CCDlowGain")
	  CCDlowGain = .true.
        endif
        call release(GaincameCol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! quadrant high voltage [1198 1199 1200 1201]
      write(colname,'(a2,i3.3)') 'F1',198+quadrant
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! r32
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pmh32rCol => real32data(tempInCol)
        call getHKstat(pmh32rCol,istart,iend,-200.0,-100.0, & 
                       V_VC_UHV_MIN,V_VC_UHV_MAX,V_VC_UHV_AVE)
        write(errstr,'(a,f8.3,a,f8.3,a,f8.3,a)') &
           'V_VC_UHVx:    average = ',V_VC_UHV_AVE,' [V],     range = [', &
            V_VC_UHV_MIN,',',V_VC_UHV_MAX,']'
        call message(AppMsg,Sparse)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! quadrant DISCLIN counter [1330 1375 1420 1465]
      write(colname,'(a2,i3.3)') 'F1',330+quadrant*45
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! i
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pmh16iCol => uint16data(tempInCol)
        xmin = MINVAL(pmh16iCol)
        call getHKstat_uint16(pmh16iCol,istart,iend,-0.1,32756.0, & 
                       An_DSLINC_MIN,An_DSLINC_MAX,An_DSLINC_AVE)
        write(errstr,'(a,f8.1,a,f8.1,a,f8.1,a)') &
           'Ax_DSLINC:    average = ',An_DSLINC_AVE,' [ ],     range = [', &
            An_DSLINC_MIN,',',An_DSLINC_MAX,']'
        call message(AppMsg,Sparse)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! CCD temperature 
      colname = 'F1128'
      if (hasColumn(pmh1table,colname)) then
        tempInCol = column(pmh1table,colname,READ) ! r32
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pmh32rCol => real32data(tempInCol)
        call getHKstat(pmh32rCol,istart,iend,-125.0,-55.0, & 
                       F1128MIN,F1128MAX,F1128AVE)
        write(errstr,'(a,f8.3,a,f8.3,a,f8.3,a)') &
           'H_CE_TTMPFPF: average = ',F1128AVE,' [deg C], range = [', &
            F1128MIN,',',F1128MAX,']'
        call message(AppMsg,Sparse)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

    else
      call message(AppMsg,Noisy,' assumed to be not valid ODF.')
    endif

    call release(pmh1table)
    call release(pmhfile)
  endif

  end subroutine readPMH




! =============================================================================
! open, read, and close auxiliary periodic housekeeping file
! =============================================================================
  subroutine readPAH(pahfilename,ntpah,pahalloc,Quadrant,CCDid,tts,tte, &
                     toffs,toffe)

! pahfileName : Name of the input ODF/PAH1 file
! ntpmh       : Number of rows in the input file
!
! further parameters to be read: 
!       An_RDNODE                F1539,1639,1739,1839 ("SMALL WINDOW") ! 12A

  character(len=*),   intent(in)   :: pahfilename, tts,tte, toffs,toffe
  integer(kind=Int16),intent(in )  :: quadrant, CCDid ! 0..3, 0..2
  integer(kind=Int32),intent(out)  :: ntpah
  logical(kind=bool) ,intent(out)  :: pahalloc

  integer(kind=Int8),  dimension(:), pointer :: pah8Col
  integer(kind=Int16), dimension(:), pointer :: pah16Col
  real(kind=single),   dimension(:), pointer :: pah32rCol
  integer(kind=Int32)              :: irow,istart,iend,xmin,xmax,nrow, &
                                      idx1,idx2, bpt_f1529, bpt_f1530

  logical(kind=bool)               :: hasPNPAH1, ok_charval
  character(len=5)                 :: colname
  character(len=12)                :: Aq_char12, An_char12
  character(len=12)                :: Aq_SENDMODE, bpt_f1528
  type(DataSetT)                   :: pahfile
  type(TableT)                     :: pah1table
  type(ColumnT)                    :: pahTimeCol, SendmodeCol, CmLinePixSetCol

  call message(AppMsg,Noisy,'... entering readPAH...')

  bpt_f1529 = 0
  bpt_f1530 = 0
  bpt_f1528 = 'NOTDEF'

  pahalloc = .false.
  pahfile = dataSet(pahfilename,READ,LOW_MEMORY)

  hasPNPAH1 = hasBlock(pahfile,'PNPAH1')
  if (hasPNPAH1) then
    pah1table = table(pahfile,'PNPAH1')
    ntpah = numberOfRows(pah1table)
  else
    ntpah = 0
    call warning('noPNPAH1','PHK file does not contain extension PNPAH1.')
  endif

!
! access PAH columns
!
  if (hasPNPAH1) then
    call message(AppMsg,Noisy,'Accessing PAH data...')

    if (correct_HK) then

! time
      colname = 'TIME '
      if (hasColumn(pah1table,colname)) then
        pahTimeCol = column(pah1table,colname,READ) ! a19 yyyy-mm-ddThh:mm:ss
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        if (stringCell(pahTimeCol,ntpah-1) < tte) then
          write(errstr,*) 'The PAH file seems to end prematurely: ', &
                'DATE-END > TIME(PAH)_max: [', &
                   trim(tte),'] [',trim(stringCell(pahTimeCol,ntpah-1)), &
                ']   The ODF is invalid and cannot be processed properly.', &
                ' Please contact the originator of this data.'
          call warning('illegalODF')
          illegal_ODF = .true.       
        endif
        do irow = 0, ntpah-1
          if (stringCell(pahTimeCol,irow) > tts) exit
        enddo
        istart = irow
        do irow = istart+1, ntpah-1
          if (stringCell(pahTimeCol,irow) > tte) exit
        enddo
        iend = irow-1
        nrow = iend-istart+1
        if (nrow > 2) then
          iend = iend-1
          istart = istart+1
          nrow = iend-istart+1
        endif
        write(errstr,*) 'PAH interval to search for HK parameters: ', tts, tte
        call message(AppMsg,Noisy)
        write(errstr,'(1x,a,a,6x,a)') &
             'PAH interval identified from time tags  : ', &
              trim(stringCell(pahTimeCol,istart)) , &
              trim(stringCell(pahTimeCol,iend))
        call message(AppMsg,Noisy)
        write(errstr,*) '  this corresponds to indices ', istart, iend
        call message(AppMsg,Noisy)
!
! offset map calculation times: get PAH indices
!
        do irow = 0, istart
          if (stringCell(pahTimeCol,irow) > toffs) exit
        enddo
        idx1 = irow
        do irow = idx1+1, istart
          if (stringCell(pahTimeCol,irow) > toffe) exit
        enddo
        idx2 = irow-1
        if (idx2-idx1 > 2) then
          idx1 = idx1+1
          idx2 = idx2-1
        endif
!
!    print*,'Interval = ', trim(stringCell(pahTimeCol,idx1)), ' ', &
!         trim(stringCell(pahTimeCol,idx2)), '   indices = ',idx1,idx2
!        
        call release(pahTimeCol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif
      
! An_MAXFRC = 20
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,38
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin == xmax) then
          An_MAXFRC = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_MAXFRC,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_CMCORR = 512
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,25
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i16
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah16Col => uint16data(tempInCol)
        xmin = MINVAL(pah16Col)
        xmin = MINVAL(pah16Col(istart:iend))
        xmax = MAXVAL(pah16Col(istart:iend))
        if (xmin == xmax) then
          An_CMCORR = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_CMCORR,1)
        endif
        xmin = MINVAL(pah16Col(idx1:idx2))
        xmax = MAXVAL(pah16Col(idx1:idx2))
        if (xmin == xmax) then
          OT_CMCORR = xmin
        else
          OT_CMCORR = xmax
          call writeHKerror(colname,xmin,xmax,OT_CMCORR,1)
          write (errstr,*) 'PAH OT interval indices: ',idx1,idx2,&
                         '  PAH OT times: ', trim(stringCell(pahTimeCol,idx1)),&
                         ' ',trim(stringCell(pahTimeCol,idx2))
          call message(AppMsg,Sparse)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_CMLOTm = 532
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,15+CCDid
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i16
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah16Col => uint16data(tempInCol)
        xmin = MINVAL(pah16Col)
        xmin = MINVAL(pah16Col(istart:iend))
        xmax = MAXVAL(pah16Col(istart:iend))
        if (xmin == xmax) then
          An_CMLOTH = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_CMLOTH,1)
        endif
        xmin = MINVAL(pah16Col(idx1:idx2))
        xmax = MAXVAL(pah16Col(idx1:idx2))
        if (xmin == xmax) then
          OT_CMLOTH = xmin
        else
          OT_CMLOTH = xmin
          call writeHKerror(colname,xmin,xmax,OT_CMLOTH,1)
          write (errstr,*) 'PAH OT interval indices: ',idx1,idx2,&
                         '  PAH OT times: ', trim(stringCell(pahTimeCol,idx1)),&
                         ' ',trim(stringCell(pahTimeCol,idx2))
          call message(AppMsg,Sparse)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_MIP = 3512
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,26
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i16
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah16Col => uint16data(tempInCol)
        xmin = MINVAL(pah16Col)
        xmin = MINVAL(pah16Col(istart:iend))
        xmax = MAXVAL(pah16Col(istart:iend))
        if (xmin == xmax) then
          An_MIP = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_MIP,1)
        endif
        xmin = MINVAL(pah16Col(idx1:idx2))
        xmax = MAXVAL(pah16Col(idx1:idx2))
        if (xmin == xmax) then
          OT_MIP = xmin
        else
          OT_MIP = xmin
          call writeHKerror(colname,xmin,xmax,OT_MIP,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_MAXMIP = 63
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,27
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin < 0) xmin = xmin+256
        if (xmax < 0) xmax = xmax+256
        if (xmin == xmax) then
          An_MAXMIP = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_MAXMIP,1)
        endif
        xmin = MINVAL(pah8Col(idx1:idx2))
        xmax = MAXVAL(pah8Col(idx1:idx2))
        if (xmin == xmax) then
          OT_MAXMIP = xmin
        else
          OT_MAXMIP = xmin
          call writeHKerror(colname,xmin,xmax,OT_MAXMIP,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_CHOP = 0..25
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,34
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin == xmax) then
          An_CHOP = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_CHOP,1)
        endif
        xmin = MINVAL(pah8Col(idx1:idx2))
        xmax = MAXVAL(pah8Col(idx1:idx2))
        if (xmin == xmax) then
          OT_CHOP = xmin
        else
          OT_CHOP = xmax
!         call writeHKerror(colname,xmin,xmax,OT_CHOP,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_EAMIPSEL = 1..2
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,36
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin == xmax) then
          An_EAMIPSEL = xmin
        else
          call writeHKerror(colname,xmin,xmax,An_EAMIPSEL,1)
        endif
        xmin = MINVAL(pah8Col(idx1:idx2))
        xmax = MAXVAL(pah8Col(idx1:idx2))
        if (xmin == xmax) then
          OT_EAMIPSEL = xmin
        else
          OT_EAMIPSEL = xmax
!         call writeHKerror(colname,xmin,xmax,OT_EAMIPSEL,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! An_SENDMODE
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,35
      if (hasColumn(pah1table,colname)) then
        SendmodeCol = column(pah1table,colname,READ) ! a12
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        ok_charval = .TRUE.
        Aq_SENDMODE = trim(stringCell(SendmodeCol,istart))
        do irow = istart+1, iend
          if (trim(stringCell(SendmodeCol,irow)) .ne. trim(Aq_SENDMODE)) then
            ok_charval = .FALSE.
            write(errstr,*) colname,' = An_SENDMODE is not constant: ', &
               Aq_SENDMODE,' and ',stringCell(SendmodeCol,irow)
            call warning('notHKconstant')
            exit
          endif
        enddo
        if (ok_charval) An_SENDMODE = Aq_SENDMODE
        ok_charval = .TRUE.
        Aq_SENDMODE = trim(stringCell(SendmodeCol,idx1))
        do irow = idx1+1, idx2
          if (trim(stringCell(SendmodeCol,irow)) .ne. trim(Aq_SENDMODE)) then
            ok_charval = .FALSE.
            write(errstr,*) colname,' = OT_SENDMODE is not constant: ', &
               Aq_SENDMODE,' and ',stringCell(SendmodeCol,irow)
            call warning('notHKconstant')
            exit
          endif
        enddo
        if (ok_charval) OT_SENDMODE = Aq_SENDMODE
        call release(SendmodeCol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! quadrant temperature [1576 1676 1776 1876]
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,76
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! r32
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah32rCol => real32data(tempInCol)
        call getHKstat(pah32rCol,istart,iend,0.0,40.0, & 
                       E_Cx_TEMPQB1_MIN,E_Cx_TEMPQB1_MAX,E_Cx_TEMPQB1_AVE)
        write(errstr,'(a,f8.3,a,f8.3,a,f8.3,a)') &
           'E_Cx_TEMPQB1: average = ',E_Cx_TEMPQB1_AVE,' [deg C], range = [', &
            E_Cx_TEMPQB1_MIN,',',E_Cx_TEMPQB1_MAX,']'
        call message(AppMsg,Sparse)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! guard ring anode curr. [1571 1572 1573 1671 1672 1673 1771 1772 1773 1871 1872 1873]
      write(colname,'(a1,i2.2,i2.2)') 'F',15+quadrant,71+CCDid
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! r32
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah32rCol => real32data(tempInCol)
        call getHKstat(pah32rCol,istart,iend,0.1,10.0, & 
                       E_Cx_IGRAx_MIN,E_Cx_IGRAx_MAX,E_Cx_IGRAx_AVE)
        write(errstr,'(a,f8.3,a,f8.3,a,f8.3,a)') &
           'E_Cx_IGRAx:   average = ',E_Cx_IGRAx_AVE,' [mu A],  range = [', &
            E_Cx_IGRAx_MIN,',',E_Cx_IGRAx_MAX,']'
        call message(AppMsg,Sparse)
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

! BPT coding F1529 F1530 F1528
      write(colname,'(a1,i4.4)') 'F',1529
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin < 0) xmin = xmin+256
        if (xmax < 0) xmax = xmax+256
        if (xmin == xmax) then
          BPT_F1529 = xmin
        else
          call writeHKerror(colname,xmin,xmax,BPT_F1529,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

      write(colname,'(a1,i4.4)') 'F',1530
      if (hasColumn(pah1table,colname)) then
        tempInCol = column(pah1table,colname,READ) ! i8
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        pah8Col => int8data(tempInCol)
        xmin = MINVAL(pah8Col)
        xmin = MINVAL(pah8Col(istart:iend))
        xmax = MAXVAL(pah8Col(istart:iend))
        if (xmin < 0) xmin = xmin+256
        if (xmax < 0) xmax = xmax+256
        if (xmin == xmax) then
          BPT_F1530 = xmin
        else
          call writeHKerror(colname,xmin,xmax,BPT_F1530,1)
        endif
        call release(tempincol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

      write(colname,'(a1,i4.4)') 'F',1528
      Aq_char12 = BPT_F1528
      if (hasColumn(pah1table,colname)) then
        CmLinePixSetCol = column(pah1table,colname,READ) ! a12
        write(errstr,*) 'Accessing column ',colname
        call message(AppMsg,Noisy)
        ok_charval = .TRUE.
        Aq_char12 = trim(stringCell(CmLinePixSetCol,istart))
        do irow = istart+1, iend
          if (trim(stringCell(CmLinePixSetCol,irow)) .ne. trim(Aq_char12)) then
            ok_charval = .FALSE.
            write(errstr,*) colname,' = An_char12 is not constant: ', &
               Aq_char12,' and ',stringCell(CmLinePixSetCol,irow)
            call warning('notHKconstant')
            exit
          endif
        enddo
        if (ok_charval) An_char12 = Aq_char12
        call release(CmLinePixSetCol)
      else
        call writeHKerror(colname,xmin,xmax,0,2)
      endif

      write(BPT_CODE,'(i1.1,a1,i3.3,a1,a)') BPT_F1529,'_',BPT_F1530,'_', &
         trim(Aq_char12)
      write(errstr,*) 'BadPixelTable code = ',BPT_CODE
      call message(AppMsg,Noisy)

      if (show_pah) then
        call message(AppMsg,Sparse,' ')
        write(errstr,*) 'EPEA setup ',trim(sciencefilename), &
                                  ': offset =',An_CMCORR, &
                         ', lower threshold =',An_CMLOTH, &
                          ', maximum frames =',An_MAXFRC, &
                                 ', chopper =',An_CHOP
        call message(AppMsg,Sparse)
        write(errstr,*) 'EPEA setup: MIP correction = ',trim(An_SENDMODE), &
                              ' , threshold =',An_MIP, &
                            ', maximum MIPs =',An_MAXMIP, &
                    ', neighbouring columns =',An_EAMIPSEL
        call message(AppMsg,Sparse)
        call message(AppMsg,Sparse,trim(messagestr))
      endif

    else
      call message(AppMsg,Noisy,' assumed to be not valid ODF.')
    endif

    call release(pah1table)
    call release(pahfile)
    call message(AppMsg,Verbose,' ')
  endif

  end subroutine readPAH




! =============================================================================
! get statistics for HK parameter
! =============================================================================
  subroutine getHKstat(hk32rCol,istart,iend,min32mask,max32mask, &
                       min32,max32,ave32)

  integer(kind=Int32),intent(in)  :: istart,iend
  real(kind=single)  ,intent(in)  :: min32mask, max32mask
  real(kind=single)  ,intent(out) :: min32, max32, ave32
  real(kind=single), dimension(:) ,intent(in) :: hk32rCol
  integer(kind=Int32)             :: nrow

  nrow = COUNT(MASK = (hk32rCol(istart:iend) > min32mask &
                 .and. hk32rCol(istart:iend) < max32mask ))
  if (nrow > 0) then
    min32 = MINVAL(hk32rCol(istart:iend), &
             MASK = (hk32rCol(istart:iend) > min32mask .and. hk32rCol(istart:iend) < max32mask))
    max32 = MAXVAL(hk32rCol(istart:iend), &
             MASK = (hk32rCol(istart:iend) > min32mask .and. hk32rCol(istart:iend) < max32mask))
    if (min32 .eq. max32) then
      ave32 = max32
    else
      ave32 = SUM(hk32rCol(istart:iend), &
             MASK = (hk32rCol(istart:iend) > min32mask .and. hk32rCol(istart:iend) < max32mask))/nrow
    endif
  else
    nrow = iend-istart+1
    min32 = MINVAL(hk32rCol(istart:iend))
    max32 = MAXVAL(hk32rCol(istart:iend))
    ave32 = SUM(hk32rCol(istart:iend))/nrow
    write(errstr,'(1x,a,f10.3,a,f10.3,a)') &
                 'HK limits [',min32mask,',',max32mask,  &
                 '] lead to empty range. Will use all entries in time interval.'
    call warning('emptyHKrange')
  endif

  end subroutine getHKstat

! =============================================================================
! get statistics for HK parameter
! =============================================================================
  subroutine getHKstat_uint16(hk16uintCol,istart,iend,min32mask,max32mask, &
                       min32,max32,ave32)

  integer(kind=Int32),intent(in)  :: istart,iend
  real(kind=single)  ,intent(in)  :: min32mask, max32mask
  real(kind=single)  ,intent(out) :: min32, max32, ave32
  integer(kind=int16), dimension(:) ,intent(in) :: hk16uintCol
  integer(kind=Int32)             :: nrow, irow, hkval, nval

  nrow = COUNT(MASK = (hk16uintCol(istart:iend) > min32mask &
                 .and. hk16uintCol(istart:iend) < max32mask ))
  if (nrow > 0) then
    min32 = MINVAL(hk16uintCol(istart:iend), &
             MASK = (hk16uintCol(istart:iend) > min32mask .and. hk16uintCol(istart:iend) < max32mask))
    max32 = MAXVAL(hk16uintCol(istart:iend), &
             MASK = (hk16uintCol(istart:iend) > min32mask .and. hk16uintCol(istart:iend) < max32mask))
    if (min32 .eq. max32) then
      ave32 = max32
    else
!     ave32 = SUM(hk16uintCol(istart:iend), &
!            MASK = (hk16uintCol(istart:iend) > min32mask .and. hk16uintCol(istart:iend) < max32mask))/nrow
      ave32 = 0.0
      nval = 0
      do irow = istart,iend
        hkval = hk16uintCol(irow)
        if (hkval>min32mask .and. hkval<max32mask) then
          ave32 = ave32+hkval
          nval = nval+1
        endif
      enddo
      ave32 = ave32/nval
    endif
  else
    nrow = iend-istart+1
    min32 = MINVAL(hk16uintCol(istart:iend))
    max32 = MAXVAL(hk16uintCol(istart:iend))
    ave32 = SUM(hk16uintCol(istart:iend))/nrow
    write(errstr,'(1x,a,f10.3,a,f10.3,a)') &
                 'HK limits [',min32mask,',',max32mask,  &
                 '] lead to empty range. Will use all entries in time interval.'
    call warning('emptyHKrange')
  endif

  end subroutine getHKstat_uint16




! =============================================================================
! write HK warnings
! =============================================================================
  subroutine writeHKerror(colname,xmin,xmax,default,flag)

  character(len=*)    ,intent(in) :: colname
  integer(kind=Int32) ,intent(in) :: xmin,xmax,default,flag

  if (flag == 1) then
    write(errstr,*) 'HK parameter ',colname,' is not constant, ', &
         'range = [',xmin,',',xmax,']. Use default: ', default
    call warning('notHKconstant')
  elseif (flag == 2) then
    write(errstr,*) 'Missing column ',colname
    call warning('misHKcolumn')
  endif

  end subroutine writeHKerror




! =============================================================================
! write HK parameters to output file (to be used in later tasks)
! =============================================================================
  subroutine writeHKattributes


  call setAttribute(outccd,"HK_TMP_B","------------------",&
                           "- Begin of selected list of HK parameters -----")

  call setAttribute(outccd,"F1030",F1030,"s","Coarse time reset")

  call setAttribute(outccd,"F1052",F1052,"s","Time wrap reset")

  if (insubmode == PRIME_FULL_WINDOW_EXTENDED) then
    call setAttribute(outccd,"F1294",F1294,"","Quadrant wait states")
  else
    call setAttribute(outccd,"F1294",0,"","Quadrant wait states")
  endif

  call setAttribute(outccd,"AX_FC0",FC0,"", &
                           "mode dependent constant livetime")

  call setAttribute(outccd,"AX_FC3",FC3,"", &
                           "chopper dependent constant livetime")

  call setAttribute(outccd,"AN_MAXFR",AN_MAXFRC,"", &
                           "An_MAXFRC F1538 F1638 F1738 F1838")

  call setAttribute(outccd,"AN_CHOP",AN_CHOP,"", &
                           "An_CHOP   F1534 F1634 F1734 F1834")

  call setAttribute(outccd,"AN_CMCOR",AN_CMCORR,"", &
                           "An_CMCORR F1525 F1625 F1725 F1825")

  call setAttribute(outccd,"AN_CMLOT",AN_CMLOTH,"", &
                           "An_CMLOTH F1515 F1516 F1517 F1615 F1616 F1617..")
                                    ! F1715 F1716 F1717 F1815 F1816 F1817")

  call setAttribute(outccd,"AN_MAXMI",AN_MAXMIP,"", &
                           "An_MAXMIP F1527 F1627 F1727 F1827")

  call setAttribute(outccd,"AN_MIP",AN_MIP,"", &
                           "An_MIP    F1526 F1626 F1726 F1826")

  call setAttribute(outccd,"AN_EAMIP",AN_EAMIPSEL,"", &
                           "An_EAMIPSEL F1536 F1636 F1736 F1836")

  call setAttribute(outccd,"AN_SENDM",AN_SENDMODE,&
                           "An_SENDMODE F1535 F1635 F1735 F1835")

  call setAttribute(outccd,"AN_DSLIN",AN_DSLINC_AVE,"", &
                           "An_DSLINC F1330 F1375 F1420 F1465")

  call setAttribute(outccd,"AX_QCCDS",AX_QUAD_CCD_SEL,"", &
                           "An_CCDSWISTAT + QUADSEL F1077 F1076 F1075 F1074")

  call setAttribute(outccd,"GAINCAME",ECACHn_GAINCAME,&
                           "ECACHn_GAINCAME F1227 F1226 F1225 F1234 F1233..")

  call setAttribute(outccd,"F1122MIN",F1122MIN,"deg", &
                           "H_CE_FWSPOS_MIN F1122")

  call setAttribute(outccd,"F1122MAX",F1122MAX,"deg", &
                           "H_CE_FWSPOS_MAX F1122")

  call setAttribute(outccd,"F1122AVE",F1122AVE,"deg", &
                           "H_CE_FWSPOS_AVE F1122")

  call setAttribute(outccd,"F1128MIN",F1128MIN,"deg C", &
                           "H_CE_TTMPFPF_MIN F1128")

  call setAttribute(outccd,"F1128MAX",F1128MAX,"deg C", &
                           "H_CE_TTMPFPF_MAX F1128")

  call setAttribute(outccd,"F1128AVE",F1128AVE,"deg C", &
                           "H_CE_TTMPFPF_AVE F1128")

  call setAttribute(outccd,"TQB1_MIN",E_Cx_TEMPQB1_MIN,"deg C", &
                           "E_Cx_TEMPQB1_MIN F1576 F1676 F1776 F1876")

  call setAttribute(outccd,"TQB1_MAX",E_Cx_TEMPQB1_MAX,"deg C", &
                           "E_Cx_TEMPQB1_MAX F1576 F1676 F1776 F1876")

  call setAttribute(outccd,"TQB1_AVE",E_Cx_TEMPQB1_AVE,"deg C", &
                           "E_Cx_TEMPQB1_AVE F1576 F1676 F1776 F1876")

  call setAttribute(outccd,"IGRA_MIN",E_Cx_IGRAx_MIN,"micro A", &
                           "E_Cx_IGRAy_MIN F1571 F1572 F1573 F1671 F1672.. ")

  call setAttribute(outccd,"IGRA_MAX",E_Cx_IGRAx_MAX,"micro A", &
                           "E_Cx_IGRAy_MAX F1571 F1572 F1573 F1671 F1672.. ")

  call setAttribute(outccd,"IGRA_AVE",E_Cx_IGRAx_AVE,"micro A", &
                           "E_Cx_IGRAy_AVE F1571 F1572 F1573 F1671 F1672.. ")

  call setAttribute(outccd,"UHV_MIN",V_VC_UHV_MIN,"V", &
                           "V_VC_UHV_MIN F1198 F1199 F1200 F1201")

  call setAttribute(outccd,"UHV_MAX",V_VC_UHV_MAX,"V", &
                           "V_VC_UHV_MAX F1198 F1199 F1200 F1201")

  call setAttribute(outccd,"UHV_AVE",V_VC_UHV_AVE,"V", &
                           "V_VC_UHV_AVE F1198 F1199 F1200 F1201")

  call setAttribute(outccd,"BPX_OCV",setup_OCV,&
                   comment="BadPixelMap + Offset Correction Vector Setup")

  call setAttribute(outccd,"SETUPSAS",setup_SAS,&
                   comment="ODF HK CCF input status words")

  call setAttribute(outccd,"OT_START",toffs, &
                           "Start Time (UTC) of offset table calculation")
  call setAttribute(outccd,"OT_STOP ",toffe, &
                           "End Time (UTC) of offset table calculation")

  call setAttribute(outccd,"OT_CHOP",OT_CHOP,"", &
                           "An_CHOP   F1534 F1634 F1734 F1834")

  call setAttribute(outccd,"OT_CMCOR",OT_CMCORR,"", &
                           "An_CMCORR F1525 F1625 F1725 F1825")

  call setAttribute(outccd,"OT_CMLOT",OT_CMLOTH,"", &
                           "An_CMLOTH F1515 F1516 F1517 F1615 F1616 F1617..")
                                    ! F1715 F1716 F1717 F1815 F1816 F1817")

  call setAttribute(outccd,"OT_MAXMI",OT_MAXMIP,"", &
                           "An_MAXMIP F1527 F1627 F1727 F1827")

  call setAttribute(outccd,"OT_MIP",OT_MIP,"", &
                           "An_MIP    F1526 F1626 F1726 F1826")

  call setAttribute(outccd,"OT_EAMIP",OT_EAMIPSEL,"", &
                           "An_EAMIPSEL F1536 F1636 F1736 F1836")

  call setAttribute(outccd,"OT_SENDM",OT_SENDMODE,&
                           "An_SENDMODE F1535 F1635 F1735 F1835")

  call setAttribute(outccd,"FROFFMAP",Frames_OffsetMap,"",&
                           "Number of frames for offset map calculation")

  call setAttribute(outccd,"MCOFFMAP",Median_OffsetMap,&
                   comment="Status of median correction for offset map calculation")

  call setAttribute(outccd,"FRAMETIM",frmtime,"ms",&
                           "nominal frame time")

  call setAttribute(outccd,"HK_TMP_E","------------------",&
                           "-- End of selected list of HK parameters ------")

  end subroutine writeHKattributes



! =============================================================================
! add offset correction vector to original amplitudes
! =============================================================================
  subroutine SAS_offsetCorrect(Quadrant,CCDid)
  integer(kind=Int16),intent(in )  :: quadrant, CCDid ! 0..3, 0..2
  integer(kind=Int32)              :: i

  ocv(0:63) = 0
  select case (setup_OCV)
    case("none")
    case("nom0")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  11
      endif
    case("nom1")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  11
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
      endif
    case("nom2")                                        ! 24.1.2000
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
      endif
    case("nom3")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 0) then
        ocv(17) =   5
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
      endif
    case("nom4")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 0) then
        ocv(17) =   5
        ocv(55) =  30
        ocv(60) =  30
        ocv(61) =  30
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
        ocv(61) =  30
      endif
    case("cal4")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 0) then
        ocv(17) =   5
        ocv(55) =  30
        ocv(60) =  30
        ocv(61) =  30
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
        ocv(61) =  30
        ocv(63) =  50
      endif
    case("nom5")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 0) then
        ocv(17) =   5
        ocv(55) =  30
        ocv(60) =  20
        ocv(61) =  10
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
        ocv(61) =  30
      endif
    case("nom6")
      if (quadrant == 0 .and. ccdid == 0) then
        ocv(18) =  16
      elseif (quadrant == 1 .and. ccdid == 1) then
        ocv(10) =   6  ! 16
        ocv(43) =  11
        ocv(60) =   6
      elseif (quadrant == 1 .and. ccdid == 2) then
        ocv(30) =   6
      elseif (quadrant == 2 .and. ccdid == 0) then
        ocv(33) =  23
        ocv(63) = 106
      elseif (quadrant == 3 .and. ccdid == 0) then
        ocv(17) =   5
        ocv(55) =  20
        ocv(60) =  20
        ocv(61) =  20
      elseif (quadrant == 3 .and. ccdid == 1) then
        ocv(26) =  76
        ocv(61) =  10
        ocv(61) =  50
      endif
    case default
      call error('wrongMode','caseDefaultError: setupbpx')
  end select
  do i = 0,ngrow-1
    arrPha(i) = arrPha(i)+ocv(arrRawX(i)-1)
  enddo
  do i = 0,nx_ccd-1
    on_offcol(i) = (ocv(i) > offcollimit)
  enddo
  i = 0
  next_offcol(i) = ( on_offcol(i+1) )
  i = nx_ccd-1
  next_offcol(i) = ( on_offcol(i-1) )
  do i = 1,nx_ccd-2
    next_offcol(i) = ( on_offcol(i-1) .or. on_offcol(i+1) )
  enddo

  end subroutine SAS_offsetCorrect



! =============================================================================
! open, read, and close `counting mode' file
! =============================================================================
  subroutine readCCX(ccxfilename,ntccx,ccxalloc,Quadrant,T_Counting_Mode)

! ccxfileName : Name of the input ODF/CCX file
! ntccx       : Number of rows in the input file

  character(len=*),   intent(in)   :: ccxfilename
  integer(kind=Int16),intent(in)   :: Quadrant
  integer(kind=Int32),intent(out)  :: ntccx
  real(kind=double),  intent(out)  :: T_Counting_Mode
  logical(kind=bool), intent(out)  :: ccxalloc

  integer(kind=Int32)              :: i,j,Stfine32,Etfine32,Dtfine32
  type(DataSetT)                   :: ccxfile
  type(TableT)                     :: ccx1table


  call message(AppMsg,Noisy,"... entering readCCX...")

  ccxfile = dataSet(ccxfilename,READ,USE_ENVIRONMENT)
  ccx1table = table(ccxfile,"PNCCX1")
  ntccx = numberOfRows(ccx1table)

!
! access CCX columns
!
  call message(AppMsg,Noisy," ")     
  call message(AppMsg,Noisy,'Accessing CCX data...')

  tempInCol = column(ccx1table,"QUADRANT",READ)
  ccx1QuadrantCol => int16Data(tempInCol)

  j = 0
  do i = 0, ntccx-1
    if (ccx1QuadrantCol(i) == Quadrant) then
      j = j+1
    endif
  enddo

  if (j == 0) then

    ccxalloc = .false.
    call message(AppMsg,Verbose,"no entries in CCX file for this Quadrant.")
    ntccx = 0

  else

    tempInCol = column(ccx1table,"STCOARSE",READ)
    ccx1StcoarseCol => int16Data(tempInCol)

    tempInCol = column(ccx1table,"STFINE",READ)
    ccx1StfineCol => uint16Data(tempInCol)

    tempInCol = column(ccx1table,"ETCOARSE",READ)
    ccx1EtcoarseCol => int16Data(tempInCol)

    tempInCol = column(ccx1table,"ETFINE",READ)
    ccx1EtfineCol => uint16Data(tempInCol)

    tempInCol = column(ccx1table,"CYCLCNTR",READ)
    ccx1NCycleCol => int16Data(tempInCol)

    ccxalloc = .true.
    allocate(ccx1Stcoarse(0:j-1))
    allocate(ccx1Stfine(0:j-1))
    allocate(ccx1Etcoarse(0:j-1))
    allocate(ccx1Etfine(0:j-1))
    allocate(ccx1NCycle(0:j-1))
    allocate(ccx1Duration(0:j-1))

    write(errstr,*) 'Total number of rows in CCX file            : ',ntccx
    call message(AppMsg,Noisy) 

    j = -1
    do i = 0, ntccx-1
      if (ccx1QuadrantCol(i) == Quadrant) then
        j = j+1      
        ccx1Stcoarse(j) = ccx1StcoarseCol(i)
        ccx1Stfine(j)   = ccx1StfineCol(i)
        ccx1Etcoarse(j) = ccx1EtcoarseCol(i)
        ccx1Etfine(j)   = ccx1EtfineCol(i)
        ccx1NCycle(j)   = ccx1NCycleCol(i)
      endif
    enddo
    ntccx = j
    T_Counting_Mode = 0.0

    write(errstr,*) 'Matching quadrant number of rows in CCX file: ',ntccx
    call message(AppMsg,Verbose)     

    do j = 0, ntccx-1
      if (ccx1Stfine(j) < 0) then
        Stfine32 = ccx1Stfine(j)+65536
      else
        Stfine32 = ccx1Stfine(j)
      endif
      if (ccx1Etfine(j) < 0) then
        Etfine32 = ccx1Etfine(j)+65536
      else
        Etfine32 = ccx1Etfine(j)
      endif
      DtFine32 = Etfine32-Stfine32

      if ((ccx1Stcoarse(j) > F1052) .or. ccx1Etcoarse(j) > F1052) then
        ccx1Duration(j) = ccx1NCycle(j)*frmtime_seconds
        write(errstr,'(a,a,I8,a,F10.4,a)') 'CCX: illegal coarse/fine times, ',&
              'duration has been estimated from NCYCLE =',ccx1NCycle(j), &
              ', FRMTIME =',frmtime,' [ms]'
        call message(AppMsg,Noisy)        
      else
        ccx1Duration(j) = ccx1Etcoarse(j)-ccx1Stcoarse(j) + Dtfine32*20.48d-6
        if (ccx1Duration(j) < 0) ccx1Duration(j) = ccx1Duration(j) + F1052
      endif

      T_COUNTING_MODE = T_COUNTING_MODE + ccx1Duration(j)

      if (show_CCX) then
!       write(errstr,'(a,2i6,a,2i6,a,f14.5,a,i5,a,i5)') &
        write(errstr,'(a,2i6,a,2i6,a,f14.5,a,i5,a,f8.2)') &
              'CCX: Start, End, Duration: ',&
               ccx1Stcoarse(j),Stfine32,' - ', &
               ccx1Etcoarse(j),Etfine32,' : ', &
               ccx1Duration(j), ' [s], Cycles nominal: ', &
               ccx1NCycle(j), ', computed: ', &
                   ccx1Duration(j)/frmtime_seconds  
!                  nint(ccx1Duration(j)/frmtime_seconds)  
        call message(AppMsg,Sparse)
      endif
    enddo

    write(errstr,'(a,i1,a,f14.5,a)') 'CCX: Total time of Quadrant ', &
                Quadrant,' in Counting Mode         : ',T_Counting_Mode,' [s]'
    call message(AppMsg,Verbose)     
    call message(AppMsg,Verbose," ")     
  endif

! call release(ccx1QuadrantCol)
! if (ccxalloc) then
!   call release(ccx1StcoarseCol)
!   call release(ccx1StfineCol)
!   call release(ccx1EtcoarseCol)
!   call release(ccx1EtfineCol)
!   call release(ccx1NCycleCol)
! endif
  call release(ccx1table)
  call release(ccxfile)

  end subroutine readCCX



! =============================================================================
! return ODF offset error of PHA column, add this value to PHA column later
! Developer's Note: currently it is not planned to reprocess ODFs that have
!                   been incorrectly created. Therefore this task has to
!                   take care of (at least part of) these errors
! =============================================================================
  subroutine ODF_offsetPHA(phaCol,ntrow,ODF_icd,ODF_offset)

  integer(kind=Int16), dimension(:), pointer :: phaCol
  character(len=*)   , intent(in)            :: ODF_icd
  integer(kind=Int32), intent(in)            :: ntrow
  integer(kind=Int16), intent(out)           :: ODF_offset
  integer                                    :: PHAmin,PHAmax
  logical                                    :: PHAwrong,ODFwrong

  call message(AppMsg,Noisy,"... entering ODF_offsetPHA...")

  PHAwrong = .false.
!
! access PHA column
!
  PHAmin = MINVAL(phaCol(0:ntrow-1))
  PHAmax = MAXVAL(phaCol(0:ntrow-1))

  call message(AppMsg,Verbose,' ')
  write(errstr,*) 'ODF PHA range: [',PHAmin,',',PHAmax,'] adu'
  call message(AppMsg,Verbose)

  if (PHAmin <  16) PHAwrong = .true.
  if (PHAmin > 128) PHAwrong = .true.

  select case (ODF_icd)
    case ('0043_0119700301_PNU002') ! TI
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0043_0119700301_PNU003') ! TI
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0043_0119700301_PNU004') ! TI
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300801_PNU004') ! BU
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0031_0116890801_PNU002') ! IM
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0032_0117500201_PNU002')
      ODFwrong  = .true.
      ODF_offset = -356
    case ('0033_0117710201_PNU002')
      ODFwrong  = .true.
      ODF_offset = -356
    case ('0033_9999990040_PNU002')
      ODFwrong  = .true.
      ODF_offset = -356
    case ('0042_0119100201_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100201_PNU003')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100201_PNU004')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100201_PNU017')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100201_PNU030')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU015')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU028')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU041')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU054')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU061')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0042_0119100301_PNU062')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300201_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300501_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300601_PNU014')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300801_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300801_PNU003')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300801_PNU005')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_0120300801_PNU006')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0045_9999990052_PNU005')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0046_0120900201_PNU014')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0046_0120900201_PNU019')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0046_0120900201_PNU032')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0046_0120900201_PNU033')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0049_0094810201_PNU002')
      ODFwrong  = .true.
      ODF_offset = -256
    case ('0051_0114090101_PNS006')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0052_0099760201_PNS003')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0026_0116160601_PNU002') ! pseudo-transparent mode
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0021_0115690201_PNU002') ! reprocessing at Vilspa incorrect
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0021_0115690201_PNU015')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0022_0115740201_PNS001')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0022_0115740201_PNU014')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0022_9999990175_PNU002')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0044_0119710201_PNS001') ! reprocessing at Vilspa incorrect
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0044_0119710201_PNS008')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0044_0119710201_PNS009')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0044_0119710201_PNU002')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0045_0120300101_PNS005')
      ODFwrong  = .true.
      ODF_offset = +256
    case ('0227_0022940101_PNU002') ! wrong commanding
      ODFwrong  = .true.
      ODF_offset = +256
    case default
      ODFwrong = .false.
      ODF_offset = 0
  end select

  if (ODFwrong .NEQV. PHAwrong) then
    write(errstr,'(a,i6,a,i5,a,a,a,i2)') 'ODF has minimum PHA of ',PHAmin, &
        ' but library contains offset of ',ODF_offset, &
        '. No correction applied. Please contact developer. ODF = ',&
           ODF_icd,' CCD = ',CCDNo
    call warning('PHAoffset')
    ODF_offset = 0
  endif

  if (ODFwrong) then
    write(errstr,'(a,sp,i5,a)') ' ODF PHA values will be shifted by ', &
             ODF_offset,' [adu]'
    call message(AppMsg,Sparse)
  endif

  call message(AppMsg,Verbose,' ')


  end subroutine ODF_offsetPHA



  subroutine SourcePositionOnDetector(sourceDir,AttTimeTag,xrawsrc,yrawsrc)

    type(EquatorialDirectionType), dimension(0:0), intent(in) :: sourceDir
    real(kind=double),             intent(in)  :: AttTimeTag
    integer,                       intent(out) :: xrawsrc, yrawsrc

    type(SpacecraftAttitudeType)               :: scAtt
    type(BoresightT)                           :: boresight
    real(kind=double), dimension(:), pointer   :: theta,phi, x,y,z
    integer(kind=int16), dimension(:), pointer :: chipX, chipY

    xrawsrc = 0
    yrawsrc = 0

    if ( OAL_getAttitude(AttTimeTag, scAtt) ) then
      boresight = CAL_getBoresight(AttTimeTag)

      call CAL_withSacCoordOrientation(boresight, scAtt)
      call message(AppMsg,Noisy,' ')
      write(errstr,*) 'AttTimeTag,SCATT = ',AttTimeTag,scAtt
      call message(AppMsg,Noisy)

      call CAL_projectOntoTelCoord(boresight, sourceDir, theta, phi)
      write(errstr,*) 'SOURCEDIR, THETA, PHI =',sourceDir, theta, phi
      call message(AppMsg,Noisy)

      call CAL_telCoordToCamCoord2(theta, phi, x, y, z)
      write(errstr,*) 'THETA, PHI, X, Y, Z = ',theta, phi, x, y, z
      call message(AppMsg,Noisy)

      call CAL_camCoord2ToCamCoord1(x, y, z, x, y, z)
      write(errstr,*) 'X, Y, Z = ',x, y, z
      call message(AppMsg,Noisy)

      call CAL_camCoord1ToChipCoord(x, y, z, chipX, chipY)
      write(errstr,*) 'chipX, chipY = ',chipX, chipY
      call message(AppMsg,Noisy)
      call message(AppMsg,Noisy,' ')

      xrawsrc = chipX(0)
      yrawsrc = chipY(0)
      call CAL_releaseMemory(theta, phi)
      call CAL_releaseMemory(x, y, z)
      call CAL_releaseMemory(chipX, chipY)
    endif

  end subroutine SourcePositionOnDetector



  subroutine SAS_DLIanalysis(dliarray,evtarray,An_SENDMODE, An_EAMIPSEL, &
                             dli_mean,dli_median,dli_var,dli_sigma)

    integer(kind=int32), intent(in), dimension(:) :: dliarray(1:EPN_MAX_Y), &
                                                     evtarray(1:EPN_MAX_Y)
    integer(kind=int32), intent(in)               :: An_EAMIPSEL
    character(len=*),   intent(in)                :: An_SENDMODE
    real(kind=double), intent(out)                :: dli_mean, dli_median, &
                                                     dli_var, dli_sigma
    real(kind=double)                             :: evt_mean, evt_median, &
                                                     evt_var, evt_sigma
    real(kind=double),  dimension(:), allocatable :: dliarrd, evtarrd
    integer(kind=int32), dimension(:), allocatable :: dlitmp
    integer                                       :: i,nn,mm,kk,iw
    real, parameter                               :: sigma_max = 5.5
    real                                          :: rd, re 
    logical                                       :: dli_warning
    character(len=1024)                           :: dli_warning_text(1:2)
    data dli_warning_text / 'Probably pixel above MIP threshold', &
                            'Probably source with extreme pile-up'/

    dli_sum    = 0
    dli_mean   = 0
    dli_median = 0
    dli_var    = 0
    dli_sigma  = 0
    dli_warning = .false.
    evt_sum    = 0
    if (trim(An_SENDMODE) == "MIP CORREC.2") then
      mm = EPN_MAX_X
    elseif (trim(An_SENDMODE) == "MIP CORREC.1") then
      mm = EPN_MAX_Y
!     call message(AppMsg,Verbose,' DLIanalysis: row-wise MIP rejection is unusual.')
    else
      return
    endif
    kk = mm-An_EAMIPSEL
    nn = kk-An_EAMIPSEL

    allocate(dliarrd(1:nn),evtarrd(1:nn),dlitmp(1:mm))
    do i = 1+An_EAMIPSEL,nn+An_EAMIPSEL
      dli_sum = dli_sum + dliarray(i)
      evt_sum = evt_sum + evtarray(i)
    enddo

    call message(AppMsg,Sparse,' ')
    write(errstr,*) 'EPN DLI analysis: MIPSEL = ',An_EAMIPSEL
    call message(AppMsg,Sparse)
    if (verbosity(AppMsg) > Verbose) then
      do i=1,mm
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
    else
      do i=1,An_EAMIPSEL+3
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
      do i=nn+An_EAMIPSEL-2,mm
        write(errstr,'(1x,a,i7,i7,i7)') '  DLI for X =',i,dliarray(i),evtarray(i)
        call message(AppMsg,Sparse)
      enddo
    endif
    dli_mean = dble(dli_sum)/dble(nn)
    dliarrd(1:nn) = dliarray(1+An_EAMIPSEL:nn+An_EAMIPSEL)
    call median0(dliarrd,nn,dli_median)
    dli_var = sqrt(dli_median)
    if (dli_var > 0) then
      dli_sigma = abs(dli_mean-dli_median)/dli_var
    else
      dli_sigma = 0.0
    endif

    evt_mean = dble(evt_sum)/dble(nn)
    evtarrd(1:nn) = evtarray(1+An_EAMIPSEL:nn+An_EAMIPSEL)
    call median0(evtarrd,nn,evt_median)
    evt_var = sqrt(evt_median)
    if (evt_var > 0) then
      evt_sigma = abs(evt_mean-evt_median)/evt_var
    else
      evt_sigma = 0.0
    endif

    write(errstr,'(1x,a,i7,a,f7.1,a,f7.1,a,f7.3)') 'DLI analysis: sum = ', &
                 dli_sum,'  mean = ',dli_mean,'  median = ',&
                 dli_median,'  deviation = ',dli_sigma
    call message(AppMsg,Sparse)
    write(errstr,'(1x,a,i7,a,f7.1,a,f7.1,a,f7.3)') 'EVT analysis: sum = ', &
                 evt_sum,'  mean = ',evt_mean,'  median = ',&
                 evt_median,'  deviation = ',evt_sigma
    call message(AppMsg,Sparse)

    dlitmp(1:mm) = dliarray(1:mm)
    if (An_EAMIPSEL == 1) then
      dlitmp(1) = dlitmp(1)+dli_median*0.20
      dlitmp(mm) = dlitmp(mm)+dli_median*0.20
    elseif (An_EAMIPSEL == 2) then
      dlitmp(1) = dlitmp(1)+dli_median*0.280
      dlitmp(2) = dlitmp(2)+dli_median*0.118
      dlitmp(mm-1) = dlitmp(mm-1)+dli_median*0.118
      dlitmp(mm) = dlitmp(mm)+dli_median*0.280
    endif

    if (dli_var > 0) then
!     do i=1+An_EAMIPSEL,nn+An_EAMIPSEL
      do i=1,mm
        rd = abs(dlitmp(i)-dli_median)/dli_var
        if (i.le.An_EAMIPSEL.or.i.gt.kk) then
          rd = rd*0.3
          if (i.lt.An_EAMIPSEL.or.i.gt.kk+1) rd = rd*0.4
        endif
        if (rd > sigma_max) then
          if (.not.dli_warning) call message(AppMsg,Sparse,' ')
          dli_warning = .true.
          re = (evtarray(i)-evt_median)/dli_var
          if (re > sigma_max) then
            iw = 2
          else
            iw = 1
          endif
          write(errstr,*) ' X =',i,'  DLI =',dlitmp(i),'  EVT =',evtarray(i), &
                          '  RD =',nint(rd),'  RE =',nint(re), &
                          '  Reason: ',trim(dli_warning_text(iw))
          call message(AppMsg,Sparse)
        endif
      enddo
    endif
!   if (dli_warning) call warning('DLIanalysis ', &
    if (dli_warning) call message(AppMsg,Sparse,' DLIanalysis: '// &
         'Large inhomogeneity in DLI map. '// &
         'Exposure entries of whole quadrant may be incorrect.')

    call message(AppMsg,Sparse,' ')
    deallocate(dliarrd,evtarrd,dlitmp)

  end subroutine SAS_DLIanalysis



SUBROUTINE median0(x, n, xmed)

! Find the median of X(1), ... , X(N), using as much of the quicksort
! algorithm as is needed to isolate it.
! N.B. On exit, the array X is partially ordered.
!
!     Latest revision - 26 November 1996 [alan miller]
!     minor adaption for XMMSAS 24-05-02 [mjf]
!
IMPLICIT NONE

INTEGER, INTENT(IN)                            :: n
REAL(kind=double), INTENT(INOUT), DIMENSION(:) :: x(1:n)
REAL(kind=double), INTENT(OUT)                 :: xmed

! Local variables

REAL(kind=double)    :: temp, xhi, xlo, xmax, xmin
LOGICAL              :: odd
INTEGER              :: hi, lo, nby2, nby2p1, mid, i, j, k

nby2 = n / 2
nby2p1 = nby2 + 1
odd = .true.
xmed = 0.0

! HI & LO are position limits encompassing the median.

IF (n == 2 * nby2) odd = .false.
lo = 1
hi = n
IF (n < 3) THEN
  IF (n < 1) THEN
    xmed = 0.0
    RETURN
  END IF
  xmed = x(1)
  IF (n == 1) RETURN
  xmed = 0.5*(xmed + x(2))
  RETURN
END IF

! Find median of 1st, middle & last values.

10 mid = (lo + hi)/2
xmed = x(mid)
xlo = x(lo)
xhi = x(hi)
IF (xhi < xlo) THEN          ! Swap xhi & xlo
  temp = xhi
  xhi = xlo
  xlo = temp
END IF
IF (xmed > xhi) THEN
  xmed = xhi
ELSE IF (xmed < xlo) THEN
  xmed = xlo
END IF

! The basic quicksort algorithm to move all values <= the sort key (XMED)
! to the left-hand end, and all higher values to the other end.

i = lo
j = hi
50 DO
  IF (x(i) >= xmed) EXIT
  i = i + 1
END DO
DO
  IF (x(j) <= xmed) EXIT
  j = j - 1
END DO
IF (i < j) THEN
  temp = x(i)
  x(i) = x(j)
  x(j) = temp
  i = i + 1
  j = j - 1

!     Decide which half the median is in.

  IF (i <= j) GO TO 50
END IF

IF (.NOT. odd) THEN
  IF (j == nby2 .AND. i == nby2p1) GO TO 130
  IF (j < nby2) lo = i
  IF (i > nby2p1) hi = j
  IF (i /= j) GO TO 100
  IF (i == nby2) lo = nby2
  IF (j == nby2p1) hi = nby2p1
ELSE
  IF (j < nby2p1) lo = i
  IF (i > nby2p1) hi = j
  IF (i /= j) GO TO 100

! Test whether median has been isolated.

  IF (i == nby2p1) RETURN
END IF
100 IF (lo < hi - 1) GO TO 10

IF (.NOT. odd) THEN
  xmed = 0.5*(x(nby2) + x(nby2p1))
  RETURN
END IF
temp = x(lo)
IF (temp > x(hi)) THEN
  x(lo) = x(hi)
  x(hi) = temp
END IF
xmed = x(nby2p1)
RETURN

! Special case, N even, J = N/2 & I = J + 1, so the median is
! between the two halves of the series.   Find max. of the first
! half & min. of the second half, then average.

130 xmax = x(1)
DO k = lo, j
  xmax = MAX(xmax, x(k))
END DO
xmin = x(n)
DO k = i, hi
  xmin = MIN(xmin, x(k))
END DO
xmed = 0.5*(xmin + xmax)

RETURN
END SUBROUTINE median0

end module epframes_mod
